/*
    Copyright (C) 2013 Nicola L.C. Talbot
    www.dickimaw-books.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
package com.dickimawbooks.makejmlrbookgui;

import java.util.*;
import java.io.*;
import java.util.regex.*;
import java.text.*;
import java.net.URL;

import javax.swing.JOptionPane;
import javax.swing.tree.*;

import com.dickimawbooks.texparserlib.*;
import com.dickimawbooks.texparserlib.aux.*;
import com.dickimawbooks.texparserlib.html.*;
import com.dickimawbooks.texparserlib.latex.*;
import com.dickimawbooks.texparserlib.latex.jmlr.*;

/**
 * Represents a LaTeX document containing <code>\documentclass</code>.
*/

public class BookDocument extends BookFile
{
   public BookDocument(MakeJmlrBookGUI application, FileElement parent, String name)
   {
      super(application, parent, name, false);

      packageList = new Hashtable<String,Calendar>();

      if (name.endsWith(".tex") || name.endsWith(".ltx"))
      {
         basename = name.substring(0, name.lastIndexOf("."));
      }
      else
      {
         basename = name;
         name += ".tex";
      }

      setName(name);

      importNode = new BookImportNode(this);
   }

   public BookDocument toBookDocument()
   {
      return this;
   }

   public TreeNode getImportNode(Book book)
   {
      importNode.setBook(book);
      return importNode;
   }

   public TreeNode getImportNode()
   {
      return importNode;
   }

   public void clearErrorMessages()
   {
      errorMessages   = null;
      warningMessages = null;
      numBibErrors    = 0;
      numBibWarnings  = 0;
      numErrors       = 0;
      numUndefRef     = 0;
      numUndefChar    = 0;
      useBookmark     = false;
      useBookmark     = false;
   }

   public void clearWarnings()
   {
      warningMessages = null;
      numUndefRef     = 0;
   }

   public String getInformationLink(String label)
   {
      String content = null;

      if (getApplication().showDetailsIcon())
      {
         String location = "/toolbarButtonGraphics/general/Information16.gif";

         URL imageURL = getClass().getResource(location);

         if (imageURL == null)
         {
            MakeJmlrBookGUI.debug("Can't find resource '"+location
              +"' switching to text only setting for info link");

            // switch to text only setting

            getApplication().getProperties().setShowDetailsText(true);
            getApplication().getProperties().setShowDetailsIcon(false);
         }
         else
         {
            content = String.format(
             "<img alt=\"%s\" border=\"0\" src=\"%s\"/>",
             MakeJmlrBookGUI.getLabel("info.alt"),
             imageURL.toString());
         }
      }

      if (getApplication().showDetailsText())
      {
         if (content == null)
         {
            content = MakeJmlrBookGUI.getLabel("info.text");
         }
         else
         {
            content += " "+MakeJmlrBookGUI.getLabel("info.text");
         }
      }

      if (content == null) content = "?";

      return String.format(" <a href=\"info:%s\">%s</a>",
         escapeHTML(label), content);
   }

   public String escapeHTML(String text)
   {
      int n = text.length();

      StringBuilder builder = new StringBuilder(n > 10 ? n : 10);

      for (int i = 0; i < n; i++)
      {
         char c = text.charAt(i);

         if (c == '&')
         {
            builder.append("&amp;");
         }
         else if (c == '<')
         {
            builder.append("&lt;");
         }
         else if (c == '>')
         {
            builder.append("&gt;");
         }
         else
         {
            builder.append(c);
         } 
      }

      return builder.toString();
   }

   public void addErrorMessage(String mess)
   {
      if (errorMessages == null)
      {
         errorMessages = mess;
      }
      else if (errorMessages.indexOf(mess) == -1)
      {
         errorMessages += "\n" + mess;
      }
   }

   public String getErrorMessages()
   {
      return errorMessages;
   }

   public void addWarningMessage(String mess)
   {
      if (!mess.contains("Warning:"))
      {
         mess = MakeJmlrBookGUI.getLabelWithValue("warning.message", mess);
      }

      if (warningMessages == null)
      {
         warningMessages = mess;
      }
      else if (warningMessages.indexOf(mess) == -1)
      {
         warningMessages += "\n" + mess;
      }
   }

   public String getWarningMessages()
   {
      return warningMessages;
   }

   private String removeControlSequences(String text)
   {
      Pattern p = Pattern.compile("(.*)\\\\[a-zA-Z]+\\s*(.*)");

      Matcher m = p.matcher(text);

      while (m.matches())
      {
         text = m.group(1)+m.group(2);

         m = p.matcher(text);
      }

      return text;
   }

   private String removeControlChars(String text)
   {
      Pattern p = Pattern.compile("(.*)\\\\[^a-zA-Z](.*)");

      Matcher m = p.matcher(text);

      while (m.matches())
      {
         text = m.group(1)+m.group(2);

         m = p.matcher(text);
      }

      return text;
   }

   public void setAuthor(AuxData authorData, TeXParser parser)
   {
      setShortAuthor(new TeXHtmlData(app, authorData.getArg(0), parser));

      TeXHtmlData data = new TeXHtmlData(app, authorData.getArg(1), parser);
      setAuthor(data);
   }

   public void setShortAuthor(TeXHtmlData authorName)
   {
      shortAuthor = authorName;
   }

   public void setAuthor(TeXHtmlData authorName)
   {
      author = authorName;

      String html = authorName.getHtml();

      Matcher m = SPAN_AUTHOR_PATTERN.matcher(html);

      if (m.matches())
      {
         setTag(m.group(1));
      }
      else
      {
         m = NOSPAN_AUTHOR_PATTERN.matcher(html);

         if (m.matches())
         {
            setTag(m.group(1));
         }
         else
         {
            int idx = html.indexOf(' ');

            if (idx > -1)
            {
               setTag(html.substring(0,idx));
            }
            else
            {
               setTag(html);
            }
         }
      }
   }

   public void setAuthor(String authorName)
   {
      setAuthor(new TeXHtmlData(app, authorName));
   }

   public TeXHtmlData getAuthor()
   {
      return author;
   }

   public TeXHtmlData getShortAuthor()
   {
      return shortAuthor;
   }

   public String getTeXAuthor()
   {
      return author == null ? "" : author.getTeXString();
   }

   public String getTeXShortAuthor()
   {
      return shortAuthor == null ? getTeXAuthor() : shortAuthor.getTeXString();
   }

   public String getBibAuthor(TeXParser parser)
   {
      if (author == null) return "";

      TeXObject data = author.getTeXObject();

      if (!(data instanceof TeXObjectList))
      {
         return author.getTeXString();
      }

      TeXObjectList list = (TeXObjectList)data;

      TeXObjectList biblist = new TeXObjectList();

      TeXObjectList fullname = null;
      TeXObjectList name = null;

      for (int i = 0, n = list.size(); i < n; i++)
      {
         TeXObject object = list.get(i);

         if (object instanceof CharObject
           && ((CharObject)object).getCharCode() == (int)',')
         {
            CharObject command = (CharObject)object;

            i++;

            while (i < n)
            {
               object = list.get(i);

               if (!(object instanceof WhiteSpace))
               {
                  break;
               }

               i++;
            }

            if (i >= n)
            {
               break;
            }

            if (object instanceof ControlSequence
             && ((ControlSequence)object).getName().toLowerCase().equals("and"))
            {
            }
            else if (object instanceof CharObject
                  && ((CharObject)object).getCharCode() == (int)'a')
            {
               if (i+3 >= n)
               {
                  TeXObject next1 = list.get(i+1);
                  TeXObject next2 = list.get(i+2);
                  TeXObject next3 = list.get(i+3);

                  if ((next1 instanceof CharObject)
                  && ((CharObject)next1).getCharCode() == (int)'n'
                  && (next2 instanceof CharObject)
                  && ((CharObject)next2).getCharCode() == (int)'d'
                  && (next3 instanceof WhiteSpace))
                  {
                     i += 3;
                  }
                  else
                  {
                     i--;
                  }
               }
               else
               {
                  i--;
               }
            }
            else
            {
               i--;
            }

            if (fullname == null)
            {
               fullname = name;
               name = null;
            }

            if (fullname != null)
            {
               if (name != null)
               {
                  fullname.add(name);
               }

               if (!biblist.isEmpty())
               {
                  biblist.add(parser.getListener().createString(" and "));
               }

               if (fullname.size() > 1)
               {
                  biblist.add(fullname.get(fullname.size()-1));
                  biblist.add(parser.getListener().getOther(','));

                  for (int j = 0, m = fullname.size()-1; j < m; j++)
                  {
                     biblist.add(parser.getListener().getSpace());
                     biblist.add(fullname.get(j));
                  }
               }
               else
               {
                  biblist.add(fullname);
               }

               fullname = null;
               name = null;
            }

         }
         else if (object instanceof WhiteSpace)
         {
            WhiteSpace space = (WhiteSpace)object;
            if (i == n-1) break;

            object = list.get(++i);

            while ((i < n-1) && object instanceof WhiteSpace)
            {
               object = list.get(++i);
            }

            if (object instanceof WhiteSpace)
            {
               break;
            }

            if ((object instanceof CharObject)
             && ((CharObject)object).getCharCode() == (int)',')
            {
               i--;
               continue;
            }

            if (name == null)
            {
               name = new TeXObjectList();
               name.add(object);
               continue;
            }

            String nameStr = name.toString(parser);

            if (nameStr.endsWith(" and"))
            {
               name.setSize(name.size()-4);

               if (fullname == null)
               {
                  fullname = new TeXObjectList();
               }

               fullname.add(name);
               nameStr = "and";
            }

            if (nameStr.toLowerCase().matches("\\\\?and"))
            {
               if (fullname != null)
               {
                  if (!biblist.isEmpty())
                  {
                     biblist.add(parser.getListener().createString(" and "));
                  }

                  if (fullname.size() > 1)
                  {
                     biblist.add(fullname.get(fullname.size()-1));
                     biblist.add(parser.getListener().getOther(','));

                     for (int j = 0, m = fullname.size()-1; j < m; j++)
                     {
                        biblist.add(parser.getListener().getSpace());
                        biblist.add(fullname.get(j));
                     }
                  }
                  else
                  {
                     biblist.add(fullname);
                  }

                  fullname = null;
               }

               name = new TeXObjectList();
               name.add(object);
            }
            else
            {
               if (!name.isEmpty())
               {
                  TeXObject obj = name.get(0);

                  if (obj instanceof CharObject
                   && Character.isLowerCase(((CharObject)obj).getCharCode()))
                  {
                     name.add(space);
                     name.add(object);
                     continue;
                  }
               }

               if (fullname == null)
               {
                  fullname = new TeXObjectList();
               }

               fullname.add(name);

               name = new TeXObjectList();
               name.add(object);
            }
         }
         else if (name == null)
         {
            name = new TeXObjectList();
            name.add(object);
         }
         else
         {
            name.add(object);
         }
      }

      if (name != null)
      {
         if (fullname == null)
         {
            fullname = new TeXObjectList();
         }

         fullname.add(name);
      }

      if (fullname != null)
      {
         if (!biblist.isEmpty())
         {
            biblist.add(parser.getListener().createString(" and "));
         }

         if (fullname.size() > 1)
         {
            biblist.add(fullname.get(fullname.size()-1));
            biblist.add(parser.getListener().getOther(','));

            for (int i = 0, m = fullname.size()-1; i < m; i++)
            {
               biblist.add(parser.getListener().getSpace());
               biblist.add(fullname.get(i));
            }
         }
         else
         {
            biblist.add(fullname);
         }
      }

      return biblist.toString(parser);
   }

   public void writeBibFieldStart(String entryTag, PrintWriter out)
     throws IOException
   {
      out.println(String.format("@InProceedings{%s,", entryTag));
   }

   public String getBibTitleField()
   {
      return "title";
   }

   public String getBibAuthorField()
   {
      return "author";
   }

   public String writeBibEntry(String entryTag, String crossref,
     PrintWriter out)
     throws IOException
   {
      AbstractParser listener = new AbstractParser
         (new BookTeXApp(getApplication()));

      TeXParser parser = new TeXParser(listener);
      parser.parse(getFile());

      writeBibFieldStart(entryTag, out);

      out.println(String.format("  %s = {%s},", 
        getBibTitleField(),
        listener.getTitle()));

      String bibAuthors = listener.getBibAuthors();

      out.print(String.format("  %s = {%s}",
         getBibAuthorField(),
         bibAuthors));

      if (getStartPage() > 0)
      {
         out.print(String.format(",%n  pages = {%d-%d}",
            getStartPage(), getEndPage()));
      }

      if (crossref != null)
      {
         out.print(String.format(",%n  crossref = {%s}", crossref));
      }

      String abs = listener.getAbstract();

      if (abs.isEmpty())
      {
         out.println();
      }
      else
      {
         out.println(String.format(",%n  abstract = {%s}", abs));
      }

      out.println("}");

      return bibAuthors;
   }

   public String getHtmlAuthor()
   {
      return author == null ? "" : author.getHtml();
   }

   public String getHtmlShortAuthor()
   {
      return shortAuthor == null ? getHtmlAuthor() : shortAuthor.getHtml();
   }

   public String getTag()
   {
      return importTag;
   }

   public void setTag(String tag)
   {
      int idx = tag.indexOf('\'');

      if (idx > -1)
      {
         importTag = tag.substring(0,idx)+tag.substring(idx+1).toLowerCase();
      }
      else
      {
         importTag = tag.toLowerCase();
      }

      String str = Normalizer.normalize(importTag, Normalizer.Form.NFD);
      importTag = str.replaceAll("\\p{M}", "");
   }

   public String getLabel()
   {
      return label;
   }

   protected void setLabel(String label)
   {
      this.label = label;
   }

   public String getBaseName()
   {
      return basename;
   }

   public String getAuxFileName(boolean printVersion, String ext)
   {
      return basename + (printVersion ? "-print." : ".") + ext;
   }

   public File getAuxFile()
   {
      return getAuxFile(false);
   }

   public File getAuxFile(boolean printVersion)
   {
      return new File(getLocation(), 
         getAuxFileName(printVersion, "aux"));
   }

   public File getLogFile()
   {
      return getLogFile(false);
   }

   public File getLogFile(boolean printVersion)
   {
      return new File(getLocation(), 
         getAuxFileName(printVersion, "log"));
   }

   public File getLogFile(String jobname)
   {
      return jobname == null 
        ? getLogFile()
        : new File(getLocation(), jobname+".log");
   }

   public File getPdfFile()
   {
      return getPdfFile(false);
   }

   public File getPdfFile(boolean printVersion)
   {
      return new File(getLocation(), 
         getAuxFileName(printVersion, "pdf"));
   }

   public File getTocFile()
   {
      return getTocFile(false);
   }

   public File getTocFile(boolean printVersion)
   {
      return new File(getLocation(), 
         getAuxFileName(printVersion, "toc"));
   }

   public File getBlgFile()
   {
      return getBlgFile(false);
   }

   public File getBlgFile(boolean printVersion)
   {
      return new File(getLocation(), 
         getAuxFileName(printVersion, "blg"));
   }

   public File getBblFile()
   {
      return getBblFile(false);
   }

   public File getBblFile(boolean printVersion)
   {
      return new File(getLocation(), 
         getAuxFileName(printVersion, "bbl"));
   }

   public File getBibFile()
   {
      if (bibdata == null) return null;

      return new File(getLocation(), bibdata+".bib");
   }

   public boolean isRunRequired()
   {
      return isRunRequired(false);
   }

   public boolean isRunRequired(long pdfMod)
   {
      return isRunRequired(false, pdfMod);
   }

   public boolean isRunRequired(boolean printVersion)
   {
      return isRunRequired(printVersion, true);
   }

   public boolean isRunRequired(boolean printVersion, boolean checkRerunTo)
   {
      File pdfFile = getPdfFile(printVersion);

      if (!pdfFile.exists())
      {
         app.debug("Run required, pdf file doesn't exist. pdf: "
            +pdfFile.getAbsolutePath());
         return true;
      }

      return isRunRequired(printVersion, pdfFile.lastModified(), checkRerunTo);
   }

   public boolean isRunRequired(boolean printVersion, long pdfMod)
   {
      return isRunRequired(printVersion, pdfMod, true);
   }

   public boolean isRunRequired(boolean printVersion, long pdfMod, boolean checkRerunTo)
   {
      if (undefCites || numUndefRef > 0)
      {
         return true;
      }

      File mainFile = getFile();

      if (mainFile.lastModified() > pdfMod)
      {
         if (app.inDebugMode())
         {
            app.printAndLogMessage(
               "Run required, main file modified since pdf modified.\n  main:"
                +mainFile.getAbsolutePath()
               +"\n  main mod time: "+(new Date(mainFile.lastModified()))
               +"\n  pdf mod time: "+(new Date(pdfMod)));
         }

         return true;
      }

      File auxFile = getAuxFile(printVersion);

      if (!auxFile.exists())
      {
         app.debug("Run required, aux file doesn't exist.\n  aux: "
            +auxFile.getAbsolutePath());
         return true;
      }

      BufferedReader in = null;

      try
      {
         parseAux(auxFile);

         File bibFile = getBibFile();
         File blgFile = getBlgFile();

         if (bibFile != null && bibFile.exists())
         {
            if (!blgFile.exists())
            {
               app.debug("Run required, blg file doesn't exist. blg: "
                 +blgFile.getAbsolutePath());
               return true;
            }

            long blgLastMod = blgFile.lastModified();

            if (bibFile.lastModified() > blgLastMod || blgLastMod > pdfMod)
            {
               if (app.inDebugMode())
               {
                  app.printAndLogMessage(
                    "Run required: either bib file modified since blg last modified"
                    +" or blg modified since pdf modified.\n  bib: "
                    +bibFile.getAbsolutePath()
                    +"\n  bib modified: "+(new Date(bibFile.lastModified()))
                    +"\n  blg modified: "+(new Date(blgLastMod))
                    +"\n  pdf modified: "+(new Date(pdfMod)));
               }
               return true;
            }
         }

         File logFile = getLogFile(printVersion);

         if (!logFile.exists())
         {
            app.debug("Run required, log file doesn't exist. log: "
                 +logFile.getAbsolutePath());
            return true;
         }

         if (checkRerunTo)
         {
            String line;

            in = new BufferedReader(new FileReader(logFile));

            while ((line = in.readLine()) != null)
            {
               if (line.contains("Rerun to "))
               {
                  app.debug("Run required, log file contains 'Rerun to '");
                  return true;
               }
            }
         }
      }
      catch (IOException e)
      {
         app.debug("Run required, io exception");
         return true;
      }
      finally
      {
         if (in != null)
         {
            try
            {
               in.close();
            }
            catch (IOException e)
            {
            }
         }

         in = null;
      }

      return false;
   }

   public boolean isBibRunRequired()
   {
      if (!hasBibData())
      {
         return false;
      }

      File bibFile = getBibFile();
      File blgFile = getBlgFile();

      if (bibFile != null && bibFile.exists())
      {
         if (!blgFile.exists())
         {
            app.debug("bibtex required, blg file doesn't exist. blg: "
              +blgFile.getAbsolutePath());
            return true;
         }

         long blgLastMod = blgFile.lastModified();

         if (bibFile.lastModified() > blgLastMod)
         {
            if (app.inDebugMode())
            {
               app.printAndLogMessage(
                 "bibtex required: bib file modified since blg last modified.\n  bib: "
                 +bibFile.getAbsolutePath()
                 +"\n  bib modified: "+(new Date(bibFile.lastModified()))
                 +"\n  blg modified: "+(new Date(blgLastMod)));
            }
            return true;
         }
      }

      return false;
   }

   protected void removeAuxFile(File dir, String fileName)
   {
      File file = new File(dir, fileName);

      if (file.exists())
      {
         app.setInfoAndDebug(app.getLabelWithValue("message.deleting",
            file.getAbsolutePath()));
         file.delete();
      }

   }

   public void clearAux()
   {
      app.debug("Clearing aux files for "+getName());
      File dir = getLocation();

      removeAuxFile(dir, basename+".aux");

      removeAuxFile(dir, basename+"-print.aux");

      removeAuxFile(dir, basename+".bbl");

      removeAuxFile(dir, basename+"-print.bbl");

      removeAuxFile(dir, basename+".out");

      removeAuxFile(dir, basename+"-print.out");

      removeAuxFile(dir, basename+".toc");

      removeAuxFile(dir, basename+"-print.toc");

      removeAuxFile(dir, basename+".log");

      removeAuxFile(dir, basename+".blg");
   }

   protected void addIfExists(File file, Vector<File> auxList)
   {
      if (file.exists())
      {
         auxList.add(file);
      }
   }

   public void addAuxFiles(Vector<File> auxList)
   {
      File dir = getLocation();

      addIfExists(new File(dir, basename+".aux"), auxList);

      addIfExists(new File(dir, basename+"-print.aux"), auxList);

      addIfExists(new File(dir, basename+".bbl"), auxList);

      addIfExists(new File(dir, basename+"-print.bbl"), auxList);

      addIfExists(new File(dir, basename+".out"), auxList);

      addIfExists(new File(dir, basename+"-print.out"), auxList);

      addIfExists(new File(dir, basename+".toc"), auxList);

      addIfExists(new File(dir, basename+"-print.toc"), auxList);

      addIfExists(new File(dir, basename+".log"), auxList);

      addIfExists(new File(dir, basename+".blg"), auxList);
   }

   public void setBibData(String bibdata)
   {
      this.bibdata = bibdata;
   }

   public String getBibData()
   {
      return bibdata;
   }

   public boolean hasBibData()
   {
      return bibdata != null && bibstyle != null;
   }

   public String checkObsoletePackages()
   {
      String mess = null;

      for (int i = 0, n = obsoletePackages.length; i < n; i++)
      {
         String packageName = obsoletePackages[i];

         if (hasLoadedPackage(packageName))
         {
            if (packageName.equals("jmlr"))
            {
               packageName = "jmlr2e";
            }

            if (mess == null)
            {
               mess = packageName;
            }
            else
            {
               mess += ", " + packageName;
            }
         }
      }

      return mess;
   }

   public boolean isProblemPackage(String name)
   {
      for (int i = 0, n = problemPackages.length; i < n; i++)
      {
         if (problemPackages[i].equals(name))
         {
            return true;
         }
      }

      return false;
   }

   public String checkProblemPackages()
   {
      String mess = null;

      for (int i = 0, n = problemPackages.length; i < n; i++)
      {
         String packageName = problemPackages[i];

         if (hasLoadedPackage(packageName))
         {
            if (mess == null)
            {
               mess = packageName;
            }
            else
            {
               mess += "\n" + packageName;
            }
         }
      }

      return mess;
   }

   public String getJmlrClsLocation()
   {
      return jmlrClsLocation;
   }

   public String getJmlrbookClsLocation()
   {
      return jmlrbookClsLocation;
   }

   public boolean isJmlrbookClsAtLeast(int year, int month, int day)
   {
      if (jmlrbookClsYear < year) return false;

      if (jmlrbookClsYear > year) return true;

      if (jmlrbookClsMonth < month) return false;

      if (jmlrbookClsMonth > month) return true;

      return jmlrbookClsDay >= day;
   }

   public int compareJmlrCls(BookDocument doc)
   {
      if (jmlrClsYear < doc.jmlrClsYear)
      {
         return -1;
      }

      if (jmlrClsYear > doc.jmlrClsYear)
      {
         return 1;
      }

      if (jmlrClsMonth < doc.jmlrClsMonth)
      {
         return -1;
      }

      if (jmlrClsMonth > doc.jmlrClsMonth)
      {
         return 1;
      }

      if (jmlrClsDay < doc.jmlrClsDay)
      {
         return -1;
      }

      if (jmlrClsDay > doc.jmlrClsDay)
      {
         return 1;
      }

      return 0;
   }

   public void addPackage(String packageName, int year, int month, int day)
   {
      // skip bookmark package information

      if (packageName.equals("bookmark"))
      {
         return;
      }

      Calendar calendar = Calendar.getInstance();

      calendar.set(year, month, day);

      packageList.put(packageName, calendar);
   }

   public Calendar getPackageDate(String packageName)
   {
      return packageList.get(packageName);
   }

   public Hashtable<String,Calendar> getPackages()
   {
      return packageList;
   }

   public void clearPackageList()
   {
      packageList.clear();
   }

   public Vector<LaTeXFile> getUserPackages()
   {
      return userPackages;
   }

   public LaTeXFile getUserPackage(String name)
   {
      if (userPackages == null) return null;

      for (LaTeXFile f : userPackages)
      {
         if (f.getName().equals(name))
         {
            return f;
         }
      }

      return null;
   }

   public String displayJmlrClsDetails()
   {
      if (jmlrClsVersion == null)
      {
         return app.getLabel("error.unknown");
      }

      return jmlrClsVersion + " ("+jmlrClsYear+"/"+jmlrClsMonth+"/"+jmlrClsDay+")";
   }

   public String displayJmlrbookClsDetails()
   {
      if (jmlrbookClsVersion == null)
      {
         return app.getLabel("error.unknown");
      }

      return jmlrbookClsVersion
        + " ("+jmlrbookClsYear+"/"+jmlrbookClsMonth+"/"+jmlrbookClsDay+")";
   }

   public boolean hasLoadedPackage(String packageName)
   {
      return packageList.get(packageName) != null;
   }

   protected boolean logParse(BufferedReader in)
     throws IOException
   {
      return logParse(in, true);
   }

   protected boolean logParse(BufferedReader in, boolean checkUndefRef)
     throws IOException
   {
      String line = in.readLine();

      if (line == null)
      {
         return false;
      }

      boolean readmore = true;

      Matcher m = brokenPathNoSpaces.matcher(line);

      if (m.matches())
      {
         String nextLine = in.readLine();

         if (nextLine != null)
         {
            line += nextLine;
         }
         else
         {
            readmore = false;
         }
      }
      else
      {
         m = brokenPathSpaces.matcher(line);

         if (m.matches())
         {
            String nextLine = in.readLine();

            if (nextLine != null)
            {
               line += nextLine;
            }
            else
            {
               readmore = false;
            }
         }
      }

      String location;

      File dir = getLocation();

      for (int i = 0; i < logPatterns.length; i++)
      {
         Matcher matcher = logPatterns[i].matcher(line);

         try
         {
            if (matcher.matches())
            {
               switch (i)
               {
                  case LOG_PATTERN_JMLRBOOK_CLS_VERSION:
                     location = matcher.group(2);

                     if (location.startsWith("./") 
                       ||location.startsWith(".\\"))
                     {
                        location = (new File(dir, location)).getAbsolutePath();
                     }

                     jmlrbookClsLocation = location;

                     jmlrbookClsYear = Integer.parseInt(matcher.group(3));
                     jmlrbookClsMonth = Integer.parseInt(matcher.group(4));
                     jmlrbookClsDay = Integer.parseInt(matcher.group(5));
                     jmlrbookClsVersion = matcher.group(6);
                  break;
                  case LOG_PATTERN_JMLRBOOK_CLS:
                     location = matcher.group(2);

                     if (location.startsWith("./") 
                       ||location.startsWith(".\\"))
                     {
                        location = (new File(dir, location)).getAbsolutePath();
                     }

                     jmlrbookClsLocation = location;
                  break;
                  case LOG_PATTERN_JMLRBOOK_VERSION:
                     jmlrbookClsYear = Integer.parseInt(matcher.group(1));
                     jmlrbookClsMonth = Integer.parseInt(matcher.group(2));
                     jmlrbookClsDay = Integer.parseInt(matcher.group(3));
                     jmlrbookClsVersion = matcher.group(4);
                  break;
                  case LOG_PATTERN_JMLR_CLS_VERSION:
                     location = matcher.group(2);

                     if (location.startsWith("./") 
                       ||location.startsWith(".\\"))
                     {
                        location = (new File(dir, location)).getAbsolutePath();
                     }

                     jmlrClsLocation = location;
                     jmlrClsYear = Integer.parseInt(matcher.group(3));
                     jmlrClsMonth = Integer.parseInt(matcher.group(4));
                     jmlrClsDay = Integer.parseInt(matcher.group(5));
                     jmlrClsVersion = matcher.group(6);
                  break;
                  case LOG_PATTERN_JMLR_CLS:
                     location = matcher.group(2);

                     if (location.startsWith("./") 
                       ||location.startsWith(".\\"))
                     {
                        location = (new File(dir, location)).getAbsolutePath();
                     }

                     jmlrClsLocation = location;
                  break;
                  case LOG_PATTERN_JMLR_VERSION:
                     jmlrClsYear = Integer.parseInt(matcher.group(1));
                     jmlrClsMonth = Integer.parseInt(matcher.group(2));
                     jmlrClsDay = Integer.parseInt(matcher.group(3));
                     jmlrClsVersion = matcher.group(4);
                  break;
                  case LOG_PATTERN_LINE_NUMBER:

                     addErrorMessage(escapeHTML(line)
                       +getInformationLink(
                         (this instanceof Book)
                         && line.matches(".*=\\{\\d+\\}\\}\\{[0-9A-E]+.*")?
                         "book.missnum" :
                         "line_number:" +matcher.group(1)+":"+getLogFile()
                        ));
                  break;
                  case LOG_PATTERN_OBS_STY:

                     addErrorMessage(escapeHTML(line)
                       +getInformationLink("obs_sty:"+matcher.group(1)));

                     numErrors++;
                  break;
                  case LOG_PATTERN_CONFLICT_STY:

                     addErrorMessage(escapeHTML(line)
                       +getInformationLink("conflict_sty:"+matcher.group(1)));

                     numErrors++;
                  break;
                  case LOG_PATTERN_MISSING_NUMBER:
                  case LOG_PATTERN_FILE_LINE_MISSING_NUMBER:

                     String error = escapeHTML(line);

                     String infoTag;

                     if (i == LOG_PATTERN_MISSING_NUMBER)
                     {
                        infoTag = "tex_error:"+getLogFile();
                     }
                     else
                     {
                        infoTag = "file_line:"+matcher.group(1)+":"
                         + matcher.group(2)+":"+matcher.group(3);
                     }

                     String nextLine = in.readLine();

                     if (nextLine == null) return false;

                     if (nextLine.startsWith("<to be read again>"))
                     {
                        nextLine = in.readLine();

                        if (nextLine == null) return false;

                        if (nextLine.matches(" +\\{"))
                        {
                           nextLine = in.readLine();

                           if (nextLine == null) return false;

                           if (nextLine.matches(
                                ".*=\\{\\d+\\}\\}\\{[0-9A-E]+.*"))
                           {
                              clearAuxAndRerun = true;
                              addErrorMessage(app.getLabel("error.missnum")
                                + getInformationLink(
                                   this instanceof Book ? 
                                      "book.missnum" : "missnum"));
                           }
                        }
                     }

                     addErrorMessage(error
                       +getInformationLink(infoTag));

                     numErrors++;

                  break;
                  case LOG_PATTERN_STY_BEFORE_CLS:

                     addErrorMessage(escapeHTML(line)
                       +getInformationLink("sty_before_cls"));
                     numErrors++;

                  break;
                  case LOG_PATTERN_ERROR:

                     addErrorMessage(escapeHTML(matcher.group(1))
                       +getInformationLink("tex_error:"+getLogFile()));
                     numErrors++;

                  break;
                  case LOG_PATTERN_FILE_LINE:

                     addErrorMessage(escapeHTML(line)
                       +getInformationLink("file_line:"+matcher.group(1)
                        +":"+matcher.group(2)+":"+matcher.group(3)));
                     numErrors++;

                  break;
                  case LOG_PATTERN_PACKAGE_INFO:
                     addPackage(matcher.group(1),
                        Integer.parseInt(matcher.group(2)),
                        Integer.parseInt(matcher.group(3)),
                        Integer.parseInt(matcher.group(4)));
                  break;
                  case LOG_PATTERN_OLD_STYLE:
                     addPackage(matcher.group(1), 0, 0, 0);
                  break;
                  case LOG_PATTERN_UNDEFINED_REF:
                     addWarningMessage(escapeHTML(line)
                      +getInformationLink("undef.ref:"+getLogFile()));
                     numUndefRef++;
                  break;
                  case LOG_PATTERN_UNDEFINED_CHAR:
                  case LOG_PATTERN_MISSING_CHAR:
                     addErrorMessage(escapeHTML(matcher.group(1))
                      +getInformationLink("undef.char:"+getName()));
                     numUndefChar++;
                  break;
                  case LOG_PATTERN_MULTIPLE_LABEL:
                     addErrorMessage(app.getLabel("error.detected.multiply_def"));
                  break;
                  case LOG_PATTERN_UNDEFINED_CITES:
                    undefCites=true;
                  break;
                  case LOG_PATTERN_RERUN_TO:

                  if (!checkUndefRef)
                  {
                     numUndefRef = 0;
                  }
                  break;
                  case LOG_PATTERN_LENDDOCA:
                     addErrorMessage(
                       app.getLabelWithValue("error.jmlrbookcls.tooold",
                        "2012/05/16"));
                  break;
                  case LOG_PATTERN_BOOKMARK:
                     useBookmark = true;
                  break;
                  case LOG_PATTERN_FILE_NOT_FOUND:
                     addErrorMessage(
                        app.getLabelWithValue("error.file_not_found", m.group(1)));
                  break;
                  case LOG_PATTERN_EMERGENCY_STOP:
                     addErrorMessage(
                        app.getLabel("error.emergency_stop"));
                  break;

               }

               break;
            }
         }
         catch (NumberFormatException e)
         {
         }
      }

      return readmore;
   }

   public void parseAux(File auxFile)
    throws IOException
   {
      AuxParser auxParser = new AuxParser(new BookTeXApp(getApplication()))
      {
         protected void addPredefined()
         {
            super.addPredefined();

            addAuxCommand("jmlr@author", 2);
            addAuxCommand("jmlr@title", 2);
            addAuxCommand("jmlrbook@info", 2);
         }
      };

      TeXParser parser = auxParser.parseAuxFile(auxFile);

      Vector<AuxData> data = auxParser.getAuxData("bibdata");

      if (data != null && data.size() > 0)
      {
         bibdata = data.firstElement().getArg(0).toString();
      }

      data = auxParser.getAuxData("bibstyle");

      if (data != null && data.size() > 0)
      {
         bibstyle = data.firstElement().getArg(0).toString();
      }

      data = auxParser.getAuxData("jmlr@author");

      if (data != null && data.size() > 0)
      {
         setAuthor(data.firstElement(), parser);
      }

      data = auxParser.getAuxData("jmlr@title");

      if (data != null && data.size() > 0)
      {
         setTitle(data.firstElement(), parser);
      }

   }

   public void blgParse(String line)
   {
      String lineNum, bib;

      for (int i = 0; i < blgPatterns.length; i++)
      {
         Matcher matcher = blgPatterns[i].matcher(line);

         if (matcher.matches())
         {
            switch (i)
            {
               case BLG_PATTERN_ERROR:
                  numBibErrors = Integer.parseInt(matcher.group(1));
               break;
               case BLG_PATTERN_UNDEFCITE:

                  String label = matcher.group(1);

                  addWarningMessage(app.getLabelWithValue(
                     "warning.bibtex_undefcite", label)
                    +getInformationLink("bibtex_undefcite:"+label));
                  numBibWarnings++;
               break;
               case BLG_PATTERN_MISSING_FIELD:

                  String field = matcher.group(1);
                  String key = matcher.group(2);

                  addWarningMessage(app.getLabelWithValues(
                     "warning.bibtex_missing_field", field, key)
                    +getInformationLink("bibtex_missing_field:"+field+":"+key));
                  numBibWarnings++;
               break;
               case BLG_PATTERN_WARNING:
                  String warnMess = matcher.group(1);

                  addWarningMessage(app.getLabelWithValue("warning.bibtex",
                    escapeHTML(warnMess))
                    +getInformationLink("bibtex_warning:"+warnMess));
                  numBibWarnings++;
               break;
               case BLG_PATTERN_MISSING_CLOSE:

                  lineNum = matcher.group(1);
                  bib = matcher.group(2);

                  addErrorMessage(app.getLabelWithValues(
                    "error.bibtex.missing_close", lineNum, bib)
                    +getInformationLink("bibtex_missing_close:"
                      +lineNum+":"+bib));
               break;
               case BLG_PATTERN_ERROR_LINE:

                  String mess = matcher.group(1);
                  lineNum = matcher.group(2);
                  bib = matcher.group(3);

                  addErrorMessage(app.getLabelWithValues(
                    "error.bibtex.file_line",
                    new String[]{escapeHTML(mess), lineNum, bib})
                    +getInformationLink("bibtex_error:"+mess+":"
                      +lineNum+":"+bib));
               break;
            }

            break;
         }
      }

   }

   public int latex()
      throws IOException,InterruptedException
   {
      return latex(null, "");
   }

   public int latex(String jobname, String predef)
      throws IOException,InterruptedException
   {
      File file = getFile();

      File dir = file.getParentFile();

      if (!file.exists())
      {
         throw new IOException(app.getLabelWithValue("error.file_cant_find",
            file.getCanonicalPath()));
      }

      String latexOptions = app.getProperties().getLaTeXOptions();

      int idx = 1;

      String[] opts = latexOptions.split(" +");
      String[] cmd;

      if (jobname != null)
      {
         cmd = new String[opts.length+4];
         cmd[idx] = "-jobname";
         idx++;
         cmd[idx] = jobname;
         idx++;
      }
      else
      {
         cmd = new String[opts.length+2];
      }

      cmd[0] = app.getLaTeXApp();

      for (int j = 0; j < opts.length; j++)
      {
         cmd[idx] = opts[j];
         idx++;
      }

      String arg = "\\batchmode";

      if (predef != null && predef.length() > 0)
      {
         arg = predef;
      }

      if (startPage != -1)
      {
         arg += "\\def\\startpage{"+startPage+"}";
      }

      if (arg.length() == 0)
      {
         cmd[idx] = getBaseName();
      }
      else
      {
         cmd[idx] = arg+"\\input{"+getBaseName()+"}";
      }

      int exitCode = app.execCommandAndWaitFor(cmd, null, dir);

      app.debug("Exit code: "+exitCode);

      if (!getLogFile(jobname).exists())
      {
         String mess = "";

         for (int i = 1; i < cmd.length; i++)
         {
            mess += "\""+cmd[i]+"\" ";
         }

         String errMess = app.getLabelWithValues("error.exit.latex.failure.nolog",
               new String[]
               {
                  ""+exitCode,
                  cmd[0],
                  mess,
                  dir.getAbsolutePath(),
                  getLogFile(jobname).getAbsolutePath()
               });

         addErrorMessage(errMess);

         throw new IOException(errMess);
      }

      return exitCode;
   }

   public int bibtex()
      throws IOException,InterruptedException
   {
      File dir = getLocation();

      String[] opts = app.getProperties().getBibTeXOptions().split(" ");

      String[] cmd = new String[opts.length+2];

      cmd[0] = app.getBibTeXApp();

      int idx = 1;

      for (int j = 0; j < opts.length; j++)
      {
         cmd[idx] = opts[j];
         idx++;
      }

      cmd[idx] = getBaseName();

      return app.execCommandAndWaitFor(cmd, null, dir);
   }

   public boolean confirmContinue(String errMess)
   {
      return true;
/*
      return JOptionPane.showConfirmDialog(app, errMess,
         app.getLabel("error.title"), JOptionPane.OK_CANCEL_OPTION,
         JOptionPane.ERROR_MESSAGE) == JOptionPane.OK_OPTION;
*/
   }

   public void process()
      throws IOException,InterruptedException
   {
      process(false, true);
   }

   public void process(String texArg)
      throws IOException,InterruptedException
   {
      process(false, true, true, texArg);
   }

   public void process(boolean printVersion, boolean doLaTeXRun)
      throws IOException,InterruptedException
   {
      process(printVersion, doLaTeXRun, true);
   }

   public void process(boolean printVersion, boolean doLaTeXRun,
     boolean checkRerunTo)
      throws IOException,InterruptedException
   {
      process(printVersion, doLaTeXRun, checkRerunTo, "");
   }

   public void process(boolean printVersion, boolean doLaTeXRun,
     boolean checkRerunTo, String texArg)
      throws IOException,InterruptedException
   {
      File file = getFile();
      File logFile = getLogFile(printVersion);
      File auxFile = getAuxFile(printVersion);
      File blgFile = getBlgFile(printVersion);

      if (getErrorMessages() != null)
      {
         doLaTeXRun = true;
      }

      try
      {
         // Parse preamble for class and packages
         clsFile = null;
         userPackages = null;
         PreambleParser listener = new PreambleParser(new BookTeXApp(app));
         TeXParser parser = listener.parse(file);
         clsFile = listener.getDocumentClass();
         userPackages = listener.getLoadedPackages();
      }
      catch (IOException e)
      {
         app.debug(e);
      }

      String jobname = null;

      if (printVersion)
      {
         jobname = basename+"-print";
      }

      int exitCode;
      BufferedReader in = null;
      String line;

      app.checkForInterrupt();

      if (doLaTeXRun || !logFile.exists() || isRunRequired(printVersion, checkRerunTo))
      {
         clearPackageList();

         exitCode = latex(jobname,
            printVersion ? "\\def\\jmlrgrayscale{1}"+texArg : texArg);

         if (exitCode != 0)
         {
            app.checkForInterrupt();

            if (!confirmContinue(app.getLabelWithValues("error.exit.latex.cont_checklog",
               new String[]
               {
                  ""+exitCode,
                  logFile.getAbsolutePath()
               }
            )))
            {
               throw new CancelledException();
            }
         }
      }

      try
      {
         clearErrorMessages();
         undefCites = false;
         useBookmark = false;
         clearAuxAndRerun = false;

         app.checkForInterrupt();

         in = new BufferedReader(new FileReader(logFile));

         app.setInfo(app.getLabelWithValue("message.processing",
            logFile.getAbsolutePath()));

         while (logParse(in, checkRerunTo))
         {
            app.checkForInterrupt();
         }

         in.close();
         in = null;

         app.checkForInterrupt();

         if (auxFile.exists())
         {
            parseAux(auxFile);
         }

         boolean doBibtex = false;

         if (undefCites)
         {
            app.debug("Has undefined citations");

            doBibtex = true;
         }
         else
         {
            app.debug("No undefined citations");
         }

         if (hasBibData())
         {
            app.debug("Has bibdata");
            doBibtex = (doBibtex && true);
         }
         else
         {
            app.debug("No bibdata");
            doBibtex = false;
         }

         if (isBibRunRequired())
         {
            app.debug("bibtex run required");
            doBibtex = true;
         }

         if (doBibtex)
         {
            app.checkForInterrupt();

            exitCode = bibtex();

            if (exitCode != 0)
            {
               app.checkForInterrupt();

               if (!confirmContinue(app.getLabelWithValues("error.exit.bibtex.check",
                  new String[]
                  {
                     ""+exitCode,
                     blgFile.getAbsolutePath()
                  })
               ))
               {
                  throw new CancelledException();
               }
            }

            undefCites = false;
            clearWarnings();

            app.checkForInterrupt();

            in = new BufferedReader(new FileReader(logFile));

            app.setInfo(app.getLabelWithValue("message.processing",
               logFile.getAbsolutePath()));

            while (logParse(in, checkRerunTo))
            {
               app.checkForInterrupt();
            }

            in.close();
            in = null;
         }

         app.checkForInterrupt();

         if (numErrors != 0)
         {
            addErrorMessage(
               numErrors == 1 ? 
                  app.getLabel("error.detected.single") : 
                  app.getLabelWithValue("error.detected.multi", numErrors)
            );
         }

         app.checkForInterrupt();

         checkCls();

         int runCount = 0;

         while (doLaTeXRun && isRunRequired(printVersion, checkRerunTo)
           && runCount <= MakeJmlrBookGUI.MAX_RERUN)
         {
            clearErrorMessages();
            app.checkForInterrupt();

            runCount++;
            latex(jobname,
               printVersion ? "\\def\\jmlrgrayscale{1}"+texArg : texArg);

            numUndefRef = 0;
            numUndefChar = 0;
            undefCites = false;
            clearWarnings();

            app.checkForInterrupt();

            in = new BufferedReader(new FileReader(logFile));

            app.setInfo(app.getLabelWithValue("message.processing",
               logFile.getAbsolutePath()));

            while (logParse(in, checkRerunTo))
            {
               app.checkForInterrupt();
            }
         }

         if (hasBibData())
         {
            app.checkForInterrupt();

            if (blgFile.exists())
            {
               in = new BufferedReader(new FileReader(blgFile));

               app.setInfo(app.getLabelWithValue("message.processing",
                  blgFile.getAbsolutePath()));

               while ((line = in.readLine()) != null)
               {
                  app.checkForInterrupt();

                  blgParse(line);
               }

               in.close();

               if (numBibErrors > 0)
               {
                  if (numBibWarnings > 0)
                  {
                     if (numBibErrors > 1 && numBibWarnings > 1)
                     {
                         addErrorMessage(app.getLabelWithValues(
                            "error.detected.bibtex.bothmulti",
                            new String[] {""+numBibErrors, ""+numBibWarnings})
                         +getInformationLink("bibtex:"+getBlgFile()));
                     }
                     else if (numBibErrors > 1)
                     {
                         addErrorMessage(app.getLabelWithValue(
                            "error.detected.bibtex.multierr.singlewarn",
                            numBibErrors)
                         +getInformationLink("bibtex:"+getBlgFile()));
                     }
                     else if (numBibWarnings > 1)
                     {
                         addErrorMessage(app.getLabelWithValue(
                            "error.detected.bibtex.singleerr.multiwarn",
                            numBibWarnings)
                         +getInformationLink("bibtex:"+getBlgFile()));
                     }
                     else
                     {
                         addErrorMessage(app.getLabel(
                            "error.detected.bibtex.bothsingle")
                         +getInformationLink("bibtex:"+getBlgFile()));
                     }
                  }
                  else
                  {
                     addErrorMessage(
                        numBibErrors == 1 ?
                        app.getLabel("error.detected.bibtex.singleerr") :
                        app.getLabelWithValue("error.detected.bibtex.multierr",
                           numBibErrors)
                         +getInformationLink("bibtex:"+getBlgFile())
                     );
                  }
               }
               else if (numBibWarnings > 0)
               {
                  addErrorMessage(
                     numBibWarnings == 1 ?
                     app.getLabel("error.detected.bibtex.singlewarn") :
                     app.getLabelWithValue("error.detected.bibtex.multiwarn",
                       numBibWarnings)
                         +getInformationLink("bibtex:"+getBlgFile())
                  );
               }
            }
         }

      }
      finally
      {
         if (in != null)
         {
            in.close();
         }
      }

      if (numUndefRef > 0)
      {
         addErrorMessage(
            (numUndefRef == 1 ?
            app.getLabel("error.detected.undef.ref") :
            app.getLabelWithValue("error.detected.undef.refs",
               numUndefRef)
            )
            +getInformationLink("undef.ref:"+getLogFile())
          );
      }

      if (useBookmark && !hasLoadedPackage("bookmark"))
      {
         addErrorMessage(app.getLabel("error.detected.bookmark_required"));
      }

      if (numUndefChar > 0)
      {
         addErrorMessage(
            numUndefChar == 1 ?
            app.getLabel("error.detected.undef.char") :
            app.getLabelWithValue("error.detected.undef.chars",
               numUndefChar)
            +getInformationLink("undef.char:"+getFile())
          );
      }

      app.checkForInterrupt();

      if (errorMessages != null)
      {
         app.debug("Error messages:\n"+errorMessages);
      }
   }

   public void checkCls()
   {
      if (jmlrClsLocation == null)
      {
         addErrorMessage(app.getLabel("error.jmlrcls.notused")
          +getInformationLink("jmlrcls.notused:"+getFile()));
      }

      if (hasLoadedPackage("color"))
      {
         addWarningMessage(app.getLabel("warning.color_sty")
          +getInformationLink("color_sty"));
      }

      if (hasLoadedPackage("ucs"))
      {
         addWarningMessage(app.getLabel("warning.ucs_sty")
          +getInformationLink("ucs_sty"));
      }

      if (hasLoadedPackage("lipsum"))
      {
         addWarningMessage(app.getLabel("warning.lipsum_sty")
          +getInformationLink("lipsum_sty"));
      }

      String list = checkObsoletePackages();

      if (list != null)
      {
         if (list.contains(","))
         {
            addErrorMessage(app.getLabelWithValue("error.obsolete.stys", list));
         }
         else
         {
            addErrorMessage(app.getLabelWithValue("error.obsolete.sty", list));
         }
      }

      list = checkProblemPackages();

      if (list != null)
      {
         if (list.contains(", "))
         {
            addErrorMessage(
                app.getLabelWithValue("error.problem.stys", list));
         }
         else
         {
            addErrorMessage(
               app.getLabelWithValue("error.problem.sty", list)
            );
         }
      }
   }

   public String getPages()
   {
      String pp = ""+startPage;

      if (startPage != endPage)
      {
         pp += "--" + endPage;
      }

      return pp;
   }

   public void setStartPage(int pageNum)
   {
      startPage = pageNum;
   }

   public void setEndPage(int pageNum)
   {
      endPage = pageNum;
   }

   public int getStartPage()
   {
      return startPage;
   }

   public int getEndPage()
   {
      return endPage;
   }

   public File[] getSupplementalFiles()
   {
      return getLocation().listFiles(new java.io.FileFilter()
      {
         public boolean accept(File file)
         {
            if ((!file.isDirectory()
              && file.getName().startsWith(getBaseName()+"-supp.")))
            {
               return true;
            }

            return false;
         }
      });
   }

   public String getSourceLink()
   {
      StringBuffer buffer = new StringBuffer();

      File texFile = getFile();

      if (texFile.exists())
      {
         buffer.append(String.format(
           "<tr><td><a href=\"%s\">%s</a></td><td>%s</td></tr>",
           texFile.toURI(),
           app.getLabel("main.edit"),
           app.getLabelWithValue("main.last_modified",
                      dateFormat.format(new Date(texFile.lastModified())))));
      }
      else
      {
         buffer.append(String.format(
           "<tr><td colspan=\"2\">%s</td></tr>",
           app.getLabelWithValue("error.file_cant_find",
                      texFile.getAbsolutePath())));
      }

      File bibFile = getBibFile();

      if (bibFile != null)
      {
         if (bibFile.exists())
         {
            buffer.append(String.format(
               "<tr><td><a href=\"%s\">%s</a></td><td>%s</td></tr>",
               bibFile.toURI(),
               app.getLabel("main.bibedit"),
               app.getLabelWithValue("main.last_modified",
                         dateFormat.format(new Date(bibFile.lastModified())))));
         }
      }

      return buffer.toString();
   }

   public String getPdfLink()
   {
      String content = "";

      File pdfFile = getPdfFile();

      if (pdfFile.exists())
      {
         content = String.format(
           "<tr><td><a href=\"%s\">%s</a></td><td>%s</td></tr>",
           pdfFile.toURI(),
           app.getLabel("main.view.pdf"),
           app.getLabelWithValue("main.last_modified", 
                    dateFormat.format(new Date(pdfFile.lastModified()))));
      }

      return content;
   }

   public String getInfoLinks()
   {
      StringBuffer buffer = new StringBuffer();

      String warnMess = getWarningMessages();

      if (warnMess != null)
      {
         warnMess = warnMess.replaceAll("\n", "<br>");
         buffer.append(String.format("<font class=warning>%s</font><br>",
           warnMess));
      }

      String errMess = getErrorMessages();

      if (errMess != null)
      {
         errMess = errMess.replaceAll("\n", "<br>");

         buffer.append(String.format("<font class=error>%s</font><br>",
           errMess));

         File logFile = getLogFile();

         buffer.append("<table>");

         if (logFile.exists())
         {
            buffer.append(String.format(
              "<tr><td><a href=\"%s\">%s</a></td><td>%s</td></tr>",
              logFile.toURI(),
              app.getLabel("main.view.log"),
              app.getLabelWithValue("main.last_modified",
                       dateFormat.format(new Date(logFile.lastModified())))));
         }
         else
         {
            buffer.append(String.format("<tr><td colspan=2>%s</td></tr>",
               app.getLabelWithValue("error.file_cant_find",
                          logFile.getAbsolutePath())));
         }

         File blgFile = getBlgFile();

         if (blgFile != null && blgFile.exists())
         {
            buffer.append(String.format(
              "<tr><td><a href=\"%s\">%s</a></td><td>%s</td></tr>",
              blgFile.toURI(),
              app.getLabel("main.view.blg"),
              app.getLabelWithValue("main.last_modified",
                       dateFormat.format(new Date(blgFile.lastModified())))));
         }
      }
      else
      {
         buffer.append("<table>");
      }

      buffer.append(getPdfLink());

      buffer.append(getSourceLink());

      File[] bibFiles = getLocation().listFiles(app.getBibFileFilter());

      if (bibFiles != null)
      {
         for (File bibFile : bibFiles)
         {
            if (!bibFile.isDirectory())
            {
               buffer.append(String.format(
                  "<tr><td><a href=\"%s\">%s</a></td><td>%s</td></tr>",
                  bibFile.toURI(),
                  app.getLabelWithValue("main.view", bibFile.getName()),
                  app.getLabelWithValue("main.last_modified",
                    dateFormat.format(new Date(bibFile.lastModified())))));
            }
         }
      }

      File[] suppFiles = getSupplementalFiles();

      if (suppFiles != null && suppFiles.length > 0)
      {
         buffer.append(String.format("<tr><td>%s</td><td>", 
           app.getLabel("main.supplemental")));

         for (File suppFile : suppFiles)
         {
            buffer.append(String.format("%s<br>%n", suppFile.getName()));
         }

         buffer.append("</td></tr>");
      }

      buffer.append("</table>");

      return buffer.toString();
   }

   public String getInfoContent()
   {
      StringBuffer buffer = new StringBuffer(String.format(
        "<h2><a name=\"%s\">%s</a></h2>",
         getBaseName(),
         app.getLabelWithValue("main.import", 
                       getBaseName())));

      if (supplemental)
      {
         buffer.append(app.getLabel("main.supplemental"));
      }

      buffer.append(String.format(
       "<table><tr><th>%s</th><td>%s</td></tr>",
       app.getLabel("main.label"), getLabel()));

      buffer.append(String.format("<tr><th>%s</th><td>%s</td></tr>",
       app.getLabel("main.directory"), getLocation()));

      buffer.append(String.format("<tr><th>%s</th><td>%s</td></tr>",
       app.getLabel("main.filename"), getFile().getName()));

      buffer.append(String.format("<tr><th>%s</th><td>%s</td></tr>",
       app.getLabel("main.author"), getHtmlAuthor()));

      buffer.append(String.format("<tr><th>%s</th><td>%s</td></tr>",
       app.getLabel("main.tag"), getTag()));

      boolean artClsFound = false;

      buffer.append(String.format("<tr><th>%s</th><td>",
       app.getLabelWithValue("main.version", "jmlr.cls")));

      String location = getJmlrClsLocation();

      if (location == null)
      {
         buffer.append(String.format("<font class=error>%s</font>",
           app.getLabel("error.unknown")));
      }
      else
      {
         artClsFound = true;

         buffer.append(String.format("%s</td></tr><tr><th>%s</th><td>%s",
           displayJmlrClsDetails(),
           app.getLabel("main.location"),
           location));
      }

      buffer.append("</td></tr>");

      int startPage = getStartPage();

      if (startPage == -1)
      {
         buffer.append(String.format(
            "<tr><th>%s</th><td><font class=error>%s</font></td></tr>",
            app.getLabel("main.pages"),
            app.getLabel("error.unknown")));
      }
      else
      {
         int endPage = getEndPage();

         if (startPage == endPage)
         {
            buffer.append(String.format("<tr><th>%s</th><td>%d",
              app.getLabel("main.page"), startPage));
         }
         else
         {
            buffer.append(String.format("<tr><th>%s</th><td>%d-%d",
              app.getLabel("main.pages"), startPage, endPage));
         }

         buffer.append("</td></tr>");
      }

      buffer.append("</table>");

      buffer.append(getInfoLinks());

      return buffer.toString();
   }

   public void setTitle(AuxData titleData, TeXParser parser)
   {
      setShortTitle(new TeXHtmlData(app, titleData.getArg(0), parser));
      setTitle(new TeXHtmlData(app, titleData.getArg(1), parser));
   }

   public void setTitle(TeXHtmlData docTitle)
   {
      title = docTitle;
   }

   public void setShortTitle(TeXHtmlData docTitle)
   {
      shortTitle = docTitle;
   }

   public void setTitle(String docTitle)
   {
      setTitle(new TeXHtmlData(app, docTitle));
   }

   public TeXHtmlData getTitle()
   {
      return title;
   }

   public String getTeXTitle()
   {
      return title == null ? "" : title.getTeXString();
   }

   public String getHtmlTitle()
   {
      return title == null ? "" : title.getHtml();
   }

   public TeXHtmlData getShortTitle()
   {
      return shortTitle;
   }

   public String getTeXShortTitle()
   {
      return shortTitle == null ? getTeXTitle() : shortTitle.getTeXString();
   }

   public String getHtmlShortTitle()
   {
      return shortTitle == null ? getHtmlTitle() : shortTitle.getHtml();
   }

   public boolean isSupplemental()
   {
      return supplemental;
   }

   public void setSupplemental(boolean isSup)
   {
      supplemental = isSup;
   }

   private String errorMessages, warningMessages;

   private int numBibErrors, numBibWarnings, numErrors, numUndefRef, numUndefChar;

   protected boolean clearAuxAndRerun = false;

   private String label, basename;

   private String importTag;

   protected String bibdata, bibstyle;

   protected TeXHtmlData author, shortAuthor, title, shortTitle;

   // Package list parsed from log file.
   // This lists all packages loaded by the document, some
   // of these packages may have been implicitly loaded by
   // other packages.
   private Hashtable<String,Calendar> packageList;

   // List of packages explicitly loaded in the document preamble
   // with \\usepackage.
   protected Vector<LaTeXFile> userPackages;

   protected LaTeXFile clsFile;

   protected String jmlrClsLocation, jmlrClsVersion;

   protected int jmlrClsYear, jmlrClsMonth, jmlrClsDay;

   protected String jmlrbookClsLocation, jmlrbookClsVersion;

   protected int jmlrbookClsYear, jmlrbookClsMonth, jmlrbookClsDay;

   private int startPage=-1, endPage;

   private boolean undefCites=false;

   private boolean supplemental=false;

   private boolean useBookmark=false;

   private static final Pattern brokenPathSpaces =
     Pattern.compile(".*\\(\"[^\"]++");

   private static final Pattern brokenPathNoSpaces =
     Pattern.compile(".*\\((\\.|/)([\\w_\\-\\.]*+/?)+");

   private static final Pattern[] logPatterns = new Pattern[]
   {
      Pattern.compile(".*\\((\"?)(.*[/|\\\\]jmlrbook\\.cls)\\1.*Document Class: jmlrbook (\\d{4})/(\\d{2})/(\\d{2}) v([0-9\\.]+).*"),
      Pattern.compile(".*\\((\"?)(.*[/|\\\\]jmlrbook\\.cls)\\1.*"),
      Pattern.compile(".*Document Class: jmlrbook (\\d{4})/(\\d{2})/(\\d{2}) v([0-9\\.]+).*"),
      Pattern.compile(".*Package: ([^ ]+) (\\d{4})/(\\d{2})/(\\d{2}).*"),
      Pattern.compile(".*\\((\"?)(.*[/|\\\\]jmlr\\.cls)\\1.*Document Class: jmlr (\\d{4})/(\\d{2})/(\\d{2}) v([0-9\\.]+).*"),
      Pattern.compile(".*\\((\"?)(.*[/|\\\\]jmlr\\.cls)\\1.*"),
      Pattern.compile(".*Document Class: jmlr (\\d{4})/(\\d{2})/(\\d{2}) v([0-9\\.]+).*"),
      Pattern.compile(".*Document Style `([^ ]+)'.*"),
      Pattern.compile(".*Warning: Reference.*undefined.*"),
      Pattern.compile(".*Keyboard character used is undefined.*"),
      Pattern.compile(".*(Missing character:.*)"),
      Pattern.compile(".*There were multiply-defined labels.*"),
      Pattern.compile(".*undefined citations.*"),
      Pattern.compile(".*Rerun to .*"),
      Pattern.compile(".*Command \\\\c@lenddoca already defined.*"),
      Pattern.compile("\\(rerunfilecheck\\)\\s+or use package `bookmark'.*"),
      Pattern.compile(".*File `([^'])' not found.*"),
      Pattern.compile(".*Class jmlr Error: Obsolete package `(.*?)' detected.*"),
      Pattern.compile(".*Class jmlr Error: Package `(.*?)' detected.*"),
      Pattern.compile("! Emergency stop\\..*"),
      Pattern.compile("! Missing number, treated as zero.*"),
      Pattern.compile("! LaTeX Error: \\\\usepackage before \\\\documentclass.*"),
      Pattern.compile("(! .*)"),
      Pattern.compile("l\\.(\\d+).*"),
      Pattern.compile("([/a-zA-Z0-9\\-\\.]+):(\\d+): (Missing number, treated as zero).*"),
      Pattern.compile("([/a-zA-Z0-9\\-\\.]+):(\\d+): (.*?)\\.?")
   };

   private static final int LOG_PATTERN_JMLRBOOK_CLS_VERSION=0;
   private static final int LOG_PATTERN_JMLRBOOK_CLS=1;
   private static final int LOG_PATTERN_JMLRBOOK_VERSION=2;
   private static final int LOG_PATTERN_PACKAGE_INFO=3;
   private static final int LOG_PATTERN_JMLR_CLS_VERSION=4;
   private static final int LOG_PATTERN_JMLR_CLS=5;
   private static final int LOG_PATTERN_JMLR_VERSION=6;
   private static final int LOG_PATTERN_OLD_STYLE=7;
   private static final int LOG_PATTERN_UNDEFINED_REF=8;
   private static final int LOG_PATTERN_UNDEFINED_CHAR=9;
   private static final int LOG_PATTERN_MISSING_CHAR=10;
   private static final int LOG_PATTERN_MULTIPLE_LABEL=11;
   private static final int LOG_PATTERN_UNDEFINED_CITES=12;
   private static final int LOG_PATTERN_RERUN_TO=13;
   private static final int LOG_PATTERN_LENDDOCA=14;
   private static final int LOG_PATTERN_BOOKMARK=15;
   private static final int LOG_PATTERN_FILE_NOT_FOUND=16;
   private static final int LOG_PATTERN_OBS_STY=17;
   private static final int LOG_PATTERN_CONFLICT_STY=18;
   private static final int LOG_PATTERN_EMERGENCY_STOP=19;
   private static final int LOG_PATTERN_MISSING_NUMBER=20;
   private static final int LOG_PATTERN_STY_BEFORE_CLS=21;
   private static final int LOG_PATTERN_ERROR=22;
   private static final int LOG_PATTERN_LINE_NUMBER=23;
   private static final int LOG_PATTERN_FILE_LINE_MISSING_NUMBER=24;
   private static final int LOG_PATTERN_FILE_LINE=25;


   private static final Pattern[] blgPatterns = new Pattern[]
   {
      Pattern.compile("\\(There w[^ ]+ ([0-9]+) error message.*\\)"),
      Pattern.compile("Warning--I didn't find a database entry for \"(.*?)\".*"),
      Pattern.compile("Warning--empty (\\w+) in (.*)"),
      Pattern.compile("Warning--(.*)"),
      Pattern.compile("I was expecting a `,' or a `}'---line (\\d+) of file (.*?.bib).*"),
      Pattern.compile("(.*)---line (\\d+) of file (.*?.bib).*")
   };

   private static final int BLG_PATTERN_ERROR=0;
   private static final int BLG_PATTERN_UNDEFCITE=1;
   private static final int BLG_PATTERN_MISSING_FIELD=2;
   private static final int BLG_PATTERN_WARNING=3;
   private static final int BLG_PATTERN_MISSING_CLOSE=4;
   private static final int BLG_PATTERN_ERROR_LINE=5;

   private static final Pattern SPAN_AUTHOR_PATTERN = Pattern.compile(
     ".*<span id=\"author\">.+?((?:[a-z]+')?\\p{IsAlphabetic}+?)</span>.*",
     Pattern.DOTALL);

   private static final Pattern NOSPAN_AUTHOR_PATTERN 
    = Pattern.compile(".*?(\\b(?:[a-z]+')?\\p{IsAlphabetic}+?)(\\s+and|,|$).*",
                      Pattern.DOTALL);

   private static final String[] obsoletePackages = new String[]
   {
      "epsfig", "psfig", "glossary"
   };

   private static final String[] problemPackages = new String[]
   {
      "subfig", "theorem", "pdfpages", "geometry", "tabularx", "jmlr2e"
   };

   private BookImportNode importNode;
}

class BookImportNode implements TreeNode
{
   public BookImportNode(BookDocument document)
   {
      this.document = document;
   }

   public void setBook(Book book)
   {
      this.book = book;
   }

   public Enumeration children() {return null; }
   public boolean getAllowsChildren() {return false;}
   public TreeNode getChildAt(int idx) {return null;}
   public int getChildCount() {return 0;}
   public int getIndex(TreeNode node) {return -1;}
   public boolean isLeaf() {return true;}

   public TreeNode getParent()
   {
      return book == null ? null : book.getBookNode();
   }

   public boolean equals(Object obj)
   {
      if (obj == null || !(obj instanceof BookImportNode))
      {
         return false;
      }

      if (this == obj)
      {
         return true;
      }

      return document.equals(((BookImportNode)obj).document);
   }

   public String toString()
   {
      return document.getBaseName();
   }

   public BookDocument getDocument()
   {
      return document;
   }

   private Book book;
   private BookDocument document;
}
