/*
    Copyright (C) 2013 Nicola L.C. Talbot
    www.dickimaw-books.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
package com.dickimawbooks.makejmlrbookgui;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;
import java.util.*;

import javax.swing.*;
import javax.swing.tree.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.text.html.*;

public class JMLRBookPanel extends JPanel
{
   public JMLRBookPanel(MakeJmlrBookGUI application)
   {
      super(new BorderLayout());

      app = application;
      setName(app.getLabel("main.title"));

      tocPanel = new TOCPanel(this);
      treeSP = new JScrollPane(tocPanel);

      infoArea = new JEditorPane();
      infoArea.setContentType("text/html");
      infoArea.setEditable(false);
      infoArea.addHyperlinkListener(app);

      HTMLEditorKit kit = new HTMLEditorKit();
      infoArea.setEditorKit(kit);

      StyleSheet styleSheet = kit.getStyleSheet();
      styleSheet.addRule("th {text-align: left;}");
      styleSheet.addRule("font.error { color: red; }");
      styleSheet.addRule("font.warning { color: navy; }");
      styleSheet.addRule("font.nonconform { color: purple; }");

      Document doc = kit.createDefaultDocument();
      infoArea.setDocument(doc);

      infoSP = new JScrollPane(infoArea);

      splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, treeSP, infoSP);
      splitPane.setResizeWeight(0.1);

      add(splitPane, "Center");

      clear();
   }

   protected JTree getTree()
   {
      return tocPanel.getTree();
   }

   public void clear()
   {
      tocPanel.resetTree();
      validate();
      repaint();
      infoArea.setText("");
   }

   public void bookLoaded()
   {
      tocPanel.resetTree();
      getTree().validate();
      updateInfo();
   }

   public void export(File htmlFile)
    throws IOException
   {
      Book book = app.getBook();

      if (book == null) return;

      if (htmlFile.exists())
      {
         if (JOptionPane.showConfirmDialog(app, 
             app.getLabelWithValue("message.confirm_overwrite", 
              htmlFile.toString()),
             app.getLabel("message.confirm"),
             JOptionPane.YES_NO_OPTION,
             JOptionPane.QUESTION_MESSAGE) != 
             JOptionPane.YES_OPTION)
         {
            return;
         }
      }

      PrintWriter out = null;

      try
      {
         out = new PrintWriter(htmlFile);

         out.println("<html>");
         out.println("<head>");
         out.println("<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\" />");
         out.println("<style type=\"text/css\">");
         out.println("th { text-align: left; }");
         out.println("font.error { color: red; }");
         out.println("font.warning { color: navy; }");
         out.println("font.nonconform { color: purple; }");
         out.println("a[href^=\"info:\"] {visibility: hidden;}");
         out.println("</style>");
         out.println("</head>");

         out.println("<body>");
         out.println(book.getInfoContent());
         out.println("</body>");
      }
      finally
      {
         if (out != null)
         {
            out.close();
         }
      }
   }

   public void updateInfo()
   {
      Book book = app.getBook();

      if (book == null) return;

      infoArea.setText(book.getInfoContent());

      JTree tree = getTree();

      if (tree == null)
      {
         return;
      }

      tree.validate();

      if (tree.getRowCount() > 0)
      {
         tree.expandRow(0);
      }
      //splitPane.validate();
      repaint();
   }

   public Book getBook()
   {
      return app.getBook();
   }

   public void scrollToReference(String reference)
   {
      infoArea.scrollToReference(reference);
   }

   private JEditorPane infoArea;

   private JScrollPane warningSP, infoSP, treeSP;

   private TOCPanel tocPanel;

   private DefaultMutableTreeNode rootNode;

   private JSplitPane splitPane;

   private MakeJmlrBookGUI app;
}

class TOCPanel extends JPanel
   implements MouseListener
{
   public TOCPanel(JMLRBookPanel bookPanel)
   {
      super(new FlowLayout(FlowLayout.LEFT));
      setBackground(Color.white);

      panel = bookPanel;

      resetTree();
   }

   public void resetTree()
   {
      removeAll();

      Book book = panel.getBook();

      if (book == null)
      {
         tree = new JTree(new DefaultMutableTreeNode());
         tree.setRootVisible(false);
      }
      else
      {
         tree = new JTree(book.getBookNode());
         tree.setRootVisible(true);
      }

      add(tree);

      tree.addMouseListener(this);
      tree.setCellRenderer(new DefaultTreeCellRenderer()
      {
         public Component getTreeCellRendererComponent(
             JTree tree, Object value, boolean sel, boolean expanded,
             boolean leaf, int row, boolean hasFocus)
         {
            Component comp = super.getTreeCellRendererComponent(tree,
               value, sel, expanded, leaf, row, hasFocus);

            if (value instanceof BookImportNode)
            {
               BookDocument doc = ((BookImportNode)value).getDocument();

               if (doc != null && doc.getErrorMessages() != null)
               {
                  comp.setForeground(Color.red);
               }
            }

            return comp;
         }
      });
   }

   public JTree getTree()
   {
      return tree;
   }

   public void mousePressed(MouseEvent e)
   {
      TreePath selPath = tree.getPathForLocation(e.getX(), e.getY());

      if (selPath != null)
      {
         TreeNode lastComp = (TreeNode)selPath.getLastPathComponent();

         if (lastComp != null && e.getClickCount() == 1)
         {
            panel.scrollToReference(lastComp.toString());
         }
      }
   }

   public void mouseExited(MouseEvent e)
   {
   }

   public void mouseEntered(MouseEvent e)
   {
   }

   public void mouseReleased(MouseEvent e)
   {
   }

   public void mouseClicked(MouseEvent e)
   {
   }

   private JMLRBookPanel panel;
   private JTree tree;
}
