/*
    Copyright (C) 2013 Nicola L.C. Talbot
    www.dickimaw-books.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
package com.dickimawbooks.makejmlrbookgui;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.MalformedURLException;
import java.nio.file.Path;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Enumeration;
import java.util.regex.*;
import java.text.Normalizer;
import java.nio.charset.Charset;
import java.nio.charset.IllegalCharsetNameException;

import javax.swing.*;
import javax.swing.table.TableModel;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableColumnModel;
import javax.swing.table.TableColumn;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import com.dickimawbooks.texparserlib.*;
import com.dickimawbooks.texparserlib.latex.*;
import com.dickimawbooks.texparserlib.latex.inputenc.InputEncSty;
import com.dickimawbooks.texparserlib.latex2latex.*;

public class NewBookDialog extends JDialog
  implements ActionListener
{
   public NewBookDialog(MakeJmlrBookGUI application)
   {
      super(application, application.getLabel("newbook.title"), true);

      app = application;

      cardLayout = new CardLayout();

      mainPanel = new JPanel(cardLayout);

      introPanel = new IntroPanel();
      mainPanel.add(introPanel);

      titleInfoPanel = new TitleInfoPanel(application.getProperties());
      mainPanel.add(titleInfoPanel);

      frontMatterPanel = new FrontMatterPanel(this);
      mainPanel.add(frontMatterPanel);

      bookPartPanel = new BookPartPanel(this);
      mainPanel.add(bookPartPanel);

      articlesPanel = new ArticlesPanel(bookPartPanel, this);
      mainPanel.add(new JScrollPane(articlesPanel));

      bookDirPanel = new BookDirectoryPanel(this);
      mainPanel.add(bookDirPanel);

      processPanel = new ProcessPanel(this);
      mainPanel.add(processPanel);

      getContentPane().add(mainPanel, BorderLayout.CENTER);

      backButton = createActionButton("back",
         KeyStroke.getKeyStroke(KeyEvent.VK_B, InputEvent.CTRL_MASK));

      nextButton = createActionButton("next",
         KeyStroke.getKeyStroke(KeyEvent.VK_B, InputEvent.CTRL_MASK));

      Box buttonPanel = Box.createHorizontalBox();

      getContentPane().add(buttonPanel, BorderLayout.SOUTH);

      buttonPanel.add(backButton);
      buttonPanel.add(Box.createHorizontalGlue());

      JButton helpButton = JMLRResources.createActionButton(
        "button", "help", null, KeyStroke.getKeyStroke(KeyEvent.VK_F1, 0));

      app.enableHelpOnButton( helpButton, "newbook");

      buttonPanel.add(helpButton);

      buttonPanel.add(Box.createHorizontalGlue());
      buttonPanel.add(nextButton);

      pack();
   }

   private JButton createActionButton(String tag)
   {
      return createActionButton(tag, null);
   }

   private JButton createActionButton(String tag,
     KeyStroke keyStroke)
   {
      return JMLRResources.createActionButton(
        app.getLabel("newbook."+tag),
        app.getMnemonic("newbook."+tag),
        tag, this, keyStroke,
        app.getLabelWithAlt("newbook."+tag+".tooltip", null));
   }

   public void display()
   {
      setLocationRelativeTo(app);
      cardLayout.first(mainPanel);
      currentComponentIndex = 0;
      backButton.setVisible(false);
      nextButton.setVisible(true);

      setVisible(true);
   }

   public NewBookPanel getCurrentPanel()
   {
      NewBookPanel currentPanel = null;

      Component component = mainPanel.getComponent(currentComponentIndex);

      if (component instanceof JScrollPane)
      {
         currentPanel = (NewBookPanel)
                 ((JScrollPane)component).getViewport().getView();
      }
      else
      {
         currentPanel = (NewBookPanel)component;
      }

      return currentPanel;
   }

   public void actionPerformed(ActionEvent evt)
   {
      String action = evt.getActionCommand();

      if (action == null) return;

      if (action.equals("next"))
      {
         NewBookPanel currentPanel = getCurrentPanel();

         if (currentPanel.canMoveForwards())
         {
            cardLayout.next(mainPanel);
            backButton.setVisible(true);
            currentComponentIndex++;

            currentPanel = getCurrentPanel();

            currentPanel.updatePanel();
            nextButton.setVisible(currentPanel.hasNextButton());
            backButton.setVisible(currentPanel.hasBackButton());
         }
      }
      else if (action.equals("back"))
      {
         NewBookPanel currentPanel = getCurrentPanel();

         if (currentComponentIndex > 0 && currentPanel.hasBackButton())
         {
            cardLayout.previous(mainPanel);
            currentComponentIndex--;

            currentPanel = getCurrentPanel();

            backButton.setVisible(currentPanel.hasBackButton());
            nextButton.setVisible(currentPanel.hasNextButton());
         }
      }
   }

   public boolean isCiML()
   {
      return titleInfoPanel.isCiML();
   }

   public int getVolume()
   {
      return isCiML() ? titleInfoPanel.getCiMLVolume() :
         titleInfoPanel.getJMLRVolume();
   }

   public int getYear()
   {
      return isCiML() ? 0 : titleInfoPanel.getJMLRYear();
   }

   public String getMonth()
   {
      return isCiML() ? null : titleInfoPanel.getJMLRMonth();
   }

   public String getDay()
   {
      return isCiML() ? null : titleInfoPanel.getJMLRDay();
   }

   public String getJMLRLocation()
   {
      return isCiML() ? null : titleInfoPanel.getJMLRLocation();
   }

   public String getLogoImage()
   {
      return isCiML() ? null : titleInfoPanel.getJMLRLogoImage();
   }

   public String getLogoUrl()
   {
      return isCiML() ? null : titleInfoPanel.getJMLRLogoUrl();
   }

   public Vector<String> getEditors()
   {
      return isCiML() ? titleInfoPanel.getCiMLEditors() :
         titleInfoPanel.getJMLREditors();
   }

   public String getBookTitle()
   {
      return isCiML() ? titleInfoPanel.getCiMLTitle() :
         titleInfoPanel.getJMLRTitle();
   }

   public String getBookShortTitle()
   {
      return isCiML() ? titleInfoPanel.getCiMLShortTitle() :
         titleInfoPanel.getJMLRShortTitle();
   }

   public String getBookSubTitle()
   {
      return isCiML() ? titleInfoPanel.getCiMLSubTitle() : "";
   }

   public String getProductionEditor()
   {
      return isCiML() ? titleInfoPanel.getCiMLProductionEditor() : "";
   }

   public File getBookDirectory()
   {
      return bookDirPanel.getBookDirectory();
   }

   public String getBookFonts()
   {
      return bookDirPanel.getBookFonts();
   }

   public Vector<PartListPanel> getPartLists()
   {
      return articlesPanel.getPartLists();
   }

   public ImportPapersPanel getMainImportPapers()
   {
      return articlesPanel.getMainImportPapers();
   }

   public ImportPapersPanel getAppendixImportPapers()
   {
      return articlesPanel.getAppendixImportPapers();
   }

   public boolean hasPreface()
   {
      return frontMatterPanel.hasPreface();
   }

   public boolean hasForeword()
   {
      return frontMatterPanel.hasForeword();
   }

   public boolean hasAppendices()
   {
      return bookPartPanel.hasAppendices();
   }

   public int getPartConfig()
   {
      return bookPartPanel.getPartConfig();
   }

   public File getPrefaceFile()
   {
      return frontMatterPanel.getPrefaceFile();
   }

   public File getForewordFile()
   {
      return frontMatterPanel.getForewordFile();
   }

   public JFileChooser getFileChooser()
   {
      return app.getFileChooser();
   }

   public JFileChooser getDirChooser()
   {
      return app.getDirChooser();
   }

   public ImageFileFilter getImageFileFilter()
   {
      return app.getImageFileFilter();
   }

   public File getPackageLocation(String name)
   {
      try
      {
         return new File(app.kpsewhich(name+".sty"));
      }
      catch (IOException e)
      {
      }
      catch (InterruptedException e)
      {
      }

      return null;
   }

   public void verboseCopyFile(File srcFile, File destFile)
    throws IOException
   {
      app.verboseCopyFile(srcFile, destFile);
   }

   public String kpsewhich(String name)
     throws IOException
   {
      try
      {
         return app.kpsewhich(name);
      }
      catch (InterruptedException e)
      {
         throw new IOException(MakeJmlrBookGUI.getLabel("error.interrupted"),
           e);
      }
   }

   public void epstopdf(File epsFile, File pdfFile)
     throws IOException
   {
      try
      {
         app.epstopdf(epsFile, pdfFile);
      }
      catch (InterruptedException e)
      {
         throw new IOException(MakeJmlrBookGUI.getLabel("error.interrupted"),
           e);
      }
   }

   public void wmftoeps(File wmfFile, File epsFile)
     throws IOException
   {
      try
      {
         app.wmftoeps(wmfFile, epsFile);
      }
      catch (InterruptedException e)
      {
         throw new IOException(MakeJmlrBookGUI.getLabel("error.interrupted"),
           e);
      }
   }

   public void viewFile(File file)
     throws MalformedURLException
   {
      app.viewFile(file);
   }

   public void setAppInfo(String msg)
   {
      app.setInfo(msg);
   }

   public void finish(File file)
   {
      setVisible(false);

      app.setInfo(MakeJmlrBookGUI.getLabel("message.done"));

      if (file != null && file.exists())
      {
         try
         {
            app.load(file);
         }
         catch (IOException e)
         {
            app.error(e);
         }
      }
   }

   private MakeJmlrBookGUI app;

   private CardLayout cardLayout;

   private JComponent mainPanel;

   private IntroPanel introPanel;
   private TitleInfoPanel titleInfoPanel;
   private FrontMatterPanel frontMatterPanel;
   private BookPartPanel bookPartPanel;
   private ArticlesPanel articlesPanel;
   private BookDirectoryPanel bookDirPanel;
   private ProcessPanel processPanel;

   private int currentComponentIndex;

   private JButton nextButton, backButton;
}

interface NewBookPanel
{
   public boolean hasNextButton();
   public boolean hasBackButton();
   public boolean canMoveForwards();
   public void updatePanel();
}

class IntroPanel extends JTextArea
   implements NewBookPanel
{
   public IntroPanel()
   {
      super(10, 50);

      setText(MakeJmlrBookGUI.getLabel("newbook.intro"));
      setLineWrap(true);
      setWrapStyleWord(true);
      setEditable(false);
      setOpaque(false);
   }

   public void updatePanel() {}
   public boolean canMoveForwards() {return true;}
   public boolean hasNextButton() {return true;}
   public boolean hasBackButton() {return false;}
}

interface TitleField
{
   public JComponent getComponent();
   public JComponent getLabel();
   public int getNumValues();
   public boolean isEmpty();
}

class TitleTextField extends JPanel implements TitleField
{
   public TitleTextField(String parentTag, String tag)
   {
      super(new FlowLayout(FlowLayout.LEADING, 5, 0));
      setAlignmentX(LEFT_ALIGNMENT);

      textField = new JTextField(40);
      add(textField);

      String label = parentTag+"."+tag;

      jLabel = new JLabel(MakeJmlrBookGUI.getLabel(label));
      jLabel.setDisplayedMnemonic(MakeJmlrBookGUI.getMnemonic(label));
      jLabel.setLabelFor(textField);

      String tooltip = MakeJmlrBookGUI.getLabelWithAlt(label+".tooltip", null);

      if (tooltip != null)
      {
         textField.setToolTipText(tooltip);
         jLabel.setToolTipText(tooltip);
      }

      Dimension dim = getMaximumSize();
      dim.height = (int)getPreferredSize().getHeight();
      setMaximumSize(dim);
   }

   public JComponent getComponent()
   {
      return this;
   }

   public JComponent getLabel()
   {
      return jLabel;
   }

   public int getNumValues()
   {
      return 1;
   }

   public boolean isEmpty()
   {
      return getText().isEmpty();
   }

   public String getText()
   {
      return textField.getText();
   }

   private JLabel jLabel;
   private JTextField textField;
}

class TitleIntField extends JPanel implements TitleField
{
   public TitleIntField(String parentTag, String tag)
   {
      super(new FlowLayout(FlowLayout.LEADING, 5, 0));
      setAlignmentX(LEFT_ALIGNMENT);

      field = new NonNegativeIntField();
      add(field);

      String label = parentTag+"."+tag;

      jLabel = new JLabel(MakeJmlrBookGUI.getLabel(label));
      jLabel.setDisplayedMnemonic(MakeJmlrBookGUI.getMnemonic(label));
      jLabel.setLabelFor(field);

      String tooltip = MakeJmlrBookGUI.getLabelWithAlt(label+".tooltip", null);

      if (tooltip != null)
      {
         field.setToolTipText(tooltip);
         jLabel.setToolTipText(tooltip);
      }

      Dimension dim = getMaximumSize();
      dim.height = (int)getPreferredSize().getHeight();
      setMaximumSize(dim);
   }

   public JComponent getComponent()
   {
      return this;
   }

   public JComponent getLabel()
   {
      return jLabel;
   }

   public String getText()
   {
      return field.getText();
   }

   public int getNumValues()
   {
      return 1;
   }

   public boolean isEmpty()
   {
      return field.getText().isEmpty();
   }

   public int getIntValue()
   {
      return field.getValue();
   }

   private NonNegativeIntField field;
   private JLabel jLabel;
}

class TitleListField extends JPanel 
  implements TitleField,ActionListener
{
   public TitleListField(String parentTag, String tag,
      final String columnLabel)
   {
      super(new FlowLayout(FlowLayout.LEADING, 0, 0));
      setAlignmentX(LEFT_ALIGNMENT);
      this.columnLabel = columnLabel;

      data = new Vector<String>();
      data.add(new String());

      TableModel model = new AbstractTableModel()
      {
         public int getColumnCount() {return data.size();}
         public int getRowCount() {return 1;}
         public boolean isCellEditable(int row, int col) {return true;}

         public Object getValueAt(int row, int col)
         {
            return data.get(col);
         }

         public void setValueAt(Object value, int row, int col)
         {
            data.set(col, (String)value);
         }

         public String getColumnName(int column)
         {
            return MakeJmlrBookGUI.getLabelWithValue(columnLabel, column+1);
         }
      };

      table = new JTable(model)
      {
         public void columnSelectionChanged(ListSelectionEvent e)
         {
            super.columnSelectionChanged(e);
            updateRemoveButton();
         }

         public String getToolTipText(MouseEvent e)
         {
            return MakeJmlrBookGUI.getLabel(columnLabel+".tooltip");
         }
      };

      table.setAlignmentX(LEFT_ALIGNMENT);
      table.setColumnSelectionAllowed(true);
      table.setRowSelectionAllowed(false);
      table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
      table.putClientProperty("terminateEditOnFocusLost", Boolean.TRUE);

      ListSelectionModel selectionModel = table.getSelectionModel();
      selectionModel.setSelectionMode(
         ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);

      JScrollPane sp = new JScrollPane(table);
      sp.setBorder(BorderFactory.createEmptyBorder());
      sp.setAlignmentX(LEFT_ALIGNMENT);

      Dimension dim = sp.getPreferredSize();
      dim.height = (int)table.getPreferredSize().getHeight() 
        + 3*table.getRowHeight();
      sp.setPreferredSize(dim);

      Dimension maxDim = sp.getMaximumSize();
      maxDim.height = dim.height;
      sp.setMaximumSize(maxDim);

      TableColumn column = table.getColumnModel().getColumn(0);
      column.setPreferredWidth(Math.max(column.getPreferredWidth(),
        dim.width/3));

      table.setRowHeight(table.getRowHeight()+4);

      add(sp);

      JPanel buttonPanel = new JPanel(new FlowLayout(FlowLayout.LEADING));
      buttonPanel.setAlignmentX(LEFT_ALIGNMENT);

      JButton addButton = JMLRResources.createAddButton(this);
      buttonPanel.add(addButton);

      removeButton = JMLRResources.createRemoveButton(this);
      removeButton.setEnabled(false);
      buttonPanel.add(removeButton);

      String label = parentTag+"."+tag;

      labelComponent = Box.createVerticalBox();
      labelComponent.setAlignmentX(LEFT_ALIGNMENT);

      JLabel jLabel = new JLabel(MakeJmlrBookGUI.getLabel(label));
      jLabel.setAlignmentX(LEFT_ALIGNMENT);
      jLabel.setDisplayedMnemonic(MakeJmlrBookGUI.getMnemonic(label));

      labelComponent.add(jLabel);
      labelComponent.add(buttonPanel);
   }

   public void updateRemoveButton()
   {
      int colCount = table.getColumnCount();
      int selColCount = table.getSelectedColumnCount();

      removeButton.setEnabled(
        colCount > 1 && selColCount > 0 && selColCount < colCount);
   }

   public void actionPerformed(ActionEvent evt)
   {
      String action = evt.getActionCommand();

      if (action == null) return;

      if (action.equals("add"))
      {
         if (table.isEditing())
         {
            table.getCellEditor().stopCellEditing();
         }

         int n = data.size();

         TableColumnModel colModel = table.getColumnModel();
         int[] colWidths = new int[n+1];

         for (int i = 0; i < n; i++)
         {
            TableColumn column = colModel.getColumn(i);
            colWidths[i] = column.getPreferredWidth();
         }

         colWidths[n] = colWidths[n-1];

         data.add(new String(""));
         ((AbstractTableModel)table.getModel()).fireTableStructureChanged();

         for (int i = 0; i <= n; i++)
         {
            TableColumn column = colModel.getColumn(i);
            column.setPreferredWidth(colWidths[i]);
         }

         updateRemoveButton();
      }
      else if (action.equals("remove"))
      {
         if (table.getSelectedColumnCount() > 0)
         {
            TableColumnModel colModel = table.getColumnModel();
            Vector<Integer> colWidths = new Vector<Integer>(data.size());

            for (int i = 0, n = data.size(); i < n; i++)
            {
               TableColumn column = colModel.getColumn(i);
               colWidths.add(column.getPreferredWidth());
            }

            int[] selection = table.getSelectedColumns();

            for (int i = selection.length-1; i >= 0; i--)
            {
               data.remove(selection[i]);
               colWidths.remove(selection[i]);
            }

            ((AbstractTableModel)table.getModel()).fireTableStructureChanged();

            for (int i = 0, n = colWidths.size(); i < n; i++)
            {
               TableColumn column = colModel.getColumn(i);
               column.setPreferredWidth(colWidths.get(i));
            }

            updateRemoveButton();
         }
      }
   }

   public JComponent getComponent()
   {
      return this;
   }

   public JComponent getLabel()
   {
      return labelComponent;
   }

   public String getText()
   {
      return table.getColumnCount() == 0 ? "" : 
        table.getModel().getValueAt(0,0).toString();
   }

   public Vector<String> getEditors()
   {
      return data;
   }

   public int getNumValues()
   {
      return table.getColumnCount();
   }

   public boolean isEmpty()
   {
      TableModel model = table.getModel();
      int n = model.getColumnCount();

      for (int i = 0; i < n; i++)
      {
         Object value = model.getValueAt(0, i);

         if (value != null && !value.toString().isEmpty())
         {
            return false;
         }
      }

      return true;
   }

   private JTable table;

   private JButton removeButton;

   private Vector<String> data;

   private String columnLabel;
   private JComponent labelComponent;
}

class TitleMenuField extends JPanel implements TitleField
{
   public TitleMenuField(String parentTag, String tag,
      final String itemList)
   {
      super(new FlowLayout(FlowLayout.LEADING, 5, 0));
      setAlignmentX(LEFT_ALIGNMENT);

      menuField = new JComboBox<String>(
        MakeJmlrBookGUI.getLabel(itemList).split(","));
      add(menuField);

      String label = parentTag+"."+tag;

      jLabel = new JLabel(MakeJmlrBookGUI.getLabel(label));
      jLabel.setDisplayedMnemonic(MakeJmlrBookGUI.getMnemonic(label));
      jLabel.setLabelFor(menuField);

      String tooltip = MakeJmlrBookGUI.getLabelWithAlt(label+".tooltip", null);

      if (tooltip != null)
      {
         menuField.setToolTipText(tooltip);
         jLabel.setToolTipText(tooltip);
      }

      Dimension dim = getMaximumSize();
      dim.height = (int)getPreferredSize().getHeight();
      setMaximumSize(dim);
   }

   public JComponent getComponent()
   {
      return this;
   }

   public JComponent getLabel()
   {
      return jLabel;
   }

   public int getNumValues()
   {
      return 1;
   }

   public boolean isEmpty()
   {
      return getText().isEmpty();
   }

   public String getText()
   {
      return menuField.getSelectedItem().toString();
   }

   private JLabel jLabel;
   private JComboBox<String> menuField;
}

class TitleFileField extends FileField implements TitleField
{
   public TitleFileField(String parentTag, String tag,
      final String filterType, File defaultDir)
   {
      super(null, null);
      setAlignmentX(LEFT_ALIGNMENT);
      setBorder(BorderFactory.createEmptyBorder(0, 5, 0, 5));

      setFileChooser(createFileChooser(filterType, defaultDir));

      String label = parentTag+"."+tag;

      jLabel = new JLabel(MakeJmlrBookGUI.getLabel(label));
      jLabel.setDisplayedMnemonic(MakeJmlrBookGUI.getMnemonic(label));
      jLabel.setLabelFor(getTextField());

      String tooltip = MakeJmlrBookGUI.getLabelWithAlt(label+".tooltip", null);

      if (tooltip != null)
      {
         getTextField().setToolTipText(tooltip);
         jLabel.setToolTipText(tooltip);
      }

      Dimension dim = getMaximumSize();
      dim.height = (int)getPreferredSize().getHeight();
      setMaximumSize(dim);
   }

   private JFileChooser createFileChooser(String filterType, File defaultDir)
   {
      JFileChooser chooser = new JFileChooser();

      if (filterType.equals("image"))
      {
         ImageFileFilter filter = new ImageFileFilter();
         chooser.addChoosableFileFilter(filter);
         chooser.setFileFilter(filter);
      }
      else if (filterType.equals("tex"))
      {
         TeXFileFilter filter = new TeXFileFilter();
         chooser.addChoosableFileFilter(filter);
         chooser.setFileFilter(filter);
      }

      chooser.setCurrentDirectory(defaultDir);

      return chooser;
   }

   public JComponent getComponent()
   {
      return this;
   }

   public JComponent getLabel()
   {
      return jLabel;
   }

   public int getNumValues()
   {
      return 1;
   }

   public boolean isEmpty()
   {
      return getFileName().isEmpty();
   }

   public String getText()
   {
      return getFileName();
   }

   private JLabel jLabel;
}

class TitleInfoPanel extends JPanel
  implements ActionListener,NewBookPanel
{
   public TitleInfoPanel(JMLRBookProperties properties)
   {
      super(new BorderLayout());

      File defaultDir = new File(properties.getDefaultDirectory());

      JPanel bookTypePanel = new JPanel();

      add(bookTypePanel, BorderLayout.NORTH);

      bookTypePanel.add(new JLabel(MakeJmlrBookGUI.getLabel("newbook.type")));

      ButtonGroup bg = new ButtonGroup();

      jmlrButton = JMLRResources.createRadioButton("newbook", "jmlrwcp", 
        this, true);
      bg.add(jmlrButton);
      bookTypePanel.add(jmlrButton);

      cimlButton = JMLRResources.createRadioButton("newbook", "ciml", 
        this, false);
      bg.add(cimlButton);
      bookTypePanel.add(cimlButton);

      cardLayout = new CardLayout();

      cardPanel = new JPanel(cardLayout);
      add(cardPanel, BorderLayout.CENTER);

      Box jmlrComp = Box.createVerticalBox();
      cardLayout.addLayoutComponent(jmlrComp, "jmlrwcp");
      cardPanel.add(jmlrComp);

      JComponent[] labels = new JComponent[jmlrData.length];
      jmlrFields = new JComponent[jmlrData.length];

      int maxWidth = 0;

      for (int i = 0; i < jmlrData.length; i++)
      {
         Box box = Box.createHorizontalBox();
         jmlrComp.add(box);

         switch (jmlrData[i].getType())
         {
            case INT:
              jmlrFields[i] = new TitleIntField("newbook.jmlrwcp",
                 jmlrData[i].getTag());
            break;
            case TEXT:
              jmlrFields[i] = new TitleTextField("newbook.jmlrwcp",
                 jmlrData[i].getTag());
            break;
            case LIST:
              jmlrFields[i] = new TitleListField("newbook.jmlrwcp",
                 jmlrData[i].getTag(), jmlrData[i].getArg());
            break;
            case MENU:
              jmlrFields[i] = new TitleMenuField("newbook.jmlrwcp",
                 jmlrData[i].getTag(), jmlrData[i].getArg());
            break;
            case FILE:
              jmlrFields[i] = new TitleFileField("newbook.jmlrwcp",
                 jmlrData[i].getTag(), jmlrData[i].getArg(), defaultDir);
            break;
         }

         labels[i] = ((TitleField)jmlrFields[i]).getLabel();

         box.add(labels[i]);

         box.add(jmlrFields[i]);

         if (!jmlrData[i].isRequired())
         {
            box.add(new JLabel(MakeJmlrBookGUI.getLabel("newbook.optional")));
         }

         Dimension dim = labels[i].getPreferredSize();
         maxWidth = Math.max(maxWidth, dim.width);

         jmlrComp.add(Box.createVerticalStrut(2));
      }

      jmlrComp.add(Box.createVerticalGlue());

      for (int i = 0; i < labels.length; i++)
      {
         Dimension dim = labels[i].getPreferredSize();
         dim.width = maxWidth;
         labels[i].setPreferredSize(dim);
      }

      Box cimlComp = Box.createVerticalBox();
      cardLayout.addLayoutComponent(cimlComp, "ciml");
      cardPanel.add(cimlComp);

      labels = new JComponent[cimlData.length];
      cimlFields = new JComponent[cimlData.length];

      maxWidth = 0;

      for (int i = 0; i < cimlData.length; i++)
      {
         Box box = Box.createHorizontalBox();
         cimlComp.add(box);

         switch (cimlData[i].getType())
         {
            case INT:
               cimlFields[i] = new TitleIntField("newbook.ciml",
                  cimlData[i].getTag());
            break;
            case TEXT:
               cimlFields[i] = new TitleTextField("newbook.ciml",
                  cimlData[i].getTag());
            break;
            case LIST:
               cimlFields[i] = new TitleListField("newbook.ciml",
                  cimlData[i].getTag(), cimlData[i].getArg());
            break;
            case MENU:
               cimlFields[i] = new TitleMenuField("newbook.ciml",
                  cimlData[i].getTag(), cimlData[i].getArg());
            break;
            case FILE:
               cimlFields[i] = new TitleFileField("newbook.ciml",
                  cimlData[i].getTag(), cimlData[i].getArg(), defaultDir);
            break;
         }

         labels[i] = ((TitleField)cimlFields[i]).getLabel();

         box.add(labels[i]);

         box.add(cimlFields[i]);

         if (!cimlData[i].isRequired())
         {
            box.add(new JLabel(MakeJmlrBookGUI.getLabel("newbook.optional")));
         }

         Dimension dim = labels[i].getPreferredSize();
         maxWidth = Math.max(maxWidth, dim.width);

         cimlComp.add(Box.createVerticalStrut(2));
      }

      cimlComp.add(Box.createVerticalGlue());

      for (int i = 0; i < labels.length; i++)
      {
         Dimension dim = labels[i].getPreferredSize();
         dim.width = maxWidth;
         labels[i].setPreferredSize(dim);
      }
   }

   public void actionPerformed(ActionEvent evt)
   {
      String action = evt.getActionCommand();

      if (action == null) return;

      if (action.equals("jmlrwcp"))
      {
         cardLayout.show(cardPanel, "jmlrwcp");
      }
      else if (action.equals("ciml"))
      {
         cardLayout.show(cardPanel, "ciml");
      }
   }

   public boolean canMoveForwards()
   {
      String message = "";

      if (jmlrButton.isSelected())
      {
         for (int i = 0; i < jmlrData.length; i++)
         {
            TitleField field = (TitleField)jmlrFields[i];

            if (jmlrData[i].isRequired() && field.isEmpty())
            {
               String fieldLabel = MakeJmlrBookGUI.getLabelWithAlt(
                    "newbook.jmlrwcp."+jmlrData[i].getTag(),
                    jmlrData[i].getTag());

               int idx = fieldLabel.lastIndexOf(":");

               if (idx > -1)
               {
                  fieldLabel = fieldLabel.substring(0, idx);
               }

               message += MakeJmlrBookGUI.getLabelWithValue(
                 "error.missing_field", fieldLabel)+"\n";
            }
         }
      }
      else
      {
         for (int i = 0; i < cimlData.length; i++)
         {
            TitleField field = (TitleField)cimlFields[i];

            if (cimlData[i].isRequired() && field.isEmpty())
            {
               String fieldLabel = MakeJmlrBookGUI.getLabelWithAlt(
                   "newbook.ciml."+cimlData[i].getTag(),
                   cimlData[i].getTag());

               int idx = fieldLabel.lastIndexOf(":");

               if (idx > -1)
               {
                  fieldLabel = fieldLabel.substring(0, idx);
               }

               message += MakeJmlrBookGUI.getLabelWithValue(
                 "error.missing_field", fieldLabel)+"\n";
            }
         }
      }

      if (message.isEmpty())
      {
         return true;
      }

      JMLRResources.error(this, message);

      return false;
   }

   public boolean isCiML()
   {
      return cimlButton.isSelected();
   }

   public int getCiMLVolume()
   {
      return ((TitleIntField)cimlFields[CIML_VOLUME]).getIntValue();
   }

   public int getJMLRVolume()
   {
      return ((TitleIntField)jmlrFields[JMLR_VOLUME]).getIntValue();
   }

   public int getJMLRYear()
   {
      return ((TitleIntField)jmlrFields[JMLR_YEAR]).getIntValue();
   }

   public String getJMLRMonth()
   {
      return ((TitleMenuField)jmlrFields[JMLR_MONTH]).getText();
   }

   public String getJMLRDay()
   {
      return ((TitleTextField)jmlrFields[JMLR_DAY]).getText();
   }

   public String getJMLRLocation()
   {
      return ((TitleTextField)jmlrFields[JMLR_LOCATION]).getText();
   }

   public String getJMLRLogoImage()
   {
      return ((TitleFileField)jmlrFields[JMLR_LOGO_IMAGE]).getText();
   }

   public String getJMLRLogoUrl()
   {
      return ((TitleTextField)jmlrFields[JMLR_LOGO_URL]).getText();
   }

   public Vector<String> getCiMLEditors()
   {
      return ((TitleListField)cimlFields[CIML_EDITOR]).getEditors();
   }

   public Vector<String> getJMLREditors()
   {
      return ((TitleListField)jmlrFields[JMLR_EDITOR]).getEditors();
   }

   public String getJMLRTitle()
   {
      return ((TitleTextField)jmlrFields[JMLR_TITLE]).getText();
   }

   public String getJMLRShortTitle()
   {
      return ((TitleTextField)jmlrFields[JMLR_SHORT_TITLE]).getText();
   }

   public String getCiMLTitle()
   {
      return ((TitleTextField)cimlFields[CIML_TITLE]).getText();
   }

   public String getCiMLShortTitle()
   {
      return ((TitleTextField)cimlFields[CIML_SHORT_TITLE]).getText();
   }

   public String getCiMLSubTitle()
   {
      return ((TitleTextField)cimlFields[CIML_SUBTITLE]).getText();
   }

   public String getCiMLProductionEditor()
   {
      return ((TitleTextField)cimlFields[CIML_PRODEDITOR]).getText();
   }

   public void updatePanel() {}

   public boolean hasNextButton() {return true;}

   public boolean hasBackButton() {return true;}

   private CardLayout cardLayout;

   private JComponent cardPanel;

   private JRadioButton jmlrButton, cimlButton;

   private JComponent[] jmlrFields, cimlFields;

   private static final int TEXT=0, INT=1, LIST=2, MENU=3, FILE=4;

   private static final int JMLR_VOLUME=0;
   private static final int JMLR_TITLE=1;
   private static final int JMLR_SHORT_TITLE=2;
   private static final int JMLR_YEAR=3;
   private static final int JMLR_MONTH=4;
   private static final int JMLR_DAY=5;
   private static final int JMLR_LOCATION=6;
   private static final int JMLR_LOGO_IMAGE=7;
   private static final int JMLR_LOGO_URL=8;
   private static final int JMLR_EDITOR=9;

   private static final TitleInfoData[] jmlrData =
   {
      new TitleInfoData("volume", null, true, INT),
      new TitleInfoData("title", null, true, TEXT),
      new TitleInfoData("shorttitle", null, true, TEXT),
      new TitleInfoData("year", null, true, INT),
      new TitleInfoData("month", "newbook.months", true, MENU),
      new TitleInfoData("day", null, false, TEXT),
      new TitleInfoData("location", null, false, TEXT),
      new TitleInfoData("logoimage", "image", false, FILE),
      new TitleInfoData("logourl", null, false, TEXT),
      new TitleInfoData("editor", "newbook.editor", true, LIST)
   };

   private static final int CIML_VOLUME=0;
   private static final int CIML_TITLE=1;
   private static final int CIML_SHORT_TITLE=2;
   private static final int CIML_SUBTITLE=3;
   private static final int CIML_PRODEDITOR=4;
   private static final int CIML_EDITOR=5;

   private static final TitleInfoData[] cimlData =
   {
      new TitleInfoData("volume", null, true, INT),
      new TitleInfoData("title", null, true, TEXT),
      new TitleInfoData("shorttitle", null, false, TEXT),
      new TitleInfoData("subtitle", null, false, TEXT),
      new TitleInfoData("prodeditor", null, false, TEXT),
      new TitleInfoData("editor", "newbook.editor", true, LIST)
   };
}

class TitleInfoData
{
   public TitleInfoData(String tag, String arg, boolean required, int type)
   {
      this.tag = tag;
      this.arg = arg;
      this.required = required;
      this.type = type;
   }

   public String getTag()
   {
      return tag;
   }

   public String getArg()
   {
      return arg;
   }

   public boolean isRequired()
   {
      return required;
   }

   public int getType()
   {
      return type;
   }

   private String tag, arg;
   private boolean required;
   private int type;
}

class FrontMatterPanel extends JPanel
  implements ActionListener,NewBookPanel
{
   public FrontMatterPanel(NewBookDialog dialog)
   {
      super(null);

      this.dialog = dialog;

      setLayout(new BoxLayout(this, BoxLayout.PAGE_AXIS));
      setAlignmentX(LEFT_ALIGNMENT);

      JLabel label 
        = new JLabel(MakeJmlrBookGUI.getLabel("newbook.frontmatter.preface"));
      label.setAlignmentX(LEFT_ALIGNMENT);

      add(label);

      ButtonGroup bg = new ButtonGroup();

      prefaceNoButton = 
         JMLRResources.createRadioButton("newbook.frontmatter", "prefaceno",
         this, false);
      prefaceNoButton.setAlignmentX(LEFT_ALIGNMENT);
      bg.add(prefaceNoButton);
      add(prefaceNoButton);

      prefaceYesButton = 
         JMLRResources.createRadioButton("newbook.frontmatter", "prefaceyes",
         this, true);
      prefaceYesButton.setAlignmentX(LEFT_ALIGNMENT);
      bg.add(prefaceYesButton);
      add(prefaceYesButton);

      Box box = Box.createHorizontalBox();
      box.setAlignmentX(LEFT_ALIGNMENT);
      add(box);

      box.add(Box.createHorizontalStrut(INDENT));

      bg = new ButtonGroup();

      prefaceNewButton = JMLRResources.createRadioButton(
         "newbook.frontmatter", "prefacenew", this, true);
      bg.add(prefaceNewButton);
      box.add(prefaceNewButton);
      prefaceNewButton.setAlignmentX(LEFT_ALIGNMENT);

      box = Box.createHorizontalBox();
      box.setAlignmentX(LEFT_ALIGNMENT);
      add(box);

      box.add(Box.createHorizontalStrut(INDENT));

      prefaceTeXButton = JMLRResources.createRadioButton(
        "newbook.frontmatter", "prefacetex", this, false);
      bg.add(prefaceTeXButton);
      box.add(prefaceTeXButton);
      prefaceTeXButton.setAlignmentX(LEFT_ALIGNMENT);

      prefaceFileField = new FileField(this, dialog.getFileChooser());
      box.add(prefaceFileField);

      add(Box.createVerticalGlue());

      forewordPanel = new JPanel(null);
      forewordPanel.setLayout(new BoxLayout(forewordPanel, 
         BoxLayout.PAGE_AXIS));
      forewordPanel.setAlignmentX(LEFT_ALIGNMENT);

      add(forewordPanel);

      label = 
        new JLabel(MakeJmlrBookGUI.getLabel("newbook.frontmatter.foreword"));
      label.setAlignmentX(LEFT_ALIGNMENT);

      forewordPanel.add(label);

      bg = new ButtonGroup();

      forewordNoButton = 
         JMLRResources.createRadioButton("newbook.frontmatter", "forewordno",
         this, true);
      forewordNoButton.setAlignmentX(LEFT_ALIGNMENT);
      bg.add(forewordNoButton);
      forewordPanel.add(forewordNoButton);

      forewordYesButton = 
         JMLRResources.createRadioButton("newbook.frontmatter", "forewordyes",
         this, false);
      forewordYesButton.setAlignmentX(LEFT_ALIGNMENT);
      bg.add(forewordYesButton);
      forewordPanel.add(forewordYesButton);

      box = Box.createHorizontalBox();
      box.setAlignmentX(LEFT_ALIGNMENT);
      forewordPanel.add(box);

      box.add(Box.createHorizontalStrut(INDENT));

      bg = new ButtonGroup();

      forewordNewButton = JMLRResources.createRadioButton(
         "newbook.frontmatter", "forewordnew", this, true);
      bg.add(forewordNewButton);
      box.add(forewordNewButton);
      forewordNewButton.setAlignmentX(LEFT_ALIGNMENT);
      forewordNewButton.setEnabled(false);

      box = Box.createHorizontalBox();
      box.setAlignmentX(LEFT_ALIGNMENT);
      forewordPanel.add(box);

      box.add(Box.createHorizontalStrut(INDENT));

      forewordTeXButton = JMLRResources.createRadioButton(
        "newbook.frontmatter", "forewordtex", this, false);
      bg.add(forewordTeXButton);
      box.add(forewordTeXButton);
      forewordTeXButton.setAlignmentX(LEFT_ALIGNMENT);
      forewordTeXButton.setEnabled(false);

      forewordFileField = new FileField(this, dialog.getFileChooser());
      forewordFileField.setEnabled(false);
      box.add(forewordFileField);

      add(Box.createVerticalGlue());
      add(Box.createVerticalStrut(100));
   }

   public void actionPerformed(ActionEvent evt)
   {
      String action = evt.getActionCommand();

      if (action == null) return;

      if (action.equals("prefaceno"))
      {
         prefaceNewButton.setEnabled(false);
         prefaceTeXButton.setEnabled(false);
         prefaceFileField.setEnabled(false);
      }
      else if (action.equals("prefaceyes"))
      {
         prefaceNewButton.setEnabled(true);
         prefaceTeXButton.setEnabled(true);
         prefaceFileField.setEnabled(prefaceTeXButton.isSelected());
      }
      else if (action.equals("prefacenew"))
      {
         prefaceFileField.setEnabled(prefaceTeXButton.isSelected());
      }
      else if (action.equals("prefacetex"))
      {
         prefaceFileField.setEnabled(prefaceTeXButton.isSelected());
      }
      else if (action.equals("forewordno"))
      {
         forewordNewButton.setEnabled(false);
         forewordTeXButton.setEnabled(false);
         forewordFileField.setEnabled(false);
      }
      else if (action.equals("forewordyes"))
      {
         forewordNewButton.setEnabled(true);
         forewordTeXButton.setEnabled(true);
         forewordFileField.setEnabled(forewordTeXButton.isSelected());
      }
      else if (action.equals("forewordnew"))
      {
         forewordFileField.setEnabled(forewordTeXButton.isSelected());
      }
      else if (action.equals("forewordtex"))
      {
         forewordFileField.setEnabled(forewordTeXButton.isSelected());
      }
   }

   public void updatePanel()
   {
      forewordPanel.setVisible(dialog.isCiML());
   }

   public boolean canMoveForwards()
   {
      if (hasPreface() && prefaceTeXButton.isSelected())
      {
          File file = prefaceFileField.getFile();

          if (file == null)
          {
             JMLRResources.error(this, 
                MakeJmlrBookGUI.getLabel("newbook.frontmatter.missing_pref"));

             return false;
          }
          else if (!file.exists())
          {
             JMLRResources.error(this, 
                MakeJmlrBookGUI.getLabelWithValue("error.file_not_found",
                file.getAbsolutePath()));

             return false;
          }
      }

      if (hasForeword() && forewordTeXButton.isSelected())
      {
          File file = forewordFileField.getFile();

          if (file == null)
          {
             JMLRResources.error(this, 
                MakeJmlrBookGUI.getLabel("newbook.frontmatter.missing_fore"));

             return false;
          }
          else if (!file.exists())
          {
             JMLRResources.error(this, 
                MakeJmlrBookGUI.getLabelWithValue("error.file_not_found",
                file.getAbsolutePath()));

             return false;
          }
      }

      return true;
   }

   public boolean hasNextButton() {return true;}

   public boolean hasBackButton() {return true;}

   public boolean hasPreface()
   {
      return prefaceYesButton.isSelected();
   }

   public boolean hasForeword()
   {
      return forewordYesButton.isSelected();
   }

   public File getPrefaceFile()
   {
      return prefaceTeXButton.isSelected() ? 
             prefaceFileField.getFile() : null ;
   }

   public File getForewordFile()
   {
      return forewordTeXButton.isSelected() ? 
             forewordFileField.getFile() : null ;
   }

   private JRadioButton prefaceTeXButton, prefaceNewButton,
      forewordTeXButton, forewordNewButton, prefaceNoButton,
      prefaceYesButton, forewordNoButton, forewordYesButton;

   private FileField prefaceFileField, forewordFileField;

   private JComponent forewordPanel;

   private NewBookDialog dialog;

   private final static int INDENT = 20;
}

class BookPartPanel extends JPanel
  implements NewBookPanel
{
   public BookPartPanel(NewBookDialog dialog)
   {
      super(null);
      this.dialog = dialog;

      setLayout(new BoxLayout(this, BoxLayout.PAGE_AXIS));
      setAlignmentX(LEFT_ALIGNMENT);

      JLabel label = 
        new JLabel(MakeJmlrBookGUI.getLabel("newbook.parts.query"));
      label.setAlignmentX(LEFT_ALIGNMENT);
      add(label);

      ButtonGroup bg = new ButtonGroup();

      noPartButton = JMLRResources.createRadioButton(
        "newbook.parts", "partno", null, true);
      bg.add(noPartButton);
      add(noPartButton);
      noPartButton.setAlignmentX(LEFT_ALIGNMENT);

      tocButton = JMLRResources.createRadioButton(
        "newbook.parts", "parttoc", null, false);
      bg.add(tocButton);
      add(tocButton);
      tocButton.setAlignmentX(LEFT_ALIGNMENT);

      yesPartButton = JMLRResources.createRadioButton(
        "newbook.parts", "partyes", null, false);
      bg.add(yesPartButton);
      add(yesPartButton);
      yesPartButton.setAlignmentX(LEFT_ALIGNMENT);

      add(Box.createVerticalStrut(INDENT));
      add(Box.createVerticalGlue());

      label = new JLabel(MakeJmlrBookGUI.getLabel("newbook.parts.appendices"));
      label.setAlignmentX(LEFT_ALIGNMENT);
      add(label);

      JTextArea textArea = new JTextArea();
      textArea.setRows(2);
      textArea.setLineWrap(true);
      textArea.setWrapStyleWord(true);
      textArea.setEditable(false);
      textArea.setOpaque(false);
      textArea.setAlignmentX(LEFT_ALIGNMENT);
      textArea.setText(MakeJmlrBookGUI.getLabel("newbook.parts.apdinfo"));
      add(textArea);

      Dimension dim = textArea.getMaximumSize();
      dim.height = (int)textArea.getPreferredSize().getHeight();
      textArea.setMaximumSize(dim);

      bg = new ButtonGroup();

      noApdButton = JMLRResources.createRadioButton(
         "newbook.parts", "apdno", null, true);
      noApdButton.setAlignmentX(LEFT_ALIGNMENT);
      bg.add(noApdButton);
      add(noApdButton);

      yesApdButton = JMLRResources.createRadioButton(
         "newbook.parts", "apdyes", null, false);
      yesApdButton.setAlignmentX(LEFT_ALIGNMENT);
      bg.add(yesApdButton);
      add(yesApdButton);

      add(Box.createVerticalGlue());
      add(Box.createVerticalStrut(200));
   }

   public boolean hasAppendices()
   {
      return yesApdButton.isSelected();
   }

   public int getPartConfig()
   {
      if (noPartButton.isSelected())
      {
         return PARTS_NO;
      }
      else if (tocButton.isSelected())
      {
         return PARTS_TOC;
      }

      return PARTS_YES;
   }

   public void updatePanel()
   {
      yesPartButton.setEnabled(dialog.isCiML());
   }

   public boolean canMoveForwards() {return true;}

   public boolean hasNextButton() {return true;}

   public boolean hasBackButton() {return true;}

   private static final int INDENT = 20;

   private JRadioButton noPartButton, yesPartButton, tocButton,
    noApdButton, yesApdButton;

   public static final int PARTS_NO=0, PARTS_TOC=1, PARTS_YES=2;

   private NewBookDialog dialog;
}

class ArticlesPanel extends JPanel
  implements ActionListener,NewBookPanel
{
   public ArticlesPanel(BookPartPanel bookPartPanel, NewBookDialog dialog)
   {
      super(null);
      this.bookPartPanel = bookPartPanel;
      this.dialog = dialog;

      setLayout(new BoxLayout(this, BoxLayout.PAGE_AXIS));
      setAlignmentX(LEFT_ALIGNMENT);

      JLabel label;
      Box box;

      cardLayout = new CardLayout();
      topComp = new JPanel(cardLayout);
      topComp.setAlignmentX(LEFT_ALIGNMENT);
      add(topComp);

      partVector = new Vector<PartListPanel>();

      partComp = new JPanel(new BorderLayout());
      partComp.setAlignmentX(LEFT_ALIGNMENT);
      cardLayout.addLayoutComponent(partComp, "part");
      topComp.add(partComp);

      partListsComp = Box.createVerticalBox();
      partListsComp.setAlignmentX(LEFT_ALIGNMENT);
      partComp.add(partListsComp, BorderLayout.CENTER);

      JButton addButton = JMLRResources.createActionButton("newbook.articles",
       "addpart", this, null);
      JPanel buttonPanel = new JPanel();
      buttonPanel.add(addButton);
      partComp.add(buttonPanel, BorderLayout.SOUTH);

      addPart();

      mainComp = Box.createVerticalBox();
      mainComp.setAlignmentX(LEFT_ALIGNMENT);
      mainComp.setBorder(BorderFactory.createLoweredSoftBevelBorder());
      cardLayout.addLayoutComponent(mainComp, "main");
      topComp.add(mainComp);

      label = new JLabel(MakeJmlrBookGUI.getLabel("newbook.articles.main"));
      label.setAlignmentX(LEFT_ALIGNMENT);
      mainComp.add(label);

      mainImpPapersPanel = new ImportPapersPanel(dialog);
      mainComp.add(mainImpPapersPanel, BorderLayout.CENTER);

      apdComp = new JPanel(new BorderLayout());
      apdComp.setAlignmentX(LEFT_ALIGNMENT);
      apdComp.setBorder(BorderFactory.createLoweredSoftBevelBorder());
      add(apdComp);

      label = new JLabel(MakeJmlrBookGUI.getLabel("newbook.articles.appendix"));
      label.setAlignmentX(LEFT_ALIGNMENT);
      apdComp.add(label, BorderLayout.NORTH);

      apdImpPapersPanel = new ImportPapersPanel(dialog);
      apdComp.add(apdImpPapersPanel, BorderLayout.CENTER);
   }

   private void addPart()
   {
      if (partVector.size() > 0)
      {
         partVector.lastElement().setMovePartDownEnabled(true);
      }

      PartListPanel partPanel = new PartListPanel(this, partVector.size(),
         dialog);
      partPanel.setBorder(BorderFactory.createLoweredSoftBevelBorder());

      partVector.add(partPanel);
      partListsComp.add(partPanel);

      partPanel = partVector.firstElement();

      if (partVector.size() == 1)
      {
         partPanel.setMovePartUpEnabled(false);
         partPanel.setRemoveEnabled(false);
      }
      else
      {
         partPanel.setRemoveEnabled(true);
      }
   }

   public void removePart(int index)
   {
      partListsComp.remove(index);
      partVector.remove(index);

      for (int i = index, n = partVector.size(); i < n; i++)
      {
         PartListPanel partPanel = partVector.get(i);
         partPanel.setIndex(i);
      }

      if (partVector.size() == 1)
      {
         partVector.firstElement().setRemoveEnabled(false);
      }

      partVector.firstElement().setMovePartUpEnabled(false);
      partVector.lastElement().setMovePartDownEnabled(false);

      revalidate();
   }

   public void actionPerformed(ActionEvent evt)
   {
      String action = evt.getActionCommand();

      if (action == null) return;

      if (action.equals("addpart"))
      {
         addPart();
         revalidate();
      }
   }

   public boolean hasNextButton() {return true;}

   public boolean hasBackButton() {return true;}

   public void updatePanel()
   {
      if (bookPartPanel.getPartConfig() == BookPartPanel.PARTS_NO)
      {
         cardLayout.show(topComp, "main");

         mainImpPapersPanel.updatePanel();
      }
      else
      {
         cardLayout.show(topComp, "part");

         for (PartListPanel panel : partVector)
         {
            panel.updatePanel();
         }
      }

      apdComp.setVisible(bookPartPanel.hasAppendices());

      if (apdComp.isVisible())
      {
         apdImpPapersPanel.updatePanel();
      }
   }

   public Vector<PartListPanel> getPartLists()
   {
      return partVector;
   }

   public ImportPapersPanel getMainImportPapers()
   {
      return mainImpPapersPanel;
   }

   public ImportPapersPanel getAppendixImportPapers()
   {
      return apdImpPapersPanel;
   }

   public boolean canMoveForwards()
   {
      int total = 0;

      if (bookPartPanel.getPartConfig() == BookPartPanel.PARTS_NO)
      {
         total = mainImpPapersPanel.countPapers();
      }
      else
      {
         for (PartListPanel panel : partVector)
         {
            total += panel.countPapers();
         }
      }

      if (total > 0)
      {
         return true;
      }

      JMLRResources.error(this,
          MakeJmlrBookGUI.getLabel("error.newbook.no_main_papers"));

      return false;
   }

   private BookPartPanel bookPartPanel;

   private JComponent partComp, mainComp, apdComp, topComp,
     partListsComp;

   private ImportPapersPanel mainImpPapersPanel, apdImpPapersPanel;

   private Vector<PartListPanel> partVector;

   private CardLayout cardLayout;

   private NewBookDialog dialog;
}

class PartListPanel extends JPanel
  implements ActionListener
{
   public PartListPanel(ArticlesPanel articlesPanel, int partNum,
     NewBookDialog dialog)
   {
      super(new BorderLayout());
      this.articlesPanel = articlesPanel;
      this.partNum = partNum;

      JPanel topPanel = new JPanel();
      add(topPanel, BorderLayout.NORTH);

      String partNumStr = ""+(partNum+1);

      partLabel = new JLabel(MakeJmlrBookGUI.getLabelWithValue(
         "newbook.articles.part", partNumStr));
      partLabel.setDisplayedMnemonic(
         partNumStr.charAt(partNumStr.length()-1));
      topPanel.add(partLabel);

      partField = new JTextField(20);
      partField.setText(MakeJmlrBookGUI.getLabel(
        "newbook.articles.partdefault"));
      partLabel.setLabelFor(partField);
      topPanel.add(partField);

      movePartUpButton = JMLRResources.createUpButton(this);
      topPanel.add(movePartUpButton);

      movePartDownButton = JMLRResources.createDownButton(this);
      topPanel.add(movePartDownButton);
      movePartDownButton.setEnabled(false);

      removePartButton = JMLRResources.createImageButton(
         "newbook.articles", "removepart", "general/Delete24.gif",
         this);

      topPanel.add(removePartButton);

      impPapersPanel = new ImportPapersPanel(dialog);

      add(impPapersPanel, BorderLayout.CENTER);

      partField.requestFocusInWindow();
   }

   public void setIndex(int index)
   {
      partNum = index;
      String partNumStr = ""+(partNum+1);

      partLabel.setText(MakeJmlrBookGUI.getLabelWithValue(
         "newbook.articles.part", partNumStr));
      partLabel.setDisplayedMnemonic(
         partNumStr.charAt(partNumStr.length()-1));
   }

   public void setRemoveEnabled(boolean enabled)
   {
      removePartButton.setEnabled(enabled);
   }

   public void setMovePartUpEnabled(boolean enabled)
   {
      movePartUpButton.setEnabled(enabled);
   }

   public void setMovePartDownEnabled(boolean enabled)
   {
      movePartDownButton.setEnabled(enabled);
   }

   public void actionPerformed(ActionEvent evt)
   {
      String action = evt.getActionCommand();

      if (action == null) return;

      if (action.equals("removepart"))
      {
         if (JOptionPane.showConfirmDialog(this,
          MakeJmlrBookGUI.getLabelWithValue("newbook.articles.confirmrmpart",
            partField.getText()),
          MakeJmlrBookGUI.getLabel("message.confirm"),
          JOptionPane.YES_NO_OPTION)
          == JOptionPane.YES_OPTION)
         {
            articlesPanel.removePart(partNum);
         }
      }
   }

   public void updatePanel()
   {
      impPapersPanel.updatePanel();
   }

   public ImportPapersPanel getImportPapersPanel()
   {
      return impPapersPanel;
   }

   public String getPartTitle()
   {
      return partField.getText();
   }

   public int countPapers()
   {
      return impPapersPanel.countPapers();
   }

   private ArticlesPanel articlesPanel;

   private JLabel partLabel;
   private JTextField partField;
   private JButton removePartButton, movePartUpButton, movePartDownButton;

   private ImportPapersPanel impPapersPanel;

   private int partNum;
}

class ImportPapersPanel extends JPanel
{
   public ImportPapersPanel(NewBookDialog dialog)
   {
      super(null);
      this.dialog = dialog;

      setLayout(new BoxLayout(this, BoxLayout.PAGE_AXIS));

      addRow();
   }

   public void addRow()
   {
      add(new ImportPaperPanel(this, dialog));
      updateButtons();
   }

   public void addRow(int index)
   {
      add(new ImportPaperPanel(this, dialog), index);
      updateButtons();
   }

   public void addRowAfter(Component comp)
   {
      int index = getComponentIndex(comp)+1;

      if (index >= getComponentCount())
      {
         addRow();
      }
      else
      {
         addRow(index);
      }
   }

   public void updatePanel()
   {
      for (int i = 0, n = getComponentCount(); i < n; i++)
      {
         ImportPaperPanel panel = (ImportPaperPanel)getComponent(i);
         panel.updatePanel();
      }
   }

   public int getComponentIndex(Component comp)
   {
      for (int i = 0, n = getComponentCount(); i < n; i++)
      {
         if (getComponent(i) == comp) return i;
      }

      return -1;
   }

   public void moveRowUp(int index)
   {
      if (index <= 0) return;

      Component comp = getComponent(index);

      remove(comp);
      add(comp, index-1);

      updateButtons();
   }

   public void moveRowDown(int index)
   {
      int n = getComponentCount();

      if (index >= n-1) return;

      Component comp = getComponent(index);
      add(comp, index+1);

      updateButtons();
   }

   public void updateButtons()
   {
      for (int i = 0, n = getComponentCount()-1; i <= n; i++)
      {
         ImportPaperPanel panel = (ImportPaperPanel)getComponent(i);

         panel.enableUpButton(i > 0);
         panel.enableDownButton(i < n);
         panel.enableRemoveButton(n > 0);
      }

      revalidate();
   }

   public int countPapers()
   {
      int total = 0;

      for (int i = 0, n = getComponentCount()-1; i <= n; i++)
      {
         ImportPaperPanel panel = (ImportPaperPanel)getComponent(i);

         if (panel.hasFile()) total++;
      }

      return total;
   }

   private NewBookDialog dialog;
}

class ImportPaperPanel extends JPanel
   implements ActionListener
{
   public ImportPaperPanel(ImportPapersPanel parent, NewBookDialog dialog)
   {
      super(null);
      setLayout(new BoxLayout(this, BoxLayout.PAGE_AXIS));
      setBorder(BorderFactory.createRaisedSoftBevelBorder());

      this.parent = parent;
      this.dialog = dialog;

      JPanel mainPanel = new JPanel(new FlowLayout(FlowLayout.LEADING));
      add(mainPanel);

      mainPanel.add(
         new JLabel(MakeJmlrBookGUI.getLabel("newbook.articles.import")));

      fileField = new FileField(this, dialog.getFileChooser());
      mainPanel.add(fileField);

      mainPanel.add(Box.createHorizontalStrut(20));

      upButton = JMLRResources.createUpButton(this);
      mainPanel.add(upButton);
      downButton = JMLRResources.createDownButton(this);
      mainPanel.add(downButton);

      mainPanel.add(JMLRResources.createImageButton("newbook.articles",
        "addarticle", "table/RowInsertAfter24.gif", this));

      mainPanel.add(Box.createHorizontalStrut(20));

      removeButton = JMLRResources.createRemoveButton(this);
      mainPanel.add(removeButton);

      reprintPanel = new JPanel();
      add(reprintPanel);

      reprintBox = new JCheckBox(
         MakeJmlrBookGUI.getLabel("newbook.articles.isreprint"), false);
      reprintBox.addActionListener(this);
      reprintBox.setActionCommand("isreprint");

      reprintPanel.add(reprintBox);

      pageRangeLabel = new JLabel(
         MakeJmlrBookGUI.getLabel("newbook.articles.reprintpages"));

      reprintPanel.add(pageRangeLabel);

      startPageField = new JTextField(3);
      reprintPanel.add(startPageField);

      pageRangeDashLabel = new JLabel("--");

      reprintPanel.add(pageRangeDashLabel);

      endPageField = new JTextField(3);
      reprintPanel.add(endPageField);

      enableReprintPages(false);

      setMaximumSize(getPreferredSize());

      reprintPanel.setVisible(dialog.isCiML());
   }

   public void actionPerformed(ActionEvent evt)
   {
      String action = evt.getActionCommand();

      if (action == null) return;

      if (action.equals("remove"))
      {
         parent.remove(this);
         parent.updateButtons();
      }
      else if (action.equals("addarticle"))
      {
         parent.addRowAfter(this);
      }
      else if (action.equals("up"))
      {
         parent.moveRowUp(parent.getComponentIndex(this));
      }
      else if (action.equals("down"))
      {
         parent.moveRowDown(parent.getComponentIndex(this));
      }
      else if (action.equals("isreprint"))
      {
         boolean isReprint = reprintBox.isSelected();

         enableReprintPages(isReprint);

         if (isReprint)
         {
            startPageField.requestFocusInWindow();
         }
      }
   }

   public void enableRemoveButton(boolean enabled)
   {
      removeButton.setEnabled(enabled);
   }

   public void enableUpButton(boolean enabled)
   {
      upButton.setEnabled(enabled);
   }

   public void enableDownButton(boolean enabled)
   {
      downButton.setEnabled(enabled);
   }

   public void updatePanel()
   {
      reprintPanel.setVisible(dialog.isCiML());
      revalidate();
   }

   public void enableReprintPages(boolean enabled)
   {
      pageRangeLabel.setEnabled(enabled);
      pageRangeDashLabel.setEnabled(enabled);
      startPageField.setEnabled(enabled);
      endPageField.setEnabled(enabled);
   }

   public File getFile()
   {
      return fileField.getFile();
   }

   public boolean hasFile()
   {
      String filename = fileField.getFileName();

      return (filename != null && !filename.isEmpty());
   }

   public boolean isReprint()
   {
      return reprintPanel.isVisible() && reprintBox.isSelected();
   }

   public String getStartPage()
   {
      return startPageField.getText();
   }

   public String getEndPage()
   {
      return endPageField.getText();
   }

   public Vector<LaTeXFile> parsePreamble(TeXApp app, File bookDir)
      throws IOException,NewBookException,IllegalCharsetNameException
   {
      File file = getFile();

      Path orgParentPath = file.getParentFile().toPath();

      String base = file.getName();

      int idx = base.lastIndexOf(".");

      if (idx > -1)
      {
         base = base.substring(0, idx);
      }

      charset = null;

      Vector<LaTeXFile> styList = null;

      String authortag = null;

      PreambleParser preambleListener = new PreambleParser(app)
      {
         public void newcommand(byte overwrite,
          String type, String csName, boolean isShort,
          int numParams, TeXObject defValue, TeXObject definition)
         throws IOException
         {
         }
      };

      TeXParser parser = new TeXParser(preambleListener);
      parser.parse(file);

      String encoding = preambleListener.getInputEncoding();
      styList = preambleListener.getLoadedPackages();

      if (encoding != null && !encoding.equals("utf8"))
      {
         charset = InputEncSty.getCharSet(encoding);
      }

      TeXObjectList author = preambleListener.getAuthor();

      if (author != null && author.size() > 1)
      {
         if (author.firstElement() instanceof ControlSequence)
         {
            ControlSequence cs = (ControlSequence)author.firstElement();

            idx = 1;

            while (author.size() > idx && 
                   author.get(idx) instanceof Ignoreable)
            {
               idx++;
            }

            if (idx < author.size())
            {
               TeXObject authorName = author.get(idx);

               if (authorName instanceof CharObject
               && ((CharObject)authorName).getCharCode() == '[')
               {
                  for ( ; idx < author.size(); idx++)
                  {
                     TeXObject obj = author.get(idx);

                     if (obj instanceof CharObject
                     && ((CharObject)obj).getCharCode() == ']')
                     {
                        idx++;
                        break;
                     }
                  }

                  while (author.size() > idx && 
                         author.get(idx) instanceof Ignoreable
                       ||author.get(idx) instanceof WhiteSpace)
                  {
                     idx++;
                  }

                  authorName = (idx < author.size() ? author.get(idx) : 
                   (TeXObject)null);
               }

               if (authorName instanceof CharObject
                && cs.getName().equals("name"))
               {
                  TeXObjectList list = new TeXObjectList();

                  for (int j = idx; j < author.size(); j++)
                  {
                     TeXObject obj = author.get(j);

                     if (obj instanceof ControlSequence)
                     {
                        String csName = 
                          ((ControlSequence)obj).getName()
                             .toLowerCase();

                        if (csName.equals("email")
                          ||csName.equals("addr")
                          ||csName.equals("and")
                          ||csName.equals("thanks")
                          ||csName.equals("\\"))
                        {
                           break;
                        }
                     }

                     list.add(obj);
                  }

                  authorName = list;
               }

               if (authorName instanceof TeXObjectList)
               {
                  TeXObjectList list = (TeXObjectList)authorName;
                  authortag = "";

                  for (int j = list.size()-1; j >= 0; j--)
                  {
                     TeXObject obj = list.get(j);

                     if (obj instanceof WhiteSpace
                      && !authortag.equals(""))
                     {
                        break;
                     }

                     if (obj instanceof Ignoreable
                      || obj instanceof ControlSequence)
                     {
                        continue;
                     }

                     if (j > 0 && 
                       list.get(j-1) instanceof ControlSequence
                    && ((ControlSequence)list.get(j-1)).getName().equals("nametag"))
                     {
                        j--;
                        break;
                     }

                     if (obj instanceof Group)
                     {
                        obj = ((Group)obj).toList();
                     }

                     // remove any "von" part

                     String str = obj.toString(parser).replaceFirst("^[a-z]+ ", "");

                     authortag = Normalizer.normalize(
                       str, Normalizer.Form.NFD).replaceAll(
                         "[^a-zA-Z]", "").toLowerCase() + authortag;
                  }

               }
            }
         }
      }

      if (authortag.equals(""))
      {
         authortag = null;
      }

      File impDir = null;

      if (authortag == null)
      {
         impDir = new File(bookDir, file.getParentFile().getName());
      }
      else
      {
         authortag = authortag + (dialog.getYear()%100);

         for (char c = 'a'; c <= 'z'; c++)
         {
            impDir = new File(bookDir, authortag+c);

            if (!impDir.exists())
            {
               authortag = impDir.getName();
               break;
            }
         }
      }

      if (impDir.exists() && !impDir.getName().equals(base))
      {
         impDir = new File(bookDir, base);
      }

      if (!impDir.mkdir())
      {
         throw new NewBookException(MakeJmlrBookGUI.getLabelWithValue(
           "newbook.process.cant_mkdir",
            impDir.getAbsolutePath()));
      }

      impFile = new File(impDir, 
        (authortag == null ? base : authortag)+".tex");

      return styList;
   }

   public File getToImportFile()
   {
      return impFile;
   }

   public Charset getCharSet()
   {
      return charset;
   }

   private File impFile;
   private Charset charset;

   private FileField fileField;

   private JCheckBox reprintBox;

   private JTextField startPageField, endPageField;

   private JLabel pageRangeLabel, pageRangeDashLabel;

   private JComponent reprintPanel;

   private JButton upButton, downButton, removeButton;

   private ImportPapersPanel parent;

   private NewBookDialog dialog;
}

class BookDirectoryPanel extends JPanel
 implements NewBookPanel
{
   public BookDirectoryPanel(NewBookDialog dialog)
   {
      super(null);
      this.dialog = dialog;
      setLayout(new BoxLayout(this, BoxLayout.PAGE_AXIS));

      Box box = Box.createHorizontalBox();
      box.setAlignmentX(Component.LEFT_ALIGNMENT);
      add(box);

      nameField = new JTextField(12);

      int volNum = dialog.getVolume();

      dirField = new FileField(this,
         MakeJmlrBookGUI.getLabelWithValue("newbook.default_dir", 
           volNum == 0 ? "" : ""+volNum),
         dialog.getDirChooser(),
         JFileChooser.DIRECTORIES_ONLY);

      JLabel label = new JLabel(MakeJmlrBookGUI.getLabel("newbook.dir"));
      label.setDisplayedMnemonic(MakeJmlrBookGUI.getMnemonic("newbook.dir"));
      label.setLabelFor(nameField);

      Dimension dim = nameField.getPreferredSize();
      dim.width = Integer.MAX_VALUE;
      nameField.setMaximumSize(dim);

      box.add(label);
      box.add(nameField);

      label = new JLabel(MakeJmlrBookGUI.getLabel("newbook.dir_in"));
      label.setDisplayedMnemonic(MakeJmlrBookGUI.getMnemonic("newbook.dir_in"));
      label.setLabelFor(dirField.getTextField());

      box.add(label);
      box.add(dirField);

      add(Box.createVerticalStrut(10));

      box = Box.createHorizontalBox();
      box.setAlignmentX(Component.LEFT_ALIGNMENT);
      add(box);

      bookFontsArea = new JTextArea(String.format("%s%n%s%n%s%n%s%n%n",
       "\\usepackage{mathptmx}",
       "\\usepackage[scaled=0.9]{helvet}",
       "\\usepackage{courier}",
       "\\usepackage{cmap}"));

      label = new JLabel(MakeJmlrBookGUI.getLabel("newbook.fonts"));
      label.setDisplayedMnemonic(MakeJmlrBookGUI.getMnemonic("newbook.fonts"));
      label.setLabelFor(bookFontsArea);

      dim = bookFontsArea.getPreferredSize();
      dim.width = Integer.MAX_VALUE;

      JScrollPane sp = new JScrollPane(bookFontsArea);
      sp.setMaximumSize(dim);

      label.setAlignmentX(Component.LEFT_ALIGNMENT);
      label.setAlignmentY(Component.TOP_ALIGNMENT);
      sp.setAlignmentY(Component.TOP_ALIGNMENT);

      box.add(label);
      box.add(sp);
      box.setMaximumSize(dim);

   }

   public boolean canMoveForwards() 
   {
      if (nameField.getText().isEmpty())
      {
         JMLRResources.error(this,
            MakeJmlrBookGUI.getLabel("newbook.missing_base"));
         return false;
      }

      File file = getBookDirectory();

      if (file.exists())
      {
         JMLRResources.error(this, 
            MakeJmlrBookGUI.getLabelWithValue("error.file_exists",
            file.getAbsolutePath()));

         return false;
      }

      if (!file.mkdir())
      {
         JMLRResources.error(this,
           MakeJmlrBookGUI.getLabelWithValue("error.io.cant_mkdir",
              file.getAbsolutePath()));

         return false;
      }

      return true;
   }

   public boolean hasNextButton() {return true;}

   public boolean hasBackButton() {return true;}

   public void updatePanel()
   {
      JFileChooser fc = dialog.getDirChooser();

      int volNum = dialog.getVolume();

      nameField.setText(
         MakeJmlrBookGUI.getLabelWithValue("newbook.default_dir", 
           volNum == 0 ? "" : ""+volNum));

      File file = fc.getCurrentDirectory();

      dirField.setFile(file);
   } 

   public File getBookDirectory()
   {
      return new File(dirField.getFile(), nameField.getText());
   }

   public String getBookFonts()
   {
      return bookFontsArea.getText();
   }

   private NewBookDialog dialog;

   private JTextField nameField;

   private FileField dirField;

   private JTextArea bookFontsArea;
}

class ProcessPanel extends JPanel
  implements NewBookPanel,Runnable,ActionListener,LaTeXApp
{
   public ProcessPanel(NewBookDialog dialog)
   {
      super(new BorderLayout());
      this.dialog = dialog;

      errMessageBuffer = new StringBuffer();

      messageArea = new MessagePane();

      add(new JScrollPane(messageArea), BorderLayout.CENTER);

      buttonPanel = new JPanel();

      buttonPanel.add(JMLRResources.createActionButton("newbook.process",
        "copy", this, 
        KeyStroke.getKeyStroke(KeyEvent.VK_C, InputEvent.CTRL_MASK)));

      buttonPanel.add(JMLRResources.createActionButton("newbook.process",
        "view", this, 
        KeyStroke.getKeyStroke(KeyEvent.VK_E, InputEvent.CTRL_MASK)));

      buttonPanel.add(JMLRResources.createActionButton("newbook.process",
        "finish", this, null));

      add(buttonPanel, BorderLayout.SOUTH);
   }

   public void actionPerformed(ActionEvent evt)
   {
      String action = evt.getActionCommand();

      if (action == null) return;

      if (action.equals("copy"))
      {
         messageArea.selectAll();
         messageArea.copy();
      }
      else if (action.equals("finish"))
      {
         dialog.finish(bookFile);
      }
      else if (action.equals("view"))
      {
         if (bookFile == null)
         {
            JMLRResources.error(dialog,
              MakeJmlrBookGUI.getLabel("error.load_failed"));
         }
         else if (bookFile.exists())
         {
            try
            {
               dialog.viewFile(bookFile);
            }
            catch (MalformedURLException e)
            {
               JMLRResources.error(dialog, e);
            }
         }
         else
         {
            JMLRResources.error(dialog,
              MakeJmlrBookGUI.getLabelWithValue("error.file_not_found",
                 bookFile.getAbsolutePath()));
         }
      }
   }

   public boolean canMoveForwards() {return false;}

   public boolean hasNextButton() {return false;}

   public boolean hasBackButton() {return false;}

   public void updatePanel()
   {
      buttonPanel.setVisible(false);

      errMessageBuffer.setLength(0);
      messageArea.clearMessages();

      Thread thread = new Thread(this);
      thread.start();
      thread = null;
   }

   public void copyFile(File src, File dest)
     throws IOException
   {
      dialog.verboseCopyFile(src, dest);
   }

   public void epstopdf(File epsFile, File pdfFile)
     throws IOException
   {
      dialog.epstopdf(epsFile, pdfFile);
   }

   public void wmftoeps(File wmfFile, File epsFile)
     throws IOException
   {
      dialog.wmftoeps(wmfFile, epsFile);
   }

   public String kpsewhich(String name)
     throws IOException
   {
      return dialog.kpsewhich(name);
   }

   public Vector<String> getEditors()
   {
      return dialog.getEditors();
   }

   public void run()
   {
      reader = null;
      writer = null;
      currentFile = null;
      bookDir = dialog.getBookDirectory();
      base = bookDir.getName();
      prefaceFile = null;
      forewordFile = null;

      styNames = new Hashtable<String,PackageInfo>();

      message("h1",
        MakeJmlrBookGUI.getLabelWithValue("newbook.process.title", base));

      hasPreface = dialog.hasPreface();
      hasForeword = dialog.hasForeword();
      hasAppendices = dialog.hasAppendices();

      if (hasPreface || hasForeword)
      {
         doFrontMatter();
      }

      partConfig = dialog.getPartConfig();

      doMainMatter();

      if (hasAppendices)
      {
         doAppendices();
      }

      writeBookFile();

      if (errMessageBuffer.length() > 0)
      {
         JMLRResources.error(dialog, 
            MakeJmlrBookGUI.getLabelWithValue(
              "newbook.completed_with_errors",
              errMessageBuffer.toString()));
      }
      else
      {
         JOptionPane.showMessageDialog(dialog, 
           MakeJmlrBookGUI.getLabel("newbook.completed"),
           MakeJmlrBookGUI.getLabel("newbook.completed.title"),
           JOptionPane.INFORMATION_MESSAGE);
      }

      buttonPanel.setVisible(true);

      message(MakeJmlrBookGUI.getLabel("message.done"));
      dialog.setAppInfo(MakeJmlrBookGUI.getLabel("message.done"));
   }

   private void doFrontMatter()
   {
      message("h2", MakeJmlrBookGUI.getLabel("newbook.frontmatter.label"));
      frontMatterDir = new File(bookDir, FRONTMATTER);

      message(MakeJmlrBookGUI.getLabelWithValue("newbook.process.mkdir",
        frontMatterDir.getAbsolutePath()));

      if (!frontMatterDir.mkdir())
      {
         error(MakeJmlrBookGUI.getLabelWithValue(
           "newbook.process.cant_mkdir",
        frontMatterDir.getAbsolutePath()));

         hasPreface = false;
         hasForeword = false;
      }

      if (hasPreface)
      {
         doPreface();
      }

      if (hasForeword)
      {
         doForeword();
      }
   }

   private void doPreface()
   {
      message("h3",
        MakeJmlrBookGUI.getLabel("newbook.frontmatter.preface.label"));

      prefaceFile = dialog.getPrefaceFile();
      currentFile = prefaceFile;

      try
      {
         if (prefaceFile == null)
         {
            prefaceFile = new File(frontMatterDir, PREFACE+".tex");

            message(MakeJmlrBookGUI.getLabelWithValue("message.writing",
               prefaceFile.getAbsolutePath()));

            writer = new PrintWriter(prefaceFile);

            Vector<String> editors = getEditors();

            String first = editors.firstElement();

            String[] names = first.split(" ");

            first = names[names.length-1].toLowerCase();
            int year = dialog.getYear();

            writer.println("\\begin{preface}["+first+year+"]");
            writer.println("");

            writer.println("\\begin{signoff}{Insert Date Here}");

            for (String editor : editors)
            {
               writer.println("% Editor details: ");
               writer.println("\\Editor{" + editor+"\\\\");
               writer.println("Insert Address Here\\\\");
               writer.println("\\mailto{insert@email.here}}");
            }

            writer.println("\\end{signoff}");
            writer.println("\\end{preface}");
         }
         else
         {
            PrefaceParser prefParser = new PrefaceParser(this,
               frontMatterDir, getEditors());

            TeXParser parser = new TeXParser(prefParser);

            parser.parse(prefaceFile);

            prefaceFile = new File(frontMatterDir, prefaceFile.getName());
         }
      }
      catch (IOException e)
      {
         prefaceFile = null;
         hasPreface = false;
         JMLRResources.error(dialog, e);
      }
      finally
      {
         if (writer != null)
         {
            writer.close();
            writer = null;
         }
      }
   }

   private void doForeword()
   {
      message("h3",
        MakeJmlrBookGUI.getLabel("newbook.frontmatter.foreword.label"));

      forewordFile = dialog.getForewordFile();
      currentFile = forewordFile;

      try
      {
         if (forewordFile == null)
         {
            forewordFile = new File(frontMatterDir, FOREWORD+".tex");

            message(MakeJmlrBookGUI.getLabelWithValue("message.writing",
               forewordFile.getAbsolutePath()));

            writer = new PrintWriter(forewordFile);

            writer.println("\\chapter{Foreword}");

            writer.println();
            writer.println("\\begin{authorsignoff}");
            writer.println("\\Author{Insert Name Here\\\\");
            writer.println("Insert Affiliation Here}");
            writer.println("\\end{authorsignoff}");
         }
         else
         {
            ForewordParser forewordParser = new ForewordParser(this,
               frontMatterDir);

            TeXParser parser = new TeXParser(forewordParser);

            parser.parse(forewordFile);

            forewordFile = new File(frontMatterDir, forewordFile.getName());
         }
      }
      catch (IOException e)
      {
         forewordFile = null;
         hasForeword = false;
         JMLRResources.error(dialog, e);
      }
      finally
      {
         if (writer != null)
         {
            writer.close();
            writer = null;
         }
      }

   }

   private void doMainMatter()
   {
      message("h2", MakeJmlrBookGUI.getLabel("newbook.articles.main"));

      if (partConfig == BookPartPanel.PARTS_NO)
      {
         ImportPapersPanel mainImportPanel = dialog.getMainImportPapers();

         doImportPapers(mainImportPanel);
      }
      else
      {
         Vector<PartListPanel> partPanels = dialog.getPartLists();

         for (PartListPanel partPanel : partPanels)
         {
            message("h2", partPanel.getPartTitle());

            ImportPapersPanel impPanel = partPanel.getImportPapersPanel();

            doImportPapers(impPanel);
         }
      }
   }

   private void doAppendices()
   {
      message("h2", MakeJmlrBookGUI.getLabel("newbook.articles.appendix"));

      ImportPapersPanel impPanel = dialog.getAppendixImportPapers();

      doImportPapers(impPanel);
   }

   public PackageInfo addPackage(String styName, String options)
   {
      PackageInfo pkgInfo;

      if (styName.equals("algpseudocode"))
      {
         styNames.remove("algorithmicx");
      }
      else if (styName.equals("algorithmicx"))
      {
         if (styNames.containsKey("algpseudocode"))
         {
            return null;
         }
      }

      if (styNames.containsKey(styName))
      {
         pkgInfo = styNames.get(styName);

         pkgInfo.addOption(options);
      }
      else
      {
         pkgInfo = new PackageInfo(styName, options, getPackageLocation(styName));

         styNames.put(styName, pkgInfo);
      }

      return pkgInfo;
   }

   private void doImportPapers(ImportPapersPanel impPanel)
   {
      for (int i = 0, n = impPanel.getComponentCount(); i < n; i++)
      {
         ImportPaperPanel paper = (ImportPaperPanel)impPanel.getComponent(i);

         File file = paper.getFile();
         currentFile = file;

         Vector<LaTeXFile> styList = null;

         try
         {
            styList = paper.parsePreamble(this, bookDir);
         }
         catch (Exception e)
         {
            error(e);
            continue;
         }

         File impFile = paper.getToImportFile();

         try
         {
            ArticleParser artParser = new ArticleParser(this,
               styList, impFile);

            TeXParser parser = new TeXParser(artParser);

            artParser.usepackage(null, "graphicx");

            parser.parse(file, paper.getCharSet());
         }
         catch (IOException e)
         {
            error(e);
         }
      }
   }

   private void writeFontFile()
     throws IOException
   {
      File fontFile = new File(bookDir, "bookfonts.tex");

      message(MakeJmlrBookGUI.getLabelWithValue("message.writing",
          fontFile.getAbsolutePath()));

      PrintWriter out = null;

      try
      {
         out = new PrintWriter(fontFile);

         out.println("% Book fonts");

         out.println(dialog.getBookFonts());
      }
      finally
      {
         if (out != null)
         {
            out.close();
            out = null;
         }
      }
   }

   private void writeBookFile()
   {
      try
      {
         writeFontFile();

         String logo = dialog.getLogoImage();

         if (logo != null && !"".equals(logo))
         {
            File logoFile = new File(logo);
            logo = logoFile.getName();

            if (logoFile.exists())
            {
               copyFile(logoFile, new File(bookDir, logo));
            }
         }

         bookFile = new File(bookDir, base+".tex");
         Path bookPath = bookDir.toPath();

         message(MakeJmlrBookGUI.getLabelWithValue("message.writing",
            bookFile.getAbsolutePath()));
         writer = new PrintWriter(bookFile);

         writer.println("% "+MakeJmlrBookGUI.getLabelWithValue(
           "newbook.comment.prehyperref", "\\jmlrprehyperref"));
         writer.println();
         writer.println("\\def\\jmlrprehyperref{%");
         writer.println("% "
           + MakeJmlrBookGUI.getLabel("newbook.comment.import_sty"));

         boolean algorithmic = false;
         boolean algorithmicx = false;

         for (Enumeration<String> en=styNames.keys(); en.hasMoreElements();)
         {
            String key = en.nextElement();

            PackageInfo pkgInfo = styNames.get(key);

            if (key.equals("inputenc"))
            {
               if (!pkgInfo.hasOption("utf8"))
               {
                  message(MakeJmlrBookGUI.getLabelWithValue(
                    "error.inputenc.notutf8", pkgInfo.getOptions()));
               }

               continue;
            }

            if (key.equals("fontenc"))
            {
               if (!pkgInfo.hasOption("T1"))
               {
                  message(MakeJmlrBookGUI.getLabelWithValue(
                    "error.inputenc.notT1", pkgInfo.getOptions()));
               }

               continue;
            }

            if (key.equals("algpseudocode") || key.equals("algorithmicx"))
            {
               algorithmicx = true;
               writer.println(" % "
                + MakeJmlrBookGUI.getLabelWithValue("newbook.comment.patch_sty",
                     key));
               writer.println(" % "
                +MakeJmlrBookGUI.getLabelWithValue(
                    "newbook.comment.save_sty_cs", "algorithm2e"));
               writer.println(" \\let\\ALGOFor\\For");
               writer.println(" \\let\\ALGOForEach\\ForEach");
               writer.println(" \\let\\ALGOForAll\\ForAll");
               writer.println(" \\let\\ALGOWhile\\While");
               writer.println(" \\let\\ALGOReturn\\Return");
               writer.println(" \\let\\ALGOBegin\\Begin");
               writer.println(" \\let\\ALGOIf\\If");
               writer.println(" \\let\\ALGOElse\\Else");
               writer.println(" \\let\\ALGOSwitch\\Switch");
               writer.println(" \\let\\ALGOCase\\Case");
               writer.println(" \\let\\ALGOOther\\Other");

               writer.println(" \\def\\switchtoALGO{%");
               writer.println("  \\let\\For\\ALGOFor");
               writer.println("  \\let\\ForEach\\ALGOForEach");
               writer.println("  \\let\\ForAll\\ALGOForAll");
               writer.println("  \\let\\While\\ALGOWhile");
               writer.println("  \\let\\Return\\ALGOReturn");
               writer.println("  \\let\\Begin\\ALGOBegin");
               writer.println("  \\let\\If\\ALGOIf");
               writer.println("  \\let\\Else\\ALGOElse");
               writer.println("  \\let\\Switch\\ALGOSwitch");
               writer.println("  \\let\\Case\\ALGOCase");
               writer.println("  \\let\\Other\\ALGOOther");
               writer.println(" }");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValues(
                 "newbook.comment.undef_sty_cs", "algorithm2e", key));

               writer.println(" \\undef\\For");
               writer.println(" \\undef\\ForEach");
               writer.println(" \\undef\\ForAll");
               writer.println(" \\undef\\While");
               writer.println(" \\undef\\Return");
               writer.println(" \\undef\\Begin");
               writer.println(" \\undef\\If");
               writer.println(" \\undef\\Else");
               writer.println(" \\undef\\Switch");
               writer.println(" \\undef\\Case");
               writer.println(" \\undef\\Other");

               if (algorithmic)
               {
                  writer.println(" % "+MakeJmlrBookGUI.getLabelWithValues(
                     "newbook.comment.undef_sty_cs", "algorithmic", key));

                  writer.println(" \\undef\\algorithmicindent");
                  writer.println(" \\undef\\algorithmic");
                  writer.println(" \\undef\\endalgorithmic");
                  writer.println(" \\undef\\algorithmiccomment");
                  writer.println(" \\undef\\algorithmicrequire");
                  writer.println(" \\undef\\algorithmicensure");
                  writer.println(" \\undef\\algorithmicend");
                  writer.println(" \\undef\\algorithmicelse");
                  writer.println(" \\undef\\algorithmicfor");
                  writer.println(" \\undef\\algorithmicforall");
                  writer.println(" \\undef\\algorithmicdo");
                  writer.println(" \\undef\\algorithmicwhile");
                  writer.println(" \\undef\\algorithmicloop");
                  writer.println(" \\undef\\algorithmicrepeat");
                  writer.println(" \\undef\\algorithmicuntil");
                  writer.println(" \\undef\\algorithmicreturn");
                  writer.println(" \\undef\\algorithmicif");
                  writer.println(" \\undef\\algorithmicthen");
               }

               writer.print(" ");
               pkgInfo.write(writer);

               writer.println(" \\let\\ALGPSEUDOCODEalgorithmic\\algorithmic");
               writer.println(" \\let\\endALGPSEUDOCODEalgorithmic\\endalgorithmic");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicindent\\algorithmicindent");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmiccomment\\algorithmiccomment");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicrequire\\algorithmicrequire");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicensure\\algorithmicensure");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicend\\algorithmicend");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicelse\\algorithmicelse");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicfor\\algorithmicfor");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicforall\\algorithmicforall");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicdo\\algorithmicdo");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicwhile\\algorithmicwhile");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicloop\\algorithmicloop");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicrepeat\\algorithmicrepeat");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicuntil\\algorithmicuntil");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicreturn\\algorithmicreturn");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicif\\algorithmicif");
               writer.println(" \\let\\ALGPSEUDOCODEalgorithmicthen\\algorithmicthen");
               writer.println(" \\let\\ALGPSEUDOCODEFor\\For");
               writer.println(" \\let\\ALGPSEUDOCODEForEach\\ForEach");
               writer.println(" \\let\\ALGPSEUDOCODEForAll\\ForAll");
               writer.println(" \\let\\ALGPSEUDOCODEWhile\\While");
               writer.println(" \\let\\ALGPSEUDOCODEReturn\\Return");
               writer.println(" \\let\\ALGPSEUDOCODEBegin\\Begin");
               writer.println(" \\let\\ALGPSEUDOCODEIf\\If");
               writer.println(" \\let\\ALGPSEUDOCODEElse\\Else");
               writer.println(" \\let\\ALGPSEUDOCODESwitch\\Switch");
               writer.println(" \\let\\ALGPSEUDOCODECase\\Case");
               writer.println(" \\let\\ALGPSEUDOCODEOther\\Other");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.restore_sty_cs", key));
               writer.println(" \\def\\switchtoALGPSEUDOCODE{%");
               writer.println("  \\let\\algorithmic\\ALGPSEUDOCODEalgorithmic");
               writer.println("  \\let\\endalgorithmic\\endALGPSEUDOCODEalgorithmic");
               writer.println("  \\let\\algorithmicindent\\ALGPSEUDOCODEalgorithmicindent");
               writer.println("  \\let\\algorithmiccomment\\ALGPSEUDOCODEalgorithmiccomment");
               writer.println("  \\let\\algorithmicrequire\\ALGPSEUDOCODEalgorithmicrequire");
               writer.println("  \\let\\algorithmicensure\\ALGPSEUDOCODEalgorithmicensure");
               writer.println("  \\let\\algorithmicend\\ALGPSEUDOCODEalgorithmicend");
               writer.println("  \\let\\algorithmicelse\\ALGPSEUDOCODEalgorithmicelse");
               writer.println("  \\let\\algorithmicfor\\ALGPSEUDOCODEalgorithmicfor");
               writer.println("  \\let\\algorithmicforall\\ALGPSEUDOCODEalgorithmicforall");
               writer.println("  \\let\\algorithmicdo\\ALGPSEUDOCODEalgorithmicdo");
               writer.println("  \\let\\algorithmicwhile\\ALGPSEUDOCODEalgorithmicwhile");
               writer.println("  \\let\\algorithmicloop\\ALGPSEUDOCODEalgorithmicloop");
               writer.println("  \\let\\algorithmicrepeat\\ALGPSEUDOCODEalgorithmicrepeat");
               writer.println("  \\let\\algorithmicuntil\\ALGPSEUDOCODEalgorithmicuntil");
               writer.println("  \\let\\algorithmicreturn\\ALGPSEUDOCODEalgorithmicreturn");
               writer.println("  \\let\\algorithmicif\\ALGPSEUDOCODEalgorithmicif");
               writer.println("  \\let\\algorithmicthen\\ALGPSEUDOCODEalgorithmicthen");
               writer.println(" \\let\\For\\ALGPSEUDOCODEFor");
               writer.println(" \\let\\ForEach\\ALGPSEUDOCODEForEach");
               writer.println(" \\let\\ForAll\\ALGPSEUDOCODEForAll");
               writer.println(" \\let\\While\\ALGPSEUDOCODEWhile");
               writer.println(" \\let\\Return\\ALGPSEUDOCODEReturn");
               writer.println(" \\let\\Begin\\ALGPSEUDOCODEBegin");
               writer.println(" \\let\\If\\ALGPSEUDOCODEIf");
               writer.println(" \\let\\Else\\ALGPSEUDOCODEElse");
               writer.println(" \\let\\Switch\\ALGPSEUDOCODESwitch");
               writer.println(" \\let\\Case\\ALGPSEUDOCODECase");
               writer.println(" \\let\\Other\\ALGPSEUDOCODEOther");
               writer.println(" }%");

            }
            else if (key.equals("algorithmic"))
            {
               algorithmic = true;
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.patch_sty", key));

               if (algorithmic)
               {
                  writer.println(" % "+MakeJmlrBookGUI.getLabelWithValues(
                    "ewbook.comment.undef_sty_cs", "algorithmicx", key));
                  writer.println(" \\undef\\algorithmicindent");
                  writer.println(" \\undef\\algorithmiccomment");
                  writer.println(" \\undef\\algorithmicrequire");
                  writer.println(" \\undef\\algorithmicensure");
                  writer.println(" \\undef\\algorithmicend");
                  writer.println(" \\undef\\algorithmicelse");
                  writer.println(" \\undef\\algorithmicfor");
                  writer.println(" \\undef\\algorithmicforall");
                  writer.println(" \\undef\\algorithmicdo");
                  writer.println(" \\undef\\algorithmicwhile");
                  writer.println(" \\undef\\algorithmicloop");
                  writer.println(" \\undef\\algorithmicrepeat");
                  writer.println(" \\undef\\algorithmicuntil");
                  writer.println(" \\undef\\algorithmicreturn");
                  writer.println(" \\undef\\algorithmicif");
                  writer.println(" \\undef\\algorithmicthen");
                  writer.println(" \\undef\\algorithmicendif");
                  writer.println(" \\undef\\algorithmicendfor");
                  writer.println(" \\undef\\algorithmicendwhile");
                  writer.println(" \\undef\\algorithmicendloop");
                  writer.println(" \\undef\\algorithmicelsif");
                  writer.println(" \\undef\\algorithmicprint");
                  writer.println(" \\undef\\algorithmicand");
                  writer.println(" \\undef\\algorithmicor");
                  writer.println(" \\undef\\algorithmicxor");
                  writer.println(" \\undef\\algorithmicnot");
                  writer.println(" \\undef\\algorithmicto");
                  writer.println(" \\undef\\algorithmicinputs");
                  writer.println(" \\undef\\algorithmicoutputs");
                  writer.println(" \\undef\\algorithmicglobals");
                  writer.println(" \\undef\\algorithmicbody");
                  writer.println(" \\undef\\algorithmictrue");
                  writer.println(" \\undef\\algorithmicfalse");
               }

               writer.print(" ");

               pkgInfo.write(writer);

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.save_sty_cs", key));
               writer.println("\\letcs{\\ALGORITHMICialgorithmic}{\\string\\algorithmic}%");
               writer.println(" \\let\\ALGORITHMICalgorithmic\\algorithmic");
               writer.println(" \\let\\endALGORITHMICalgorithmic\\endalgorithmic");
               writer.println(" \\let\\ALGORITHMICalgorithmicindent\\algorithmicindent");
               writer.println(" \\let\\ALGORITHMICalgorithmiccomment\\algorithmiccomment");
               writer.println(" \\let\\ALGORITHMICalgorithmicrequire\\algorithmicrequire");
               writer.println(" \\let\\ALGORITHMICalgorithmicensure\\algorithmicensure");
               writer.println(" \\let\\ALGORITHMICalgorithmicend\\algorithmicend");
               writer.println(" \\let\\ALGORITHMICalgorithmicelse\\algorithmicelse");
               writer.println(" \\let\\ALGORITHMICalgorithmicfor\\algorithmicfor");
               writer.println(" \\let\\ALGORITHMICalgorithmicforall\\algorithmicforall");
               writer.println(" \\let\\ALGORITHMICalgorithmicdo\\algorithmicdo");
               writer.println(" \\let\\ALGORITHMICalgorithmicwhile\\algorithmicwhile");
               writer.println(" \\let\\ALGORITHMICalgorithmicloop\\algorithmicloop");
               writer.println(" \\let\\ALGORITHMICalgorithmicrepeat\\algorithmicrepeat");
               writer.println(" \\let\\ALGORITHMICalgorithmicuntil\\algorithmicuntil");
               writer.println(" \\let\\ALGORITHMICalgorithmicreturn\\algorithmicreturn");
               writer.println(" \\let\\ALGORITHMICalgorithmicif\\algorithmicif");
               writer.println(" \\let\\ALGORITHMICalgorithmicthen\\algorithmicthen");
               writer.println(" \\let\\ALGORITHMICalgorithmicendif\\algorithmicendif");
               writer.println(" \\let\\ALGORITHMICalgorithmicendfor\\algorithmicendfor");
               writer.println(" \\let\\ALGORITHMICalgorithmicendwhile\\algorithmicendwhile");
               writer.println(" \\let\\ALGORITHMICalgorithmicendloop\\algorithmicendloop");
               writer.println(" \\let\\ALGORITHMICalgorithmicelsif\\algorithmicelsif");
               writer.println(" \\let\\ALGORITHMICalgorithmicprint\\algorithmicprint");
               writer.println(" \\let\\ALGORITHMICalgorithmicand\\algorithmicand");
               writer.println(" \\let\\ALGORITHMICalgorithmicor\\algorithmicor");
               writer.println(" \\let\\ALGORITHMICalgorithmicxor\\algorithmicxor");
               writer.println(" \\let\\ALGORITHMICalgorithmicnot\\algorithmicnot");
               writer.println(" \\let\\ALGORITHMICalgorithmicto\\algorithmicto");
               writer.println(" \\let\\ALGORITHMICalgorithmicinputs\\algorithmicinputs");
               writer.println(" \\let\\ALGORITHMICalgorithmicoutputs\\algorithmicoutputs");
               writer.println(" \\let\\ALGORITHMICalgorithmicglobals\\algorithmicglobals");
               writer.println(" \\let\\ALGORITHMICalgorithmicbody\\algorithmicbody");
               writer.println(" \\let\\ALGORITHMICalgorithmictrue\\algorithmictrue");
               writer.println(" \\let\\ALGORITHMICalgorithmicfalse\\algorithmicfalse");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.restore_sty_cs", key));
               writer.println(" \\def\\switchtoALGORITHMIC{");
               writer.println("  \\cslet{\\string\\algorithmic}{\\ALGORITHMICialgorithmic}%");
               writer.println("  \\let\\algorithmic\\ALGORITHMICalgorithmic");
               writer.println("  \\let\\endalgorithmic\\endALGORITHMICalgorithmic");
               writer.println("  \\let\\algorithmicindent\\ALGORITHMICalgorithmicindent");
               writer.println("  \\let\\algorithmiccomment\\ALGORITHMICalgorithmiccomment");
               writer.println("  \\let\\algorithmicrequire\\ALGORITHMICalgorithmicrequire");
               writer.println("  \\let\\algorithmicensure\\ALGORITHMICalgorithmicensure");
               writer.println("  \\let\\algorithmicend\\ALGORITHMICalgorithmicend");
               writer.println("  \\let\\algorithmicelse\\ALGORITHMICalgorithmicelse");
               writer.println("  \\let\\algorithmicfor\\ALGORITHMICalgorithmicfor");
               writer.println("  \\let\\algorithmicforall\\ALGORITHMICalgorithmicforall");
               writer.println("  \\let\\algorithmicdo\\ALGORITHMICalgorithmicdo");
               writer.println("  \\let\\algorithmicwhile\\ALGORITHMICalgorithmicwhile");
               writer.println("  \\let\\algorithmicloop\\ALGORITHMICalgorithmicloop");
               writer.println("  \\let\\algorithmicrepeat\\ALGORITHMICalgorithmicrepeat");
               writer.println("  \\let\\algorithmicuntil\\ALGORITHMICalgorithmicuntil");
               writer.println("  \\let\\algorithmicreturn\\ALGORITHMICalgorithmicreturn");
               writer.println("  \\let\\algorithmicif\\ALGORITHMICalgorithmicif");
               writer.println("  \\let\\algorithmicthen\\ALGORITHMICalgorithmicthen");
               writer.println("  \\let\\algorithmicendif\\ALGORITHMICalgorithmicendif");
               writer.println("  \\let\\algorithmicendfor\\ALGORITHMICalgorithmicendfor");
               writer.println("  \\let\\algorithmicendwhile\\ALGORITHMICalgorithmicendwhile");
               writer.println("  \\let\\algorithmicendloop\\ALGORITHMICalgorithmicendloop");
               writer.println("  \\let\\algorithmicelsif\\ALGORITHMICalgorithmicelsif");
               writer.println("  \\let\\algorithmicprint\\ALGORITHMICalgorithmicprint");
               writer.println("  \\let\\algorithmicand\\ALGORITHMICalgorithmicand");
               writer.println("  \\let\\algorithmicor\\ALGORITHMICalgorithmicor");
               writer.println("  \\let\\algorithmicxor\\ALGORITHMICalgorithmicxor");
               writer.println("  \\let\\algorithmicnot\\ALGORITHMICalgorithmicnot");
               writer.println("  \\let\\algorithmicto\\ALGORITHMICalgorithmicto");
               writer.println("  \\let\\algorithmicinputs\\ALGORITHMICalgorithmicinputs");
               writer.println("  \\let\\algorithmicoutputs\\ALGORITHMICalgorithmicoutputs");
               writer.println("  \\let\\algorithmicglobal\\ALGORITHMICalgorithmicglobal");
               writer.println("  \\let\\algorithmicbody\\ALGORITHMICalgorithmicbody");
               writer.println("  \\let\\algorithmictrue\\ALGORITHMICalgorithmictrue");
               writer.println("  \\let\\algorithmicfalse\\ALGORITHMICalgorithmicfalse");
               writer.println("  \\undef\\REQUIRE");
               writer.println("  \\undef\\ENSURE");
               writer.println("  \\undef\\PRINT");
               writer.println("  \\undef\\TRUE");
               writer.println("  \\undef\\FALSE");
               writer.println("  \\undef\\AND");
               writer.println("  \\undef\\OR");
               writer.println("  \\undef\\NOT");
               writer.println("  \\undef\\STATE");
               writer.println("  \\undef\\STMT");
               writer.println("  \\undef\\COMMENT");
               writer.println("  \\undef\\INPUTS");
               writer.println("  \\undef\\OUTPUTS");
               writer.println("  \\undef\\ENDOUTPUTS");
               writer.println("  \\undef\\GLOBALS");
               writer.println("  \\undef\\BODY");
               writer.println("  \\undef\\ENDBODY");
               writer.println("  \\undef\\IF");
               writer.println("  \\undef\\ENDIF");
               writer.println("  \\undef\\ELSE");
               writer.println("  \\undef\\ELSIF");
               writer.println("  \\undef\\FOR");
               writer.println("  \\undef\\FORALL");
               writer.println("  \\undef\\ENDFOR");
               writer.println("  \\undef\\WHILE");
               writer.println("  \\undef\\ENDWHILE");
               writer.println("  \\undef\\REPEAT");
               writer.println("  \\undef\\UNTIL");
               writer.println("  \\undef\\LOOP");
               writer.println("  \\undef\\ENDLOOP");
               writer.println("  \\undef\\RETURN");

               writer.println(" }");

            }
            else if (key.equals("mdwlist"))
            {
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.patch_sty", key));
               pkgInfo.write(writer);
               writer.println("  \\csletcs{mdwenumerate*}{enumerate*}");
               writer.println("  \\csletcs{endmdwenumerate*}{endenumerate*}");
               writer.println("  \\csundef{enumerate*}");
               writer.println("  \\csundef{endenumerate*}");
            }
            else if (key.equals("amssymb"))
            {
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.symbol_conflict", key));
               writer.println("  \\undef\\ggg");
               writer.println("  \\undef\\lll");
               writer.println("  \\undef\\checkmark");
               writer.println("  \\undef\\centerdot");
               writer.println("  \\undef\\veebar");
               writer.println("  \\undef\\barwedge");
               writer.println("  \\undef\\doublebarwedge");
               writer.println("  \\undef\\circledS");
               writer.println("  \\undef\\dotplus");
               writer.println("  \\undef\\complement");
               writer.println("  \\undef\\divideontimes");
               writer.println("  \\undef\\thicksim");
               writer.println("  \\undef\\thickapprox");
               writer.println("  \\undef\\hslash");

               pkgInfo.write(writer);

               writer.println("  \\let\\AMSSYMBggg\\ggg");
               writer.println("  \\let\\AMSSYMBlll\\lll");
               writer.println("  \\let\\AMSSYMBcheckmark\\checkmark");
               writer.println("  \\let\\AMSSYMBcenterdot\\centerdot");
               writer.println("  \\let\\AMSSYMBveebar\\veebar");
               writer.println("  \\let\\AMSSYMBbarwedge\\barwedge");
               writer.println("  \\let\\AMSSYMBdoublebarwedge\\doublebarwedge");
               writer.println("  \\let\\AMSSYMBcircledS\\circledS");
               writer.println("  \\let\\AMSSYMBdotplus\\dotplus");
               writer.println("  \\let\\AMSSYMBcomplement\\complement");
               writer.println("  \\let\\AMSSYMBdivideontimes\\divideontimes");
               writer.println("  \\let\\AMSSYMBthicksim\\thicksim");
               writer.println("  \\let\\AMSSYMBthickapprox\\thickapprox");
               writer.println("  \\let\\AMSSYMBhslash\\hslash");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.restore_sty_cs", key));

               writer.println("  \\def\\switchtoAMSSYMB{%");
               writer.println("   \\let\\ggg\\AMSSYMBggg");
               writer.println("   \\let\\lll\\AMSSYMBlll");
               writer.println("   \\let\\checkmark\\AMSSYMBcheckmark");
               writer.println("   \\let\\centerdot\\AMSSYMBcenterdot");
               writer.println("   \\let\\veebar\\AMSSYMBveebar");
               writer.println("   \\let\\barwedge\\AMSSYMBbarwedge");
               writer.println("   \\let\\doublebarwedge\\AMSSYMBdoublebarwedge");
               writer.println("   \\let\\circledS\\AMSSYMBcircledS");
               writer.println("   \\let\\dotplus\\AMSSYMBdotplus");
               writer.println("   \\let\\complement\\AMSSYMBcomplement");
               writer.println("   \\let\\divideontimes\\AMSSYMBdivideontimes");
               writer.println("   \\let\\thicksim\\AMSSYMBthicksim");
               writer.println("   \\let\\thickapprox\\AMSSYMBthickapprox");
               writer.println("   \\let\\hslash\\AMSSYMBhslash");
               writer.println("  }%");
            }
            else if (key.equals("stmaryrd"))
            {
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.symbol_conflict", key));
               writer.println("  \\undef\\baro");
               writer.println("  \\undef\\lightning");
               writer.println("  \\let\\LATEXbigtriangledown\\bigtriangledown");
               writer.println("  \\let\\LATEXbigtriangleup\\bigtriangleup");

               pkgInfo.write(writer);

               writer.println("  \\let\\STMARYRDbaro\\baro");
               writer.println("  \\let\\STMARYRDlightning\\lightning");
               writer.println("  \\let\\STMARYRDbigtriangledown\\bigtriangledown");
               writer.println("  \\let\\STMARYRDbigtriangleup\\bigtriangleup");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.restore_sty_cs", key));

               writer.println("  \\def\\switchtoSTMARYRD{%");
               writer.println("   \\let\\baro\\STMARYRDbaro");
               writer.println("   \\let\\lightning\\STMARYRDlightning");
               writer.println("   \\let\\bigtriangledown\\STMARYRDbigtriangledown");
               writer.println("   \\let\\bigtriangleup\\STMARYRDbigtriangleup");
               writer.println("  }%");
               writer.println("  \\let\\bigtriangledown\\LATEXbigtriangledown");
               writer.println("  \\let\\bigtriangleup\\LATEXbigtriangleup");
            }
            else if (key.equals("wasysym"))
            {
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.symbol_conflict", key));

               writer.println("  \\undef\\Circle");
               writer.println("  \\undef\\lightning");
               writer.println("  \\undef\\Square");

               pkgInfo.write(writer);

               writer.println("  \\let\\WASYSYMCircle\\Circle");
               writer.println("  \\let\\WASYSYMlightning\\lightning");
               writer.println("  \\let\\WASYSYMSquare\\Square");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.restore_sty_cs", key));

               writer.println("  \\def\\switchtoWASYSYM{%");
               writer.println("   \\let\\Circle\\WASYSYMCircle");
               writer.println("   \\let\\lightning\\WASYSYMlightning");
               writer.println("   \\let\\Square\\WASYSYMSquare");
               writer.println("  }%");
            }
            else if (key.equals("mathabx"))
            {
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.symbol_conflict", key));

               writer.println("  \\undef\\ggg");
               writer.println("  \\undef\\lll");
               writer.println("  \\undef\\Sun");

               pkgInfo.write(writer);

               writer.println("  \\let\\MATHABXggg\\ggg");
               writer.println("  \\let\\MATHABXlll\\lll");
               writer.println("  \\let\\MATHABXSun\\Sun");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.restore_sty_cs", key));

               writer.println("  \\def\\switchtoMATHABX{%");
               writer.println("   \\let\\ggg\\MATHABXggg");
               writer.println("   \\let\\lll\\MATHABXlll");
               writer.println("   \\let\\Sun\\MATHABXSun");
               writer.println("  }%");
            }
            else if (key.equals("marvosym"))
            {
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.symbol_conflict", key));

               writer.println("  \\undef\\Cross");
               writer.println("  \\undef\\Letter");
               writer.println("  \\undef\\Lightning");
               writer.println("  \\undef\\Sun");
               writer.println("  \\undef\\Email");

               pkgInfo.write(writer);

               writer.println("  \\let\\MARVOSYMCross\\Cross");
               writer.println("  \\let\\MARVOSYMLetter\\Letter");
               writer.println("  \\let\\MARVOSYMLightning\\Lightning");
               writer.println("  \\let\\MARVOSYMSun\\Sun");
               writer.println("  \\let\\MARVOSYMEmail\\Email");
               writer.println("  \\undef\\Email");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                  "newbook.comment.restore_sty_cs", key));

               writer.println("  \\def\\switchtoMARVOSYM{%");
               writer.println("   \\let\\Cross\\MARVOSYMCross");
               writer.println("   \\let\\Letter\\MARVOSYMLetter");
               writer.println("   \\let\\Lightning\\MARVOSYMLightning");
               writer.println("   \\let\\Sun\\MARVOSYMSun");
               writer.println("  }%");
            }
            else if (key.equals("bbding"))
            {
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                  "newbook.comment.symbol_conflict", key));

               writer.println("  \\undef\\Cross");
               writer.println("  \\undef\\Square");
               writer.println("  \\undef\\TriangleDown");
               writer.println("  \\undef\\TriangleUp");

               pkgInfo.write(writer);

               writer.println("  \\let\\BBDINGCross\\Cross");
               writer.println("  \\let\\BBDINGSquare\\Square");
               writer.println("  \\let\\BBDINGTriangleDown\\TriangleDown");
               writer.println("  \\let\\BBDINGTriangleUp\\TriangleUp");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                  "newbook.comment.restore_sty_cs", key));

               writer.println("  \\def\\switchtoBBDING{%");
               writer.println("   \\let\\Cross\\BBDINGCross");
               writer.println("   \\let\\Square\\BBDINGSquare");
               writer.println("   \\let\\TriangleDown\\BBDINGTriangleDown");
               writer.println("   \\let\\TriangleUp\\BBDINGTriangleUp");
               writer.println("  }%");
            }
            else if (key.equals("ifsym"))
            {
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.symbol_conflict", key));

               writer.println("  \\undef\\Circle");
               writer.println("  \\undef\\Cross");
               writer.println("  \\undef\\Letter");
               writer.println("  \\undef\\Lightning");
               writer.println("  \\undef\\Square");
               writer.println("  \\undef\\Sun");
               writer.println("  \\undef\\TriangleDown");
               writer.println("  \\undef\\TriangleUp");

               pkgInfo.write(writer);

               writer.println("  \\let\\IFSYMCircle\\Circle");
               writer.println("  \\let\\IFSYMCross\\Cross");
               writer.println("  \\let\\IFSYMLetter\\Letter");
               writer.println("  \\let\\IFSYMLightning\\Lightning");
               writer.println("  \\let\\IFSYMSquare\\Square");
               writer.println("  \\let\\IFSYMSun\\Sun");
               writer.println("  \\let\\IFSYMTriangleDown\\TriangleDown");
               writer.println("  \\let\\IFSYMTriangleUp\\TriangleUp");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                  "newbook.comment.restore_sty_cs", key));

               writer.println("  \\def\\switchtoIFSYM{%");
               writer.println("   \\let\\Circle\\IFSYMCircle");
               writer.println("   \\let\\Cross\\IFSYMCross");
               writer.println("   \\let\\Letter\\IFSYMLetter");
               writer.println("   \\let\\Lightning\\IFSYMLightning");
               writer.println("   \\let\\Square\\IFSYMSquare");
               writer.println("   \\let\\Sun\\IFSYMSun");
               writer.println("   \\let\\TriangleDown\\IFSYMTriangleDown");
               writer.println("   \\let\\TriangleUp\\IFSYMTriangleUp");
               writer.println("  }%");
            }
            else if (key.equals("dingbat"))
            {
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                  "newbook.comment.symbol_conflict", key));

               writer.println("  \\undef\\checkmark");

               pkgInfo.write(writer);

               writer.println("  \\let\\DINGBATcheckmark\\checkmark");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                  "newbook.comment.restore_sty_cs", key));

               writer.println("  \\def\\switchtoDINGBAT{%");
               writer.println("   \\let\\checkmark\\DINGBATcheckmark");
               writer.println("  }%");
            }
            else if (key.equals("wsuipa"))
            {
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                  "newbook.comment.symbol_conflict", key));

               writer.println("  \\undef\\baro");

               pkgInfo.write(writer);

               writer.println("  \\let\\WSUIPAbaro\\baro");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                  "newbook.comment.restore_sty_cs", key));

               writer.println("  \\def\\switchtoWSUIPA{%");
               writer.println("   \\let\\baro\\WSUIPAbaro");
               writer.println("  }%");
            }
            else if (key.equals("MnSymbol"))
            {
               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                  "newbook.comment.symbol_conflict", key));

               writer.println("  \\undef\\ggg");
               writer.println("  \\undef\\lll");
               writer.println("  \\undef\\checkmark");
               writer.println("  \\undef\\centerdot");
               writer.println("  \\undef\\veebar");
               writer.println("  \\undef\\barwedge");
               writer.println("  \\undef\\doublebarwedge");
               writer.println("  \\undef\\circledS");
               writer.println("  \\undef\\dotplus");
               writer.println("  \\undef\\complement");
               writer.println("  \\undef\\divideontimes");
               writer.println("  \\undef\\thicksim");
               writer.println("  \\undef\\thickapprox");
               writer.println("  \\undef\\hslash");

               pkgInfo.write(writer);

               writer.println("  \\let\\MNSYMBOLggg\\ggg");
               writer.println("  \\let\\MNSYMBOLlll\\lll");
               writer.println("  \\let\\MNSYMBOLcheckmark\\checkmark");
               writer.println("  \\let\\MNSYMBOLcenterdot\\centerdot");
               writer.println("  \\let\\MNSYMBOLveebar\\veebar");
               writer.println("  \\let\\MNSYMBOLbarwedge\\barwedge");
               writer.println("  \\let\\MNSYMBOLdoublebarwedge\\doublebarwedge");
               writer.println("  \\let\\MNSYMBOLcircledS\\circledS");
               writer.println("  \\let\\MNSYMBOLdotplus\\dotplus");
               writer.println("  \\let\\MNSYMBOLcomplement\\complement");
               writer.println("  \\let\\MNSYMBOLdivideontimes\\divideontimes");
               writer.println("  \\let\\MNSYMBOLthicksim\\thicksim");
               writer.println("  \\let\\MNSYMBOLthickapprox\\thickapprox");
               writer.println("  \\let\\MNSYMBOLhslash\\hslash");

               writer.println(" % "+MakeJmlrBookGUI.getLabelWithValue(
                  "newbook.comment.restore_sty_cs", key));

               writer.println("  \\def\\switchtoMNSYMBOL{%");
               writer.println("   \\let\\ggg\\MNSYMBOLggg");
               writer.println("   \\let\\lll\\MNSYMBOLlll");
               writer.println("   \\let\\checkmark\\MNSYMBOLcheckmark");
               writer.println("   \\let\\centerdot\\MNSYMBOLcenterdot");
               writer.println("   \\let\\veebar\\MNSYMBOLveebar");
               writer.println("   \\let\\barwedge\\MNSYMBOLbarwedge");
               writer.println("   \\let\\doublebarwedge\\MNSYMBOLdoublebarwedge");
               writer.println("   \\let\\circledS\\MNSYMBOLcircledS");
               writer.println("   \\let\\dotplus\\MNSYMBOLdotplus");
               writer.println("   \\let\\complement\\MNSYMBOLcomplement");
               writer.println("   \\let\\divideontimes\\MNSYMBOLdivideontimes");
               writer.println("   \\let\\thicksim\\MNSYMBOLthicksim");
               writer.println("   \\let\\thickapprox\\MNSYMBOLthickapprox");
               writer.println("   \\let\\hslash\\MNSYMBOLhslash");
               writer.println("  }%");
            }
            else
            {
               writer.print(" ");

               if (!pkgInfo.isInstalled())
               {
                  writer.println("% "+MakeJmlrBookGUI.getLabelWithValue(
                    "newbook.comment.not_installed", key));
                  writer.print("%");
               }

               pkgInfo.write(writer);
            }
         }

         writer.println("}");

         writer.println();

         boolean isCiML = dialog.isCiML();

         writer.println("% "+MakeJmlrBookGUI.getLabelWithValues(
           "newbook.comment.generated", MakeJmlrBookGUI.appName,
            MakeJmlrBookGUI.appVersion));

         if (isCiML)
         {
            writer.println("% "
              +MakeJmlrBookGUI.getLabel("newbook.comment.ciml"));
            writer.println("\\documentclass[wcp,7x10]{jmlrbook}");
         }
         else
         {
            writer.println("% "
              +MakeJmlrBookGUI.getLabel("newbook.comment.jmlrwcp"));
            writer.println("\\documentclass[wcp]{jmlrbook}");
         }

         writer.println();
         writer.println("\\input{bookfonts}");
         writer.println();

         if (isCiML)
         {
            writer.println("% "
               +MakeJmlrBookGUI.getLabel("newbook.comment.title"));
         }

         String title = dialog.getBookTitle();

         String shortTitle = dialog.getBookShortTitle();

         writer.print("\\title");

         if (!shortTitle.isEmpty())
         {
            writer.print("["+shortTitle+"]");
         }

         writer.print("{");
         writer.print(title);
         writer.println("}");
         writer.println();

         Vector<String> editors = dialog.getEditors();

         writer.print("\\author");

         int n = editors.size();
         String firstEd = editors.firstElement();

         if (n > 1)
         {
            writer.print("["+firstEd+" et al.]");
         }

         writer.print("{"+firstEd);

         for (int i = 1; i < n; i++)
         {
            writer.print(i == n-1 ? " and " : ", ");
            writer.print(editors.get(i));
         }

         writer.println("}");

         if (isCiML)
         {
            writer.println();
            writer.println("\\usepackage[utf8]{inputenc}");
            writer.println("\\usepackage[T1]{fontenc}");

            writer.println();
            writer.println("\\ifprint{}{\\usepackage{bookmark}}% "
             + MakeJmlrBookGUI.getLabel("newbook.comment.load_last"));
            writer.println();

            writer.println("\\subtitle{"+dialog.getBookSubTitle()+"}");
            writer.println("\\volume{"+dialog.getVolume()+"}");
            writer.println("%\\logo{\\includegraphics{ciml-logo}}");
            writer.println("\\team{}");
            writer.println("\\productioneditor{"
              +dialog.getProductionEditor()+"}");

            writer.println();
            writer.println("% "+MakeJmlrBookGUI.getLabelWithValue(
               "newbook.comment.jmlrvolume", "\\jmlrvolume{n}"));
            writer.println("%\\jmlrvolume{0}");
         }
         else
         {
            writer.println("\\usepackage[utf8]{inputenc}");
            writer.println("\\usepackage[T1]{fontenc}");

            writer.println();
            writer.println("\\ifprint{}{\\usepackage{bookmark}}% "
             + MakeJmlrBookGUI.getLabel("newbook.comment.load_last"));
            writer.println();

            writer.println("\\jmlrvolume{"+dialog.getVolume()+"}");

            int year = dialog.getYear();

            writer.println("\\jmlryear{"+year+"}");

            writer.println();

            String month = dialog.getMonth();
            String day = dialog.getDay();

            writer.printf("\\date{%s, %d}%n", (day.isEmpty() ? month :
             month+" "+day), year);

            String location = dialog.getJMLRLocation();

            if (location != null)
            {
               writer.println("\\jmlrlocation{"+location+"}");
            }

            writer.println();

            if (logo == null || "".equals(logo))
            {
               writer.println("% "+MakeJmlrBookGUI.getLabelWithValue(
                 "newbook.comment.imagefile", "imageFile"));
               writer.println("% \\logo[url]{\\includegraphics{imageFile}}");
            }
            else
            {
               String logoUrl = dialog.getLogoUrl();

               writer.print("\\logo");

               if (logoUrl != null)
               {
                  writer.print("["+logoUrl+"]");
               }

               writer.printf("{\\includegraphics{%s}}%n", logo);
            }
         }

         writer.println();
         writer.println("\\begin{document}");
         writer.println("\\maketitle");

         if (isCiML)
         {
            writer.println("\\frontmatter");

            if (hasForeword)
            {
               writeInputCs(bookPath, forewordFile);
            }

            if (hasPreface)
            {
               writeInputCs(bookPath, prefaceFile);
            }

            writer.println("\\tableofcontents");

            writer.println("\\mainmatter");
         }
         else
         {
            writer.println("\\frontmatter");
            writer.println("\\tableofcontents");

            writer.println("\\mainmatter");

            if (hasPreface)
            {
               writeInputCs(bookPath, prefaceFile);
            }

         }

         writeMainMatter(bookPath);

         if (hasAppendices)
         {
            writeAppendices(bookPath);
         }

         writer.println("\\end{document}");
      }
      catch (IOException e)
      {
         JMLRResources.error(dialog, e);
      }
      finally
      {
         if (writer != null)
         {
            writer.close();
            writer = null;
         }
      }
   }

   protected void writeInputCs(Path basePath, File file)
     throws IOException
   {
      String path = MakeJmlrBookGUI.getTeXPath(basePath, file.toPath());

      if (path.endsWith(".tex"))
      {
         path = path.substring(0, path.length()-4);
      }

      writer.println("\\input{"+path+"}");
   }

   protected void writeMainMatter(Path bookPath)
     throws IOException
   {
      if (partConfig == BookPartPanel.PARTS_NO)
      {
         ImportPapersPanel mainImportPanel = dialog.getMainImportPapers();

         writePapers(bookPath, mainImportPanel);
      }
      else
      {
         Vector<PartListPanel> partPanels = dialog.getPartLists();

         for (PartListPanel partPanel : partPanels)
         {
            if (partConfig == BookPartPanel.PARTS_TOC)
            {
               writer.print("\\addtocpart");
            }
            else
            {
               writer.print("\\part");
            }

            writer.println("{"+partPanel.getPartTitle()+"}");

            ImportPapersPanel impPanel = partPanel.getImportPapersPanel();

            writePapers(bookPath, impPanel);
         }
      }
   }

   protected void writeAppendices(Path bookPath)
     throws IOException
   {
      writer.println();
      writer.println("\\appendix");

      ImportPapersPanel impPanel = dialog.getAppendixImportPapers();

      writePapers(bookPath, impPanel);
   }

   protected void writePapers(Path bookPath, ImportPapersPanel impPanel)
     throws IOException
   {
      writer.println();
      writer.println("\\begin{jmlrpapers}");

      for (int i = 0, n = impPanel.getComponentCount(); i < n; i++)
      {
         ImportPaperPanel paper = (ImportPaperPanel)impPanel.getComponent(i);

         File file = paper.getToImportFile();

         if (file == null)
         {
            String ref = paper.getFile().toString();

            error(MakeJmlrBookGUI.getLabelWithValue("newbook.process.null_imp",
               ref));

            MakeJmlrBookGUI.debug(ref+": null file reference for destination");
            continue;
         }

         String parent = MakeJmlrBookGUI.getTeXPath(bookPath,
           file.getParentFile().toPath());

         if (parent == null) parent = ".";

         String base = file.getName();

         if (base.endsWith(".tex"))
         {
            base = base.substring(0, base.length()-4);
         }

         if (paper.isReprint())
         {
            writer.print("\\importpubpaper{");
            writer.print(parent+"}{");
            writer.print(base);
            writer.print("}{");
            writer.print(paper.getStartPage());
            writer.print("--");
            writer.print(paper.getEndPage());
            writer.println("}");
         }
         else if (dialog.isCiML())
         {
            writer.print("\\importarticle{");
            writer.print(parent+"}{");
            writer.print(base);
            writer.println("}");
         }
         else
         {
            writer.print("\\importpaper{");
            writer.print(parent+"}{");
            writer.print(base);
            writer.println("}");
         }
      }

      writer.println("\\end{jmlrpapers}");
      writer.println();
   }

   protected void parsing(File file)
   {
      message(MakeJmlrBookGUI.getLabelWithValue("message.parsing",
        file.getAbsolutePath()));
   }

   protected String tag(String text)
   {
      return "<b><tt>"+text+"</tt></b>";
   }

   protected void message(String tag, String msg)
   {
      messageArea.update("<"+tag+">"+msg+"</"+tag+">");
   }

   protected void message(String msg)
   {
      message("p", msg);
   }

   protected void error(String msg)
   {
      errMessageBuffer.append('\n');
      errMessageBuffer.append(currentFile.getAbsolutePath()
        +": "+msg.replaceAll("<[^>]+>", ""));

      messageArea.update("<p><font color=\"red\">"+msg+"</font></p>");
   }

   public void warning(TeXParser parser, String msg)
   {
      File file = parser.getCurrentFile();
      int lineNum = parser.getLineNumber();

      if (lineNum > 0)
      {
         msg = MakeJmlrBookGUI.getLabelWithValues("error.line", ""+lineNum, msg);

         if (file != null)
         {
            msg = MakeJmlrBookGUI.getLabelWithValues("error.file",
              file.getName(), msg);
         }
      }

      messageArea.update("<p><font color=\"red\">"+msg+"</font></p>");
   }

   public void message(int msgType, String arg)
   {
      switch (msgType)
      {
         case TeXApp.MESSAGE_READING:
            message(MakeJmlrBookGUI.getLabelWithValue("message.parsing", arg));
         return;
         case TeXApp.MESSAGE_WRITING:
            message(MakeJmlrBookGUI.getLabelWithValue("message.writing", arg));
         return;
      }

      message(arg);
   }

   public void substituting(TeXParser parser, String original, String replacement)
   {
      String msg;

      if (replacement.isEmpty())
      {
         msg = MakeJmlrBookGUI.getLabelWithValue("warning.removing", 
             tag(original));
      }
      else
      {
         msg = MakeJmlrBookGUI.getLabelWithValues("warning.substituting",
              tag(original), tag(replacement));
      }

      int lineNumber = parser.getLineNumber();

      if (lineNumber > 0)
      {
         msg = MakeJmlrBookGUI.getLabelWithValues("error.line",
          ""+lineNumber, msg);
      }

      File file = parser.getCurrentFile();

      if (file != null)
      {
         msg = MakeJmlrBookGUI.getLabelWithValues("error.file",
           file.getName(), msg);
      }

      message(msg);
   }

   public void error(Exception e)
   {
      if (e instanceof LaTeXSyntaxException)
      {
         error(MakeJmlrBookGUI.getErrorMessage((LaTeXSyntaxException)e));
      }
      else if (e instanceof TeXSyntaxException)
      {
         error(MakeJmlrBookGUI.getErrorMessage((TeXSyntaxException)e));
      }
      else if (e instanceof NewBookException)
      {
         error(((NewBookException)e).getDetails());
      }
      else
      {
         error(e.getMessage());
      }
   }


   protected File getPackageLocation(String name)
   {
      return dialog.getPackageLocation(name);
   }

   public boolean isProblemSty(String name)
   {
      for (int i = 0; i < PROBLEM_STYS.length; i++)
      {
         if (PROBLEM_STYS[i].equals(name)) return true;
      }

      return false;
   }

   public boolean isIgnoreSty(String name)
   {
      for (int i = 0; i < IGNORE_STYS.length; i++)
      {
         if (IGNORE_STYS[i].equals(name)) return true;
      }

      return false;
   }

   private NewBookDialog dialog;

   private MessagePane messageArea;

   private JPanel buttonPanel;

   private StringBuffer errMessageBuffer;

   private Hashtable<String,PackageInfo> styNames;

   public static final String[] PROBLEM_STYS = new String[]
   {
      "epsfig", "psfig", "subfig", "theorem", "ntheorem",
      "amsthm", "pdfpages", "geometry", "tabularx", "algorithm"
   };

   // Packages that can be skipped because they're already loaded by
   // jmlr.cls

   public static final String[] IGNORE_STYS = new String[]
   {
      "url", "amsmath", "amssymb", "natbib", "graphicx", "xcolor",
      "algorithm2e", "hyperref", "nameref", "etoolbox", "calc",
      "xkeyval", "setspace"
   };

   private Hashtable<File,File> importFiles;

   private BufferedReader reader;
   private PrintWriter writer;
   private File currentFile;
   private File bookDir, bookFile;
   private String base;
   private File prefaceFile, forewordFile, frontMatterDir;
   private int partConfig;

   private boolean hasPreface, hasForeword, hasAppendices;

   private static final String FRONTMATTER="FrontMatter";
   private static final String PREFACE="preface";
   private static final String FOREWORD="foreword";
}

class TeXImageFilter implements FilenameFilter
{
    public TeXImageFilter(String baseName)
    {
       int idx = baseName.lastIndexOf(".");

       if (idx == -1)
       {
          this.baseName = baseName;
       }
       else
       {
          this.baseName = baseName.substring(0, idx);
       }
    }

    public boolean accept(File dir, String name)
    {
       int idx = name.lastIndexOf(".");

       if (idx == -1) return name.equals(baseName);

       String ext = name.substring(idx+1).toLowerCase();

       name = name.substring(0, idx);

       if (name.endsWith("-gray"))
       {
          name = name.substring(0, name.lastIndexOf("-"));
       }

       if (!name.equals(baseName)) return false;

       return (ext.equals("png") 
        || ext.equals("pdf")
        || ext.equals("jpg")
        || ext.equals("jpeg")
        || ext.equals("gif")
        || ext.equals("eps")
        || ext.equals("ps"));
    }

   private String baseName;
}
