/*
    Copyright (C) 2013 Nicola L.C. Talbot
    www.dickimaw-books.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
package com.dickimawbooks.makejmlrbookgui;

import java.io.*;
import java.util.Calendar;
import java.util.Hashtable;
import java.util.Vector;
import java.text.DateFormat;

import com.dickimawbooks.texparserlib.*;
import com.dickimawbooks.texparserlib.primitives.*;
import com.dickimawbooks.texparserlib.latex.*;
import com.dickimawbooks.texparserlib.latex.jmlr.*;
import com.dickimawbooks.texparserlib.html.*;
import com.dickimawbooks.texparserlib.aux.*;

public class Toc2Html extends L2HConverter
{
   protected Toc2Html(TeXApp app, Book book)
   {
      super(app);
      this.book = book;
      setIsInDocEnv(true);
   }

   protected void addPredefined()
   {
      super.addPredefined();

      putControlSequence(new TocPart());
      putControlSequence(new ContentsLine());
      putControlSequence(new ArticlePagesRef());
   }

   public void par() throws IOException
   {
   }

   public void warn(int lineNum, String msg)
   {
      ((BookTeXApp)getTeXApp()).warn(lineNum, msg);
   }

   public void unknown(int lineNum, String name) throws IOException
   {
      warn(lineNum, 
         MakeJmlrBookGUI.getLabelWithValue("error.undefined", "\\"+name));
   }

   public static void convert(PrintWriter writer, 
     BookTeXApp app, Book book)
     throws IOException
   {
      Toc2Html listener = new Toc2Html(app, book);
      TeXReader reader = null;

      try
      {
         listener.setWriter(writer);
         listener.setParser(new TeXParser(listener));

         reader = new TeXReader(book.getTocFile());

         MakeJmlrBookGUI.debug("Toc2Html parsing: "+book.getTocFile());

         LaTeXCls cls = new JmlrBookCls(null, listener, false);

         cls.load(null);

         listener.parser.parse(reader);
      }
      finally
      {
         listener.setWriter(null);
         if (reader != null)
         {
            reader.close();
         }
      }
   }

   public BookDocument getDocument(String label)
   {
      setCurrentLabel(label);
      return book.getImportFile(label);
   }

   public String getCurrentLabel()
   {
      return currentLabel;
   }

   public void setCurrentLabel(String label)
   {
      currentLabel = label;
   }

   private Book book;
   private String currentLabel;
}

class TocPart extends ControlSequence
{
   public TocPart()
   {
      this("tocpart");
   }

   public TocPart(String name)
   {
      super(name);
   }

   public Object clone()
   {
      return new TocPart(getName());
   }

   public void process(TeXParser parser)
     throws IOException
   {
      TeXObject arg = parser.popNextArg();

      ((Toc2Html)parser.getListener()).write("<tr><th>");

      arg.process(parser);

      ((Toc2Html)parser.getListener()).writeln("</th></tr>");
   }

   public void process(TeXParser parser, TeXObjectList list)
     throws IOException
   {
      TeXObject arg = list.popArg(parser);

      ((Toc2Html)parser.getListener()).write("<tr><th>");

      arg.process(parser, list);

      ((Toc2Html)parser.getListener()).writeln("</th></tr>");
   }

}

class ContentsLine extends ControlSequence
{
   public ContentsLine()
   {
      this("contentsline");
   }

   public ContentsLine(String name)
   {
      super(name);
   }

   public Object clone()
   {
      return new ContentsLine(getName());
   }

   public void process(TeXParser parser)
     throws IOException
   {
      TeXObject arg = parser.popNextArg();

      TeXObjectList expanded = null;

      if (arg instanceof Expandable)
      {
         expanded = ((Expandable)arg).expandfully(parser);
      }

      String type;

      if (expanded == null)
      {
         type = arg.toString(parser);
      }
      else
      {
         type = expanded.toString(parser);
      }

      arg = parser.popNextArg();

      parser.popNextArg();
      parser.popNextArg();

      write(parser, type, arg);
   }

   public void process(TeXParser parser, TeXObjectList list)
     throws IOException
   {
      TeXObject arg = list.popArg(parser);

      TeXObjectList expanded = null;

      if (arg instanceof Expandable)
      {
         expanded = ((Expandable)arg).expandfully(parser, list);
      }

      String type;

      if (expanded == null)
      {
         type = arg.toString(parser);
      }
      else
      {
         type = expanded.toString(parser);
      }

      arg = list.popArg(parser);

      list.popArg(parser);
      list.popArg(parser);

      write(parser, type, arg);
   }

   private void write(TeXParser parser, String type, TeXObject arg)
     throws IOException
   {
      Toc2Html listener = (Toc2Html)parser.getListener();
      Writer writer = listener.getWriter();

      if (type.equals("papertitle"))
      {
         listener.writeln("<tr><td><dl><dt>");

         arg.process(parser);

         listener.writeln("</dt>");
      }
      else if (type.equals("chapterauthor"))
      {
         writer.write(String.format("<dd>%n"));

         if (arg instanceof TeXObjectList)
         {
            writer.write("<b><i>");

            TeXObjectList list = (TeXObjectList)arg;

            while (!list.isEmpty())
            {
               TeXObject obj = list.popArg(parser);

               if (obj instanceof CharObject)
               {
                  if (((CharObject)obj).getCharCode() == (int)';')
                  {
                     writer.write(";");
                     break;
                  }
               }
               else if (obj instanceof TeXCsRef)
               {
                  obj = listener.getControlSequence(((TeXCsRef)obj).getName());
               }

               obj.process(parser, list);
            }

            writer.write("</i></b>");

            list.process(parser);
         }
         else
         {
            arg.process(parser);
         }

         listener.writeln("<br>");

         String label = listener.getCurrentLabel();

         if (label == null)
         {
            writer.write("[abs] [pdf]");
         }
         else
         {
            BookDocument doc = listener.getDocument(label);

            File pdf = doc.getPdfFile();

            String base = pdf.getName();

            int idx = base.lastIndexOf(".");

            if (idx > -1)
            {
               base = base.substring(0, idx);
            }

            writer.write(String.format("<a href=\"%s.html\">[abs]</a>%n", base));
            writer.write(String.format("<a href=\"%s.pdf\">[pdf]</a>%n", base));
         }

         writer.write("</dd>");
         writer.write(String.format("</dl></td></tr>%n"));
      }
   }
}

class ArticlePagesRef extends ControlSequence
{
   public ArticlePagesRef()
   {
      this("articlepagesref");
   }

   public ArticlePagesRef(String name)
   {
      super(name);
   }

   public Object clone()
   {
      return new ArticlePagesRef(getName());
   }

   public void process(TeXParser parser)
     throws IOException
   {
      TeXObject arg = parser.popNextArg();

      TeXObjectList expanded = null;

      if (arg instanceof Expandable)
      {
         expanded = ((Expandable)arg).expandfully(parser);
      }

      String label;

      if (expanded == null)
      {
         label = arg.toString(parser);
      }
      else
      {
         label = expanded.toString(parser);
      }

      write(parser, label);
   }

   public void process(TeXParser parser, TeXObjectList list)
     throws IOException
   {
      TeXObject arg = list.popArg(parser);

      TeXObjectList expanded = null;

      if (arg instanceof Expandable)
      {
         expanded = ((Expandable)arg).expandfully(parser, list);
      }

      String label;

      if (expanded == null)
      {
         label = arg.toString(parser);
      }
      else
      {
         label = expanded.toString(parser);
      }

      write(parser, label);
   }

   private void write(TeXParser parser, String label)
     throws IOException
   {
      BookDocument doc = ((Toc2Html)parser.getListener()).getDocument(label);

      if (doc == null)
      {
         parser.getListener().getWriteable().write("??--??");
      }
  
      int start = doc.getStartPage();
      int end = doc.getEndPage();

      parser.getListener().getWriteable().write(""+start);

      if (start != end)
      {
         parser.getListener().getWriteable().write("&#x2013;");
         parser.getListener().getWriteable().write(""+end);
      }
   }
}

